/* global describe */
/* global beforeEach */
/* global inject */
/* global expect */
/* global describe */
"use strict";
describe("Test controller", function () {
    var ctrl, $scope, pcLocalStorage;

    // Load the 'controllers' module.
    beforeEach(module('controllers'));

    // Load the 'services' module.
    beforeEach(module('services'));

    // Inject the $controller, $rootScope, and $localStorage.
    beforeEach(inject(function ($controller, $rootScope, $localStorage) {
        // Create a new scope that's a child of the $rootScope
        $scope = $rootScope.$new();

        localStorage.setItem('pocketcardEula', '1');
        // Create the controller
        ctrl = $controller('MainCtrl', {
            $scope: $scope
        });

        pcLocalStorage = $localStorage;
    }));

    // Cleanup.
    afterEach(function () {
        _.each(pcLocalStorage.findAll(), function (card) {
            pcLocalStorage.destroy(card);
        });
    });

    describe("Page title", function () {
        it("should be set to 'Pocket Card'.", function () {
            expect($scope.title).toEqual('Pocket Card');
        });
    });

    describe("_getRows", function () {
        describe("_getRows()", function () {
            it("should return an empty array", function () {
                var rows = _getRows();
                expect(rows.length).toEqual(0);
            });
        });

        describe("_getRows([1])", function () {
            it("should return an array of arrays", function () {
                var rows = _getRows([1]);
                expect(rows[0] instanceof Array).toBe(true);
            });
        });

        describe("_getRows([1,2,3])", function () {
            it("should return an array length 2", function () {
                expect(_getRows([1, 2, 3]).length).toEqual(2);
            });
        });
    });

    describe("_checkVersionChange", function () {
        beforeEach(function () {
            var card = {
                id: 1,
                version: "1.0"
            };
            pcLocalStorage.create(card);
        });

        it("should not throw execption when any of the arguments are null.", function () {
            expect(_checkVersionChange).not.toThrow();
        });

        it("should not throw execption when 1st argument is an empty array.", function () {
            var card = pcLocalStorage.find({
                id: 1
            });
            var wrapper = function () {
                _checkVersionChange([], [card], pcLocalStorage);
            };
            expect(wrapper).not.toThrow();
        });

        it("should update the cards version number.", function () {
            var serverCards = [];
            var card = {
                id: 1,
                version: "1.1"
            };
            serverCards.push(card);
            _checkVersionChange(serverCards, [pcLocalStorage.find({
                id: 1
            })], pcLocalStorage);
            card = pcLocalStorage.find(card);

            expect(card.version).toEqual("1.1");
        });
    });

    describe("_addCards", function () {
        it("should add a new card to local storage", function () {
            var serverCards = [];
            var card = {
                id: 1,
            };

            serverCards.push(card);
            _addCards(serverCards, [1], pcLocalStorage);
            card = pcLocalStorage.find({
                id: 1
            });

            expect(card).not.toBeNull();
        });

        it("should add a new card to local storage with a state of new", function () {
            var serverCards = [];
            var card = {
                id: 1,
            };

            serverCards.push(card);
            _addCards(serverCards, [1], pcLocalStorage);
            card = pcLocalStorage.find({
                id: 1
            });

            expect(card.isNew).toBe(true);
        });

        it("should add multiply new card to local storage.", function () {
            var serverCards = [];
            var card1 = {
                id: 1,
                version: "1.0"
            };
            var card2 = {
                id: 2,
                version: "1.0"
            };

            serverCards.push(card1);
            serverCards.push(card2);

            _addCards(serverCards, [1, 2], pcLocalStorage);
            var card = pcLocalStorage.find({
                id: 1
            });
            expect(card).not.toBeNull();

            card = pcLocalStorage.find({
                id: 2
            });
            expect(card).not.toBeNull();
        });
    });

    //TODO: these tests need to be fixed since _saveCard does not save a card not already in it collection (or the _saveCard method needs to be fixed)
    // describe("_saveCard", function () {
    //     var card = {
    //         id: 1
    //     };

    //     it("should save card to local storage.", function () {
    //         _saveCard(card, $scope, pcLocalStorage, 1);
    //         card = pcLocalStorage.find(card);

    //         expect(card.isSaved).toBe(true);
    //     });

    //     describe("", function () {
    //         it("should remove card from local storage if card already exist in local storage.", function () {
    //             pcLocalStorage.create(card);
    //             _saveCard(card, $scope, pcLocalStorage, 1);
    //             card = pcLocalStorage.find(card);

    //             expect(card.isSaved).toBe(false);
    //         });
    //     });
    // });

    describe("_removeCards", function () {
        beforeEach(function () {
            pcLocalStorage.create({
                id: 1
            });
            pcLocalStorage.create({
                id: 2
            });
        });

        it("should remove cards from local storage.", function () {
            var deviceAllCards = pcLocalStorage.findAll();
            _removeCards(deviceAllCards, [1, 2], pcLocalStorage);
            deviceAllCards = pcLocalStorage.findAll();

            expect(_.isEmpty(deviceAllCards)).toBe(true);
        });

        it("should return a collection containing cards removed from local storage.", function () {
            var deviceAllCards = pcLocalStorage.findAll();
            deviceAllCards = _removeCards(deviceAllCards, [1, 2], pcLocalStorage);

            expect(_.isEmpty(deviceAllCards)).toBe(false);
        });
    });

    describe("_smartUpdate", function () {
        beforeEach(function () {
            pcLocalStorage.create({
                id: 1,
                version: "1.0"
            });
            pcLocalStorage.create({
                id: 2,
                version: "1.0"
            });
        });

        it("should add a card to local storage base on cards returned from the server", function () {
            var allCards = pcLocalStorage.findAll();
            allCards.push({
                id: 3
            });
            _smartUpdate($scope, pcLocalStorage, allCards, 1);

            allCards = pcLocalStorage.findAll();
            expect(allCards.length).toEqual(3);
        });

        it("should remove a card from local storage base on cards returned from the server", function () {
            var allCards = pcLocalStorage.findAll();
            allCards.pop();
            _smartUpdate($scope, pcLocalStorage, allCards, 1);

            allCards = pcLocalStorage.findAll();
            expect(allCards.length).toEqual(1);
        });

        it("should return cards removed when card is removed from local storage", function () {
            var allCards = pcLocalStorage.findAll();
            allCards.pop();
            var removedCards = _smartUpdate($scope, pcLocalStorage, allCards, 1);

            allCards = pcLocalStorage.findAll();
            expect(allCards.length).toEqual(1);
            expect(removedCards.length).toEqual(1);
        });

        it("should update a card in local storage base on cards returned from the server", function () {
            var allCards = pcLocalStorage.findAll();
            allCards[0].version = "1.1";
            _smartUpdate($scope, pcLocalStorage, allCards, 1);

            var card = pcLocalStorage.find({
                id: 1
            });

            expect(card.version).toEqual("1.1");
        });
    });
});