/* global _ */
/* global angular */
/* global cordova */
"use strict";
var lastRefreshed = new Date().toString();

/**
 * Function to return error code string.
 *
 * @function getErrorString
 * @param  {[type]} code
 * @return {[type]}
 */
function getErrorString(code) {
    switch (code) {
    case 1:
        return "NOT_FOUND_ERR";
    case 2:
        return "SECURITY_ERR";
    case 3:
        return "ABORT_ERR";
    case 4:
        return "NOT_READABLE_ERR";
    case 5:
        return "ENCODING_ERR";
    case 6:
        return "NO_MODIFICATION_ALLOWED_ERR";
    case 7:
        return "INVALID_STATE_ERR";
    case 8:
        return "SYNTAX_ERR";
    case 9:
        return "INVALID_MODIFICATION_ERR";
    case 10:
        return "QUOTA_EXCEEDED_ERR";
    case 11:
        return "TYPE_MISMATCH_ERR";
    case 12:
        return "PATH_EXISTS_ERR";
    default:
        return "UNKNOWN";
    }
}

/**
 * Function to return a table of pocket cards.
 *
 * The view expects that data to be an array or arrays where the outer array represents
 * the number of rows and the inner array represents the number of columns. There should
 * be 2 columns for each row.
 *
 * @function _getRows
 * @param {Array} array Array of pocket card objects.
 * @param {Number} numberOfRows The number of rows.
 */
function _getRows(array, numberOfRows) {
    var rows = [],
        cols = [],
        maxRows;

    if (!array) {
        return rows;
    }

    // Display newest cards 1st
    var a = [];
    var b = [];
    _.each(array, function (obj) {
        if (obj.isNew === true || obj.isUpdated === true || obj.isSaved) {
            a.push(obj);
        } else {
            b.push(obj);
        }
    });
    a.sort(function (a, b) {
        return new Date(b.effectiveDate) - new Date(a.effectiveDate);
    });
    b.sort(function (a, b) {
        return new Date(b.effectiveDate) - new Date(a.effectiveDate);
    });
    array = [].concat(a, b);

    if (numberOfRows) {
        maxRows = numberOfRows * 2;
    } else {
        maxRows = array.length;
    }

    // We want 2 columns at most.
    for (var i = 0; i < array.length && i < maxRows; i++) {
        if ((i % 2) === 0) {
            cols = [];
            rows.push(cols);
        }
        cols.push(array[i]);
    }
    return rows;
}

/**
 * Function to check for updates to a pocket card on the device.
 *
 * @function _checkVersionChange description
 * @param  {[type]} allCards
 * @param  {[type]} deviceCards
 * @param  {[type]} $localStorage
 * @return {[type]}
 */
function _checkVersionChange(allCards, deviceCards, $localStorage) {
    if (_.isNull(allCards) || _.isNull(deviceCards) || _.isNull($localStorage)) {
        return;
    }

    var hashtable = {};
    _.each(allCards, function (card) {
        hashtable[card.id] = card;
    });

    if (_.isEmpty(hashtable)) {
        return;
    }

    // NOTE: Using for loop since underscore doesn't allow for the mutation of the collection
    // while iterating over it.
    for (var i = 0; i < deviceCards.length; i++) {
        var updatedCard = hashtable[deviceCards[i].id];
        if (new Date(deviceCards[i].effectiveDate).getTime() !== new Date(updatedCard.effectiveDate).getTime()) {
            console.log("detected version change!");

            // Add a new attribute 'isUpdated' to our updated card.
            updatedCard.isUpdated = true;
            updatedCard.isSaved = deviceCards[i].isSaved;

            // Last, save the updated card to local storage.
            $localStorage.update(updatedCard);
        }
    }
}

/**
 * Function to add a pocket card to the device.
 *
 * @function _addCards description
 * @param {[type]} allCards
 * @param {[type]} collection
 * @param {[type]} $localStorage
 */
function _addCards(allCards, collection, $localStorage) {
    var hashtable = {};
    _.each(allCards, function (card) {
        hashtable[card.id] = card;
    });

    _.each(collection, function (id) {
        var newCard = hashtable[id];
        newCard.isNew = true;
        $localStorage.create(newCard);
    });
}

/**
 * Function to remove a pocket card from the device.
 *
 * @function _removeCards
 * @param  {[type]} deviceAllCards
 * @param  {[type]} collection
 * @param  {[type]} $localStorage
 * @return {[type]}
 */
function _removeCards(deviceAllCards, collection, $localStorage) {
    var hashtable = {};
    var removedCards = [];

    _.each(deviceAllCards, function (card) {
        hashtable[card.id] = card;
    });
    _.each(collection, function (id) {
        var oldCard = hashtable[id];

        removedCards.push(oldCard);
        $localStorage.destroy(oldCard);
    });

    return removedCards;
}

/**
 * Function to update device pocket cards based on the state of server pocket cards.
 * This is done by the following:
 *
 * See if the card version number has changed for any of the cards or if any cards have
 * been added or removed.
 *
 * @function _smartUpdate
 * @param {[type]} $scope
 * @param {[type]} $localStorage
 * @param {[type]} allCards
 * @param {[type]} nbrOfRows
 * @return {[type]} Return the cards removed.
 */
function _smartUpdate($scope, $localStorage, allCards, nbrOfRows) {
    var removedCards = [];

    var deviceAllCards = $localStorage.findAll();
    if (deviceAllCards.length === 0) {
        _.each(allCards, function (card) {
            card.isSaved = false;
            $localStorage.create(card);
        });
    } else {
        var a = _.pluck(allCards, "id");
        var b = _.pluck(deviceAllCards, "id");

        // Determine what's in the device all cards set that's not in the server all cards set.
        var results = _.difference(b, a);
        if (results.length > 0) {
            removedCards = _removeCards(deviceAllCards, results, $localStorage);
            if (removedCards.length > 0) {
                deviceAllCards = $localStorage.findAll();
                b = _.pluck(deviceAllCards, "id");
            }
        }

        // Determine what's in the server all cards set that's not in the device all cards set.
        results = _.difference(a, b);
        if (results.length === 0) {
            // Sets are the same so check for version change.
            _checkVersionChange(allCards, deviceAllCards, $localStorage);
        } else {
            // New cards have been added so save them to local storage.
            _addCards(allCards, results, $localStorage);
        }
    }

    $scope.allCards = _getRows($localStorage.findAll(), nbrOfRows);
    $scope.savedCards = _getRows(_.where($localStorage.findAll(), {
        isSaved: true
    }), nbrOfRows);

    return removedCards;
}

/**
 * Function to retrieve pocket cards from the server.
 *
 * @function _doRefresh
 * @param  {[type]} $scope
 * @param  {[type]} $webService
 * @param  {[type]} nbrOfRows
 * @return {[type]}
 */
function _doRefresh($scope, $localStorage, $webService, $cacheFactory, nbrOfRows) {
    // Remove cached request to get fresh data.
    $cacheFactory.get('$http').removeAll();

    // Retrieve all pocket cards.
    $webService.getPocketCards().then(function (allCards) {
        lastRefreshed = new Date().toString();
        $scope.lastRefreshed = lastRefreshed;

        // Determine if saved cards need to be updated.
        var removedCards = _smartUpdate($scope, $localStorage, allCards, nbrOfRows);
        displayRemovedCards($scope, $webService, removedCards);
    }, function (response) {
        console.log(response);
    }).finally(function () {
        $scope.$broadcast('scroll.refreshComplete');
    });
}

/**
 * Function to save a pocket card to local storage.
 *
 * @function _saveCard
 * @param  {[type]} card
 * @param  {[type]} $scope
 * @param  {[type]} $localStorage
 * @return {[type]}
 */
function _saveCard(card, $scope, $localStorage, nbrOfRows) {
    console.log('saveCard');

    var savedCard = $localStorage.find(card);
    if (savedCard) {
        if (savedCard.isSaved === true) {
            savedCard.isSaved = false;
            savedCard.filename = "";

            $scope.card.isSaved = false;
            $localStorage.update(savedCard);
        } else {
            var filename = "card-" + savedCard.id + ".txt";

            savedCard.isSaved = true;
            savedCard.filename = filename;
            $scope.card.isSaved = true;

            $localStorage.update(savedCard);
        }
    }

    // Refresh view by updating both collections where cards are stored.
    $scope.allCards = _getRows($localStorage.findAll(), nbrOfRows);
    $scope.savedCards = _getRows(_.where($localStorage.findAll(), {
        isSaved: true
    }), nbrOfRows);
}

/**
 * Function to hide the menu popover. This function should be called when navigating
 * to a new view when the menu was displayed on the previous view.
 *
 * @function _hidePopover
 * @param  {[type]} $scope
 * @return {[type]}
 */
function _hidePopover($scope) {
    if ($scope.popover) {
        $scope.popover.hide();
    }
}

/**
 * Function to display cards removed from local storage when the card has been removed on
 * the server.
 *
 * @function displayRemovedCards
 * @param { } $scope
 * @param  {[type]} removedCards
 * @return {[type]}
 */
function displayRemovedCards($scope, $webService, removedCards) {
    var collection = _.filter(removedCards, function (card) {
        if (card.isSaved === true) {
            return card;
        }
    });

    if (collection.length > 0) {
        // Get the reason for removing the card.
        var ids = _.pluck(collection, "id");
        var _scope = $scope;
        $webService.getPocketCard(ids).then(function (cards) {
            _scope.removedCards = cards;
            _scope.modal.show();
        }, function (error) {
            console.log("Error: " + error.status);
        });
    }
}

/**
 * Function to get pocket card from the server.
 *
 * @function _getPocketCards description
 * @param  {[type]} $scope
 * @param  {[type]} $localStorage
 * @param  {[type]} $webService
 * @param  {[type]} nbrOfRows
 * @return {[type]}
 */
function _getPocketCards($scope, $localStorage, $webService, nbrOfRows) {
    var useCache = true;
    $webService.getPocketCards(useCache).then(function (allCards) {
        var removedCards = _smartUpdate($scope, $localStorage, allCards, nbrOfRows);
        displayRemovedCards($scope, removedCards);
    }, function (error) {
        // Error object keys:
        // error.data;
        // error.status;
        // error.headers;
        // error.config;
        // error.statusText;        

        $scope.allCards = _getRows($localStorage.findAll(), nbrOfRows);
        $scope.savedCards = _getRows(_.where($localStorage.findAll(), {
            isSaved: true
        }), nbrOfRows);
    });
}

/**
 * Function to display pocket card detail.
 *
 * @function _showCardDetail
 * @param  {[type]} card
 * @param  {[type]} $state
 * @param  {[type]} $localStorage
 * @return {[type]}
 */
function _showCardDetail(card, $scope, $state, $localStorage) {
    // Remove 'new' badge when the card is view.
    if (card.isNew === true) {
        card.isNew = false;
        $localStorage.update(card);
    }

    // Remove 'updated' badge when the card is view.
    if (card.isUpdated === true) {
        card.isUpdated = false;
        $localStorage.update(card);
    }

    // Using isScreenReaderRunning defined in root scope.
    if (window.isScreenReaderRunning) {
        $state.go("card-front", {
            id: card.id
            // id: card._id
        });
    } else {
        if ($state.current.name === "tab.home") {
            $state.go("tab.home.detail", {
                id: card.id
                // id: card._id
            });
            // $state.go("tab.card-detail", {
            //     id: card.id
            //     // id: card._id
            // });
        } else if ($state.current.name === "tab.all-cards") {
            $state.go("tab.all-cards.detail", {
                id: card.id
                // id: card._id
            });
        } else if ($state.current.name === "tab.search") {
            $state.go("tab.search.detail", {
                id: card.id
                // id: card._id
            });
        }
    }
}

/**
 * Function to search for a card by tags or title.
 *
 * @function _search
 * @param  {[type]} query
 * @param  {[type]} $localStorage
 * @return {[type]}
 */
function _search(query, $localStorage) {
    console.log(query);

    var results = [];
    if (query.length === 0) {
        return results;
    }

    query = query.toLowerCase();
    var allCards = $localStorage.findAll();
    results = _.filter(allCards, function (card) {
        var tags = card.tags.toString().toLowerCase();
        if (tags.search(query) !== -1 || card.title.toLowerCase().search(query) !== -1) {
            return card;
        }
    });
    return results;
}

/**
 * Function to write to a file.
 *
 * @function _writeToFile
 * @param  {[type]} $q
 * @param  {[type]} filename
 * @param  {[type]} content
 * @return {[type]}
 */
function _writeToFile($q, filename, content) {
    var deferred = $q.defer();

    // Define a callback to handle successfull access to the file system.
    var onFileSystemSuccess = function (fileSystem) {
        fileSystem.getFile(filename, {
            create: true,
            exclusive: false
        }, gotFileEntry, fail);
    };

    // Define a callback to handle successfull retreival of a file.
    var gotFileEntry = function (fileEntry) {
        console.log("fileEntry: " + fileEntry);
        console.log("fileEntry.toURL: " + fileEntry.toURL());

        fileEntry.createWriter(gotFileWriter, fail);
    };

    // Define a callback to handle successfull creation of a file writer.
    var gotFileWriter = function (writer) {
        // Define a callback to handle a successful write to the filesystem.
        writer.onwriteend = function (evt) {
            deferred.resolve();
        };

        // Save our generated pdf document to the file system.
        writer.write(content);
    };

    // Define a callback to handle any errors during access to the file system.
    var fail = function (error) {
        console.log(getErrorString(error.code));
        deferred.reject(getErrorString(error.code));
    };

    // Request access to the file system.
    // NOTE: Uncomment to display the location of the file.
    // console.log(cordova.file.dataDirectory);
    try {
        window.resolveLocalFileSystemURL(cordova.file.dataDirectory, onFileSystemSuccess, fail);
    } catch (error) {
        // deferred.reject(error);
        deferred.resolve();
    }
    return deferred.promise;
}

/**
 * Function to read the contents of a file.
 *
 * @function _readFile
 * @param  {[type]} $q
 * @param  {[type]} filename
 * @return {[type]}
 */
function _readFile($q, filename) {
    var deferred = $q.defer();

    // Define a callback to handle successfull access to the file system.
    var onFileSystemSuccess = function (fileSystem) {
        fileSystem.getFile(filename, {
            create: true,
            exclusive: false
        }, gotFileEntry, fail);
    };

    // Define a callback to handle successfull retreival of a file.
    var gotFileEntry = function (fileEntry) {
        console.log("fileEntry: " + fileEntry);
        console.log("fileEntry.toURL: " + fileEntry.toURL());

        fileEntry.file(success, fail);
    };

    var success = function (file) {
        var reader = new FileReader();

        // Define a callback to handle a successful read.
        reader.onloadend = function (evt) {
            console.log("read success");
            console.log(evt.target.result);

            deferred.resolve(evt.target.result);
        };
        reader.readAsText(file);
    };

    // Define a callback to handle any errors during access to the file system.
    var fail = function (error) {
        console.log(error.code);
        console.log(error);
        $q.reject('Error');
    };

    // Request access to the file system.
    // NOTE: Uncomment to display the location of the file.
    // console.log(cordova.file.dataDirectory);
    try {
        window.resolveLocalFileSystemURL(cordova.file.dataDirectory, onFileSystemSuccess, fail);
    } catch (error) {
        // deferred.reject(error);
        deferred.resolve();
    }

    return deferred.promise;
}

// 508
function _setFocus(event, viewConfig) {
    var _viewConfig = viewConfig;

    // TODO: Set focus on the back button if it's displayed
    // nav-view-direction can be none, forward, or back. If none, set focus to title,
    // if forward, set focus to back button.
    // $('ion-nav-view[nav-view-direction|="forward"]').find('.nav-bar-block[nav-bar|="active"]').find('button.back-button')
    if (window.isScreenReaderRunning) {
        console.log("508 Mode for: " + _viewConfig.title);

        // iOS Safari only accepts focus when inside a touch event handler so....
        var $view = $("ion-view[view-title|='" + _viewConfig.title + "']");
        var $title;
        if ($view) {
            $title = $view.find("div.title:contains('" + _viewConfig.title + "')");
        }

        var $ionNavView = $('ion-nav-view[nav-view-direction|="forward"]');
        var $backButton;
        if ($ionNavView) {
            var ionBarBlock = $ionNavView.find('.nav-bar-block[nav-bar|="entering"]');
            if (ionBarBlock) {
                $backButton = ionBarBlock.find('button.back-button');
            }
        }

        if ($backButton) {
            // Must set tabindex 1st!!!
            $backButton.attr("tabindex", "-1");
            $backButton.one('touchstart', function (event) {
                console.log("handle touchstart... " + $(this).html());

                $(this).focus();
            });

            console.log("fire touchstart...");
            $backButton.trigger('touchstart');
            // $backButton.focus();
        } else if ($title) {
            // Must set tabindex 1st!!!
            $title.attr("tabindex", "-1");
            $title.one('touchstart', function (event) {
                console.log("handle touchstart... " + $(this).html());

                $(this).focus();
            });

            console.log("fire touchstart...");
            $title.trigger('touchstart');
        } else {
            console.log("Error! Couldn't set focus to back button or view title.");
        }
    }
}

/**
 * @module controllers
 */
angular.module('controllers', ['services', 'ionic'])
    .controller('MainCtrl', function ($rootScope, $scope, $ionicModal, $state, $localStorage, $webService, $cacheFactory, $interval) {
        console.log('MainCtrl');

        // Set the main page title.
        $scope.title = "Pocket Card";

        $scope.lastRefreshed = lastRefreshed;

        // Hide the popover if it has been displayed.
        _hidePopover($scope);

        if (!localStorage.getItem("pocketcardEula")) {
            $state.go("eula");
        }
        // Retrieve all pocket cards.
        _getPocketCards($scope, $localStorage, $webService);

        // 508
        // NOTE: Focus search search criteria is different for main view.
        console.log("mc setfocus");
        $scope.$on("$ionicView.enter", function (event, viewConfig) {
            var _viewConfig = viewConfig;

            if (window.isScreenReaderRunning) {
                console.log("508 Mode for: " + _viewConfig.title);

                // iOS Safari only accepts focus when inside a touch event handler so....
                var $navbar = $("div.nav-bar-block[nav-bar|='entering']");
                var $title = $navbar.find("div.title:contains('" + _viewConfig.title + "')");

                if ($title) {
                    // Must set tabindex 1st!!!
                    $title.attr("tabindex", "-1");
                    $title.one('dblclick', function (event) {
                        console.log("handle touchstart... " + $(this).html());

                        event.stopPropagation();
                        $(this).focus();
                    });

                    console.log("fire touchstart...");
                    $title.trigger('dblclick');
                }
            }
        });

        $scope.saveCard = function (card) {
            _saveCard(card, $scope, $localStorage, 1);
        };

        $scope.doRefresh = function () {
            _doRefresh($scope, $localStorage, $webService, $cacheFactory, 1);
        };

        $scope.showCardDetail = function (card) {
            _showCardDetail(card, $scope, $state, $localStorage);
        };

        // Load the modal from the given template URL
        $ionicModal.fromTemplateUrl('templates/modal.html', {
            // Use our scope for the scope of the modal to keep it simple
            scope: $scope,
            // The animation we want to use for the modal entrance
            animation: 'slide-in-up'
        }).then(function (modal) {
            $scope.modal = modal;
        });

        $scope.$on('$destroy', function () {
            console.log("destroy");
            $scope.modal.remove();
        });

        $scope.$parent.$on('refresh', function () {
            $scope.doRefresh();
        });

        $scope.hasSavedCards = function () {
            // TODO: Why is this called multiply times???
            return _.where($localStorage.findAll(), {
                isSaved: true
            }).length > 0;
        };

        $scope.hasNoSavedCards = function () {
            return !($scope.hasSavedCards());
        };

        // Refresh rate is entered in hours so convert refresh rate to milliseconds
        var refreshRate = $rootScope.refreshRate * 60 * 60 * 1000;
        $interval(function () {
            console.log("Timer fired! Refreshing cards...");
            _doRefresh($scope, $localStorage, $webService, $cacheFactory, 1);
        }, refreshRate);
    })
    .controller('AboutCtrl', function ($scope) {
        console.log('AboutCtrl');

        // Need to do this after each display of popover.
        _hidePopover($scope);

        // 508
        $scope.$on("$ionicView.enter", _setFocus);
    })
    .controller('MenuCtrl', function ($scope, $ionicPopover) {
        console.log('MenuCtrl');

        $ionicPopover.fromTemplateUrl('templates/popover.html', {
            scope: $scope
        }).then(function (popover) {
            $scope.popover = popover;
        });

        // 508
        // $scope.$on("$ionicView.enter", _setFocus);

        $scope.openPopover = function ($event) {
            console.log('show');
            $scope.popover.show($event);
        };
        $scope.closePopover = function () {
            console.log('hide');
            $scope.popover.hide();
        };
        //Cleanup the popover when we're done with it!
        $scope.$on('$destroy', function () {
            console.log('destroy');
            $scope.popover.remove();
        });
        // Execute action on hide popover
        $scope.$on('popover.hidden', function () {
            // Execute action
        });
        // Execute action on remove popover
        $scope.$on('popover.removed', function () {
            // Execute action
        });
    })
    .controller('AllCardsCtrl', function ($rootScope, $scope, $ionicModal, $state, $localStorage, $webService, $cacheFactory) {
        console.log('AllCardsCtrl');

        // Need to do this after each display of popover.
        _hidePopover($scope);

        $scope.lastRefreshed = lastRefreshed;

        // Retrieve all pocket cards.
        _getPocketCards($scope, $localStorage, $webService);

        // Load the modal from the given template URL
        $ionicModal.fromTemplateUrl('templates/modal.html', function ($ionicModal) {
            $scope.modal = $ionicModal;
        }, {
            // Use our scope for the scope of the modal to keep it simple
            scope: $scope,
            // The animation we want to use for the modal entrance
            animation: 'slide-in-up'
        });

        $scope.$on('$destroy', function () {
            console.log("destroy");
            $scope.modal.remove();
        });

        // 508
        $scope.$on("$ionicView.enter", _setFocus);

        $scope.saveCard = function (card) {
            _saveCard(card, $scope, $localStorage);
        };

        $scope.showCardDetail = function (card) {
            _showCardDetail(card, $scope, $state, $localStorage);
        };

        $scope.doRefresh = function () {
            _doRefresh($scope, $localStorage, $webService, $cacheFactory);
        };

        $scope.hasCards = function () {
            return ($scope.allCards.length > 0 ? true : false);
        };

        $scope.$parent.$on('refresh', function () {
            _getPocketCards($scope, $localStorage, $webService);
        });
    })
    .controller('SavedCardsCtrl', function ($scope, $state, $localStorage) {
        console.log('SavedCardsCtrl');

        var getSavedCards = function () {
            var savedCards = _.where($localStorage.findAll(), {
                isSaved: true
            });

            $scope.savedCards = _getRows(savedCards);
        };

        // Need to do this after each display of popover.
        _hidePopover($scope);

        // Retrieve all saved pocket cards.
        getSavedCards();

        // 508
        $scope.$on("$ionicView.enter", _setFocus);

        $scope.saveCard = function (card) {
            _saveCard(card, $scope, $localStorage);
        };

        $scope.showCardDetail = function (card) {
            _showCardDetail(card, $scope, $state, $localStorage);
        };

        $scope.$parent.$on('refresh', function () {
            getSavedCards();
        });
    })
    .controller('CardDetailCtrl', function ($rootScope, $scope, $state, $stateParams, $localStorage, $webService, $q) {
        console.log('CardDetailCtrl');

        // Need to do this after each display of popover.
        _hidePopover($scope);

        var getPocketCardContent = function (c) {
            var deferred = $q.defer();
            var card = c;
            $webService.getPocketCardContent($stateParams.id, false).then(function (content) {
                card.content = atob(content.buffer);
                console.log(card.content);

                // Save the content to the file system and save reference to file
                // in local storage if the card has been saved.
                if (card.isSaved === true) {
                    var filename = "card-" + $stateParams.id + ".txt";
                    _writeToFile($q, filename, card.content).then(function () {
                        card.filename = filename;

                        // Save the filename to local storage without the content!
                        var c = $localStorage.find(card);
                        c.filename = filename;
                        $localStorage.update(c);

                        deferred.resolve();
                    }, function (error) {
                        // TODO: Maybe display message to user indicating what to do???
                        console.log("Error: " + error.code);
                        deferred.reject(getErrorString(error.code));
                    });
                } else {
                    deferred.resolve();
                }
            }, function (error) {
                console.log("Error: " + Object.keys(error));
                deferred.reject(error);
            });
            return deferred.promise;
        };

        var setContent = function (filename) {
            _readFile($q, filename).then(function (content) {
                $scope.card.content = content;
            }, function (error) {
                console.log(error);
                throw "Error! Card content not found.";
            });
        };

        // Is card in local storage?
        var card = $localStorage.find({
            id: $stateParams.id
        });
        if (!card) {
            throw "Card with id: " + $stateParams.id + " not found!";
        }

        $scope.card = card;

        // Show card content.
        if (card.filename) {
            setContent(card.filename);
        } else {
            getPocketCardContent(card);
        }

        // 508
        console.log("cd.setfocus: " + $scope.$id);
        $scope.$on("$ionicView.enter", _setFocus);

        function _flipCard() {
            if ($rootScope.isScreenReaderRunning) {
                if ($("div#myCard").hasClass("flip") === false) {
                    $("div#myCard").addClass("flip");
                    $state.go("card-back", {
                        id: card.id
                    });
                } else {
                    $("div#myCard").removeClass("flip");
                    $state.go("card-front");
                }
            } else {
                document.querySelector("#myCard").classList.toggle("flip");
            }
        }

        $scope.flip = function (card) {
            // Get content from local storage.
            if (card.isSaved === true) {
                if (card.filename) {
                    _readFile($q, card.filename).then(function (content) {
                        $scope.card.content = content;

                        _flipCard();
                        // if ($rootScope.isScreenReaderRunning) {
                        //     if ($("div#myCard").hasClass("flip") === false) {
                        //         $state.go("card-back");
                        //     } else {
                        //         $state.go("card-detail");
                        //     }
                        // } else {
                        //     document.querySelector("#myCard").classList.toggle("flip");
                        // }
                    }, function (error) {
                        console.log(error);
                    });
                } else {
                    getPocketCardContent(card).then(function () {
                        _flipCard();
                        // if ($rootScope.isScreenReaderRunning) {
                        //     $state.go("card-back");
                        // } else {
                        //     document.querySelector("#myCard").classList.toggle("flip");
                        // }
                    }, function () {
                        _flipCard();
                    });
                }
            } else {
                getPocketCardContent(card).then(function () {
                    _flipCard();
                    // if ($rootScope.isScreenReaderRunning) {
                    //     $state.go("card-back");
                    // } else {
                    //     document.querySelector("#myCard").classList.toggle("flip");
                    // }
                }, function () {
                    _flipCard();
                });
            }
        };

        $scope.saveCard = function (card) {
            console.log("save card");
            _saveCard(card, $scope, $localStorage);
            getPocketCardContent(card);

            $scope.$parent.$emit("refresh");
        };

        $scope.clearSearch = function () {
            this.query = "";
            $('div#card-content').removeHighlight();
        };

        $scope.searchContent = function () {
            console.log("search content");

            $('div#card-content').removeHighlight();
            $('div#card-content').highlight(this.query);
        };
    })
    .controller('SearchCtrl', function ($scope, $state, $localStorage) {
        // Need to do this after each display of popover.
        _hidePopover($scope);

        // 508
        $scope.$on("$ionicView.enter", _setFocus);

        $scope.saveCard = function (card) {
            _saveCard(card, $scope, $localStorage);
        };

        $scope.showCardDetail = function (card) {
            _showCardDetail(card, $scope, $state, $localStorage);
        };

        $scope.clearSearch = function () {
            $scope.query = '';
            while ($scope.searchResults && $scope.searchResults.length > 0) {
                $scope.searchResults.pop();
            }
        };

        $scope.search = function () {
            console.log("search");
            var searchResults = _search($scope.query, $localStorage);
            $scope.searchResults = _getRows(searchResults);
        };

        $scope.$parent.$on('refresh', function () {
            $scope.search();
        });
    })
    .controller('eulaCtrl', function ($scope, $state) {
        console.log('eulaCtrl');

        // Need to do this after each display of popover.
        _hidePopover($scope);

        if (localStorage.getItem("pocketcardEula") === "1") {
            console.log("true");
            $scope.isAccepted = true;
        } else {
            console.log("false");
            $scope.isAccepted = false;
        }

        // 508
        $scope.$on("$ionicView.enter", _setFocus);

        $scope.accept = function () {
            console.log('accept');
            localStorage.setItem("pocketcardEula", 1);
            $state.go("tab.home");
        };

        $scope.isNotAccepted = function () {
            var x = localStorage.getItem("pocketcardEula");
            return (x !== "1");
        };
    })
    .controller('HelpCtrl', function ($scope) {
        console.log('HelpCtrl');

        // Need to do this after each display of popover.
        _hidePopover($scope);

        // 508
        $scope.$on("$ionicView.enter", _setFocus);
    });