/* global process */
//Mongo DAO for PockCard admin web
var log4js = require('log4js');
var utils = require('../../common/utils');
var log = log4js.getLogger('AdminDao_Mongo');
var ObjectId = require('mongodb').ObjectID;

var MongoDao = function (params, dbObject) {
	if (dbObject) {
		this.mongo = dbObject;
	}
	else {
		this.mongo = require('mongodb').MongoClient;
	}
	this.config = null;
	if (params) {
		this.config = params;
		log.debug('dao setup with input parameters');
	}
	else {
		log.debug('setting up from -c or --config arguments');
		var processArgs = require('minimist');
		var args = processArgs(process.argv.slice(1));
		if (args.c) {
			args.config = args.c;
		}
		this.config = require('../' + args.config);
	}

	this.config.connectString = this.config.url + this.config.dbName;
	log.debug('MongoDao instance configured');
};

MongoDao.prototype.getCurrentUserBySysid = function (id, callback) {
	this.mongo.connect(this.config.connectString, {}, function (err, db) {
		if (err) {
			return callback(err);
		}
		var now = new Date();
		var pipeline = [
			{
				$match: {
					effectiveDate: {$lte: now}
				}
			},
			{
				$group: {
					_id: "$id",
					effectiveDate: {$max: "$effectiveDate"},
					data: {$push: {effectiveDate: "$effectiveDate", status: "$status", lastname: "$lastname", firstname: "$firstname", email:"$email", userType:"$userType", activeDirId:"$activeDirId"}}
				}
			},
			{
				$unwind: "$data"
			},
			{
				$project: {
					data: 1,
					match: {
						$and: [
							{$eq: ["$effectiveDate", "$data.effectiveDate"]},
							{$eq: ["$data.activeDirId", id]}
						]
					}
				}
			},
			{
				$match: {match: true}
			}
		];
		db.collection('adminUsers').aggregate(pipeline, {out: {inline:1}}, function (err, results) {
			db.close();
			if (err) {
				return callback(err);
			}
			if (results && results.length) {
				var ary = [];
				for (var i = results.length - 1; i >= 0; i--) {
					var user = {id: results[i]._id};
					ary.push(utils.getMergedInto(user, results[i].data));
				};
				callback(null, ary);
			}
			else {
				callback();
			}
		});
	});
};

MongoDao.prototype.getCurrentUserIdsBySysid = function (id, callback) {
	this.mongo.connect(this.config.connectString, {}, function (err, db) {
		if (err) {
			return callback(err);
		}
		var now = new Date();
		var pipeline = [
			{
				$match: {
					effectiveDate: {$lte: now}
				}
			},
			{
				$group: {
					_id: "$id",
					effectiveDate: {$max: "$effectiveDate"},
					data: {$push: {effectiveDate: "$effectiveDate", activeDirId: "$activeDirId"}}
				}
			},
			{
				$unwind: "$data"
			},
			{
				$project: {
					data: 1,
					match: {
						$and: [
							{$eq: ["$effectiveDate", "$data.effectiveDate"]},
							{$eq: ["$data.activeDirId", id]}
						]
					}
				}
			},
			{
				$match: {match: true}
			}
		];
		db.collection('adminUsers').aggregate(pipeline, {out: {inline:1}}, function (err, results) {
			db.close();
			if (err) {
				return callback(err);
			}
			if (results && results.length) {
				var ary = [];
				for (var i = results.length - 1; i >= 0; i--) {
					ary.push(results[i]._id);
				};
				callback(null, ary);
			}
			else {
				callback();
			}
		});
	});
};

MongoDao.prototype.recordAction = function (data, callback) {
	this.mongo.connect(this.config.connectString, function (err, db) {
		if (err) {
			return callback(err);
		}
		db.collection('audits', function (err, audits) {
			if (err) {
				db.close();
				return callback(err);
			}
			audits.insert(data, function (err, result) {
				db.close();
				if (err) {
					callback(err);
				}
				else {
					//log.debug(result);
					callback();
				}
			});
		});
	});
};

MongoDao.prototype.saveNewUser = function (user, data, callback) {
	var dao = this;
	var newUserId = null;
	this.mongo.connect(this.config.connectString, function (err, db) {
		if (err) {
			return callback(err);
		}
		dao.getNextId(user.id, 'user_id', db, function (err, result) {
			if (err) {
				db.close();
				return callback(err);
			}
			newUserId = result;
			data.id = newUserId;
			db.collection('adminUsers').insert(data, function (err, result) {
				db.close();
				if (err) {
					return callback(err);
				}
				//log.debug('saveNewUser result=');
				//log.debug(result);
				if (1!=result.length) {
					return callback(new Error('mdao: new user not saved'));
				}
				callback(null, newUserId);
			});
		});
	});
};

MongoDao.prototype.updateUser = function (data, callback) {
	this.mongo.connect(this.config.connectString, function (err, db) {
		if (err) {
			return callback(err);
		}
		var adminUsers = db.collection('adminUsers');
		var now = new Date();
		adminUsers.findOne({id:data.id, effectiveDate: {$lte: now}}, {sort:[['effectiveDate', -1]]}, function (err, result) {
			if (err) {
				db.close();
				return callback(err);
			}
			if (!result) {
				db.close();
				return callback(new Error('mdao: user ID not found in system'));
			}
			if (!data.effectiveDate) {
				data.effectiveDate = new Date();
			}
			if (result._id) {
				delete(result._id);
			}
			var newOb = utils.getMerged(result, data);
			adminUsers.insert(newOb, function (err, result) {
				db.close();
				if (err) {
					return callback(err);
				}
				callback();
			});
		});
	});
};

MongoDao.prototype.getCurrentUserById = function (userId, callback) {
	this.mongo.connect(this.config.connectString, function (err, db) {
		if (err) {
			return callback(err);
		}
		db.collection('adminUsers').findOne({effectiveDate: {'$lte': new Date()}, id: userId}, {'sort': [['effectiveDate', -1]]}, function (err, results) {
			db.close();
			//log.debug('getCurrentUser results:');
			//log.debug(results);
			callback(err, results);
		});
	});
}

MongoDao.prototype.saveNewCard = function (user, data, callback) {
	var dao = this;
	this.mongo.connect(this.config.connectString, function (err, db) {
		if (err) {
			return callback(err);
		}
		dao.getNextId(user.id, 'card_id', db, function (err, newId) {
			if (err) {
				db.close();
				return callback(err);
			}
			data.id = newId;
			db.collection('cards').insert(data, function (err, result) {
				db.close();
				if (err) {
					return callback(err);
				}
				if (1!=result.length) {
					return callback(new Error('mdao: new card not saved'));
				}
				callback(null, newId);
			});
		});
	});
};

MongoDao.prototype.updateCard = function (card, callback) {
	//log.debug('updateCard called with:');
	//log.debug(card);
	this.mongo.connect(this.config.connectString, function (err, db) {
		if (err) {
			return callback(err);
		}
		db.collection('cards').insert(card, function (err, result) {
			db.close();
			if (err) {
				return callback(err);
			}
			//log.debug('updateCard result:');
			//log.debug(result);
			if (1!=result.length) {
				return callback(new Error('mdao: updated card not saved'));
			}
			callback();
		});
	});
};

MongoDao.prototype.getNextId = function (userId, idType, db, callback) {
	if (!userId) {
		throw new Error('mdao: userId cannot be null');
	}
	if (!idType) {
		throw new Error('mdao: idType cannot be null');
	}
	if (!' card_id name_id'.indexOf(idType)) {
		throw new Error('mdao: invalid idType ' + idType);
	}
	var userOb = {userId: userId, count: 0, idType: idType};
	tryNextUserId(db, userOb, function (err, result) {
		callback(err, result);
	});
};

MongoDao.prototype.getCurrentCard = function (cardId, callback) {
	this.mongo.connect(this.config.connectString, function (err, db) {
		if (err) {
			return callback(err);
		}
		//db.collection('adminUsers').findOne({'activeDirId': id, 'effectiveDate': {'$lte': now}}, {'sort': [['effectiveDate', -1]]}, function (err, result) {
		var now = new Date();
		db.collection('cards').findOne({id: cardId, effectiveDate: {$lte: now}}, {sort: [['effectiveDate', -1]]}, function (err, result) {
			db.close();
			if (err) return callback(err);
			callback(null, result);
		});
	});
};

MongoDao.prototype.getCurrentCards = function (userId, callback) {
	this.mongo.connect(this.config.connectString, function (err, db) {
		if (err) {
			return callback(err);
		}
		var now = new Date();
		var finalMatch = {
			match: true
		};
		if (userId) {
			finalMatch["data.ownerId"] = userId;
		}
		var pipeline = [
			{
				$match: {
					effectiveDate: {$lte: now}
				}
			},
			{
				$group: {
					_id: "$id",
					effectiveDate: {$max: "$effectiveDate"},
					data: {$push: {effectiveDate: "$effectiveDate", status: "$status", title: "$title", descr: "$descr", version:"$version", versionComment:"$versionComment", ownerId:"$ownerId"}}
				}
			},
			{
				$unwind: "$data"
			},
			{
				$project: {
					data: 1,
					match: {$eq: ["$effectiveDate", "$data.effectiveDate"]}
				}
			},
			{
				$match: finalMatch
			},
			{
				$sort: {"data.title":1}
			}
		];
		db.collection('cards').aggregate(pipeline, {out: {inline:1}}, function (err, result) {
			db.close();
			if (err) {
				log.error(err);
				return callback(err);
			}
			log.debug('getCurrentCards aggregate results:');
			log.debug(result);
			if (result && result.length) {
				var ary = [];
				for (var i = result.length - 1; i >= 0; i--) {
					var item = utils.getMergedInto({id: result[i]._id}, result[i].data);
					ary.push(item);
				};
				callback(null, ary);
			}
			else {
				callback(null, []);
			}
		});
	});
};

MongoDao.prototype.getUsers = function (callback) {
	this.mongo.connect(this.config.connectString, function (err, db) {
		if (err) {
			return callback(err);
		}
		var now = new Date();
		var pipeline = [
			{
				$match: {
					effectiveDate: {$lte: now}
				}
			},
			{
				$group: {
					_id: "$id",
					effectiveDate: {$max: "$effectiveDate"},
					data: {$push: {effectiveDate: "$effectiveDate", status: "$status", lastname: "$lastname", firstname: "$firstname", activeDirId:"$activeDirId", userType:"$userType", email:"$email"}}
				}
			},
			{
				$unwind: "$data"
			},
			{
				$project: {
					data: 1,
					match: {$eq: ["$effectiveDate", "$data.effectiveDate"]}
				}
			},
			{
				$match: {match: true}
			},
			{
				$sort: {"data.lastname":1, "data.firstname":1}
			}
		];
		db.collection('adminUsers').aggregate(pipeline, {out: {inline:1}}, function (err, results) {
			db.close();
			if (err) {
				log.error(err);
				return callback(err);
			}
			log.debug('getUsers aggregate results:');
			log.debug(results);
			if (results && results.length) {
				var ary = [];
				for (var i = results.length - 1; i >= 0; i--) {
					var item = utils.getMergedInto({id: results[i]._id}, results[i].data);
					ary.push(item);
				};
				callback(null, ary);
			}
			else {
				callback(null, []);
			}
		});
	});
};

MongoDao.prototype.getCardFile = function (cardId, callback) {
	this.mongo.connect(this.config.connectString, function (err, db) {
		if (err) {
			return callback(err);
		}
		var now = new Date();
		db.collection('cards').findOne({id:cardId, effectiveDate: {$lte: now}}, {sort:[['effectiveDate', -1]]}, function (err, result) {
			db.close();
			if (err) {
				return callback(err);
			}
			if (!result) {
				return callback();
			}
			callback(null, result.file);
		});
	});
};

MongoDao.prototype.getCardFileInstance = function (cardId, instanceId, callback) {
	if (!instanceId) {
		return callback(new Error('mdao: no instanceId given'));
	}
	instanceId = instanceId.toString();
	if (instanceId.length!=24) {
		return callback(new Error('mdao: instanceId wrong length'));
	}
	this.mongo.connect(this.config.connectString, function (err, db) {
		if (err) {
			return callback(err);
		}
		db.collection('cards').find({id: cardId, _id: new ObjectId(instanceId)}).toArray(function (err, results) {
			db.close();
			if (err) {
				return callback(err);
			}
			if (results && results.length) {
				if (results.length>1) {
					callback(new Error('mdao: too many results'));
				}
				else {
					callback(null, results[0].file);
				}
			}
			else {
				callback();
			}
		});
	});
};

MongoDao.prototype.getCardHistory = function (cardId, callback) {
	this.mongo.connect(this.config.connectString, function (err, db) {
		if (err) {
			return callback(err);
		}
		db.collection('cards').find({id: cardId}, {'file.buffer':0}, {sort:[['effectiveDate', -1]]}).toArray(function (err, results) {
			db.close();
			if (results && results.length) {
				for (var i = results.length - 1; i >= 0; i--) {
					if (results[i]._id) {
						var instId = results[i]._id.toString();
						results[i].instanceId = instId;
						delete(results[i]._id);
					}
				};
			}
			callback(err, results);
		});
	});
};

MongoDao.prototype.getCardAuditHistory = function (cardId, callback) {
	this.mongo.connect(this.config.connectString, function (err, db) {
		if (err) {
			return callback(err);
		}
		db.collection('audits').find({itemCardId: cardId}).toArray(function (err, results) {
			db.close();
			callback(err, results);
		});
	});
};

MongoDao.prototype.setPersonNames = function (personIdNameOb, callback) {
	if (!personIdNameOb) {
		log.debug('setPersonNames null ID object, so nothing returned');
		return callback();
	}
	var ids = null;
	try {
		ids = Object.getOwnPropertyNames(personIdNameOb);
	}
	catch (err) {
		return callback(err);
	}
	if (ids && ids.length) {
		for (var i = ids.length - 1; i >= 0; i--) {
			var id = parseInt(ids[i]);
			if ('NaN'===id.toString()) {
				return callback(new Error('mdao: non-numeric ID:' + ids[i]));
			}
			ids[i] = id;
		};
	}
	else {
		log.debug('setPersonNames nothing in ID object, so nothing returned');
		return callback();
	}

	this.mongo.connect(this.config.connectString, function (err, db) {
		if (err) {
			return callback(err);
		}
		var now = new Date();
		var pipeline = [
			{
				$match: {
					effectiveDate: {$lte: now},
					id: {$in: ids}
				}
			},
			{
				$group: {
					_id: "$id",
					effectiveDate: {$max: "$effectiveDate"},
					data: {$push: {effectiveDate: "$effectiveDate", lastname: "$lastname", firstname: "$firstname"}}
				}
			},
			{
				$unwind: "$data"
			},
			{
				$project: {
					data: 1,
					match: {$eq: ["$effectiveDate", "$data.effectiveDate"]}
				}
			},
			{
				$match: {match: true}
			}
		];
		db.collection('adminUsers').aggregate(pipeline, {out: {inline:1}}, function (err, results) {
			db.close();
			if (err) {
				log.error(err);
				return callback(err);
			}
			log.debug('setPersonNames aggregate results:');
			log.debug(results);
			if (results && results.length) {
				for (var i = results.length - 1; i >= 0; i--) {
					personIdNameOb[results[i]._id] = results[i].data.firstname + ' ' + results[i].data.lastname;
				};
			}
			callback();
		});
	});
};

MongoDao.prototype.getAllCurrentUserNames = function (callback) {
	this.mongo.connect(this.config.connectString, function (err, db) {
		if (err) {
			return callback(err);
		}
		var now = new Date();
		var pipeline = [
			{
				$match: {
					effectiveDate: {$lte: now}
				}
			},
			{
				$group: {
					_id: "$id",
					effectiveDate: {$max: "$effectiveDate"},
					data: {$push: {effectiveDate: "$effectiveDate", lastname: "$lastname", firstname: "$firstname"}}
				}
			},
			{
				$unwind: "$data"
			},
			{
				$project: {
					data: 1,
					match: {$eq: ["$effectiveDate", "$data.effectiveDate"]}
				}
			},
			{
				$match: {match: true}
			}
		];
		db.collection('adminUsers').aggregate(pipeline, {out: {inline:1}}, function (err, results) {
			db.close();
			if (err) {
				log.error(err);
				return callback(err);
			}
			log.debug('getAllCurrentUserNames aggregate results:');
			log.debug(results);
			if (results && results.length) {
				var ary = [];
				for (var i = results.length - 1; i >= 0; i--) {
					var item = {id: results[i]._id, lastname: results[i].data.lastname, firstname: results[i].data.firstname};
					ary.push(item);
				};
				callback(null, ary);
			}
			else {
				callback(null, []);
			}
		});
	});
};

MongoDao.prototype.getMetaMetricsIterator = function (fromDt, toDt, callback) {
	if (fromDt && fromDt.getTime && toDt && toDt.getTime) {
		this.mongo.connect(this.config.connectString, function (err, db) {
			if (err) {
				return callback(err);
			}
			var cursor = db.collection('metrics').find({timestamp: {$gte: fromDt, $lte: toDt}, request: {$regex: /\/cards(\?.*)?$/}}, {_id:0});
			callback(null, new Iterator(db, cursor));
		});
	}
	else {
		callback(new Error('mdao: date(s) invalid'));
	}
};

MongoDao.prototype.getCardMetricsIterator = function (cardId, fromDt, toDt, callback) {
	this.mongo.connect(this.config.connectString, function (err, db) {
		if (err) {
			return callback(err);
		}
		var query = {};
		var reqString = 'cards/';
		if (cardId) {
			reqString += cardId;
		}
		else {
			reqString += '.+';
		}
		reqString += '/contents';
		query.request = {$regex: new RegExp(reqString)};

		if (fromDt || toDt) {
			query.timestamp = {};
			if (fromDt) {
				query.timestamp.$gte = fromDt;
			}
			if (toDt) {
				query.timestamp.$lte = toDt;
			}
		}
		log.debug('query=');
		log.debug(query);

		var cursor = db.collection('metrics').find(query, {_id:0});
		callback(null, new Iterator(db, cursor));
	});
};

function Iterator(db, cursor) {
	this.db = db;
	this.cursor = cursor;
	this.getNext = function (callback) {
		if (this.cursor) {
			this.cursor.nextObject(callback);
		}
		else {
			callback();
		}
	};
	this.done = function () {
		this.db.close();
		this.cursor = null;
		this.db = null;
	};
};

function tryNextUserId (db, userOb, callback) {
	if (userOb.count>2) {
		return callback(new Error('mdao: failed to get a new ID number'));
	}
	//log.debug('tryNextUserId #' + userOb.count);
	userOb.count++;
	var aDate = new Date();
	var oldDate = new Date(aDate);
	oldDate.setSeconds(oldDate.getSeconds()-5);
	db.collection('ids').update({name: userOb.idType, $or: [{locked_by: null}, {locked_at: null}, {locked_at: {$lte: oldDate}}]}, {$set: {locked_by: userOb.userId, locked_at: aDate}}, function (err, result1) {
		if (err) return callback(err);
		//log.debug('1 of 3 results:');
		//log.debug(result1);
		if (result1) {
			//log.debug('1 of 3 steps done');
			db.collection('ids').find({name: userOb.idType, locked_by: userOb.userId}).toArray(function (err, result2) {
				if (err) return callback(err);
				//log.debug('2 of 3 results');
				//log.debug(result2);
				var newValue = result2[0].value+1;
				db.collection('ids').update({name: userOb.idType, locked_by: userOb.userId}, {name: userOb.idType, value: newValue}, function (err, result3) {
					if (err) return callback(err);
					//log.debug('3 of 3 results');
					//log.debug(result3);
					callback(null, newValue);
				});
			});
		}
		else {
			setTimeout(function () {
				tryNextUserId(db, userOb, callback);
			}, 500);
		}
	});
};

module.exports = MongoDao;
