//VisualStateModel.js
var Daoi = require('../dao/IAdminDao.js');
var log4js = require('log4js');
var log = log4js.getLogger('VisualStateModel');
var global = require('./global');
var clientserverCommon = require('../public/js/clientserverCommon');

var VisualStateModel = function(aDao) {
	if (aDao) {
		this.dao =  new Daoi(aDao);
	}
	else {
		throw new Error('no database dao present');
	}
};

VisualStateModel.prototype.authorize = function (stateObject, userObject, callback) {
	assertStateObject(stateObject);
	if (userObject && userObject.id) {
		userObject.id = userObject.id.toLowerCase();
	}
	if (stateObject.userId && 
		stateObject.userType && 
		stateObject.activeDirId && 
		stateObject.activeDirId==userObject.id && 
		(new Date().getTime() - stateObject.lastCheckTimestamp) < (900000)) { //15 minutes

		callback();
	}
	else {
		this.clearState(stateObject);
		this.resetState(stateObject);
		this.retrieveAndSetupState(stateObject, userObject, callback);
	}
};

VisualStateModel.prototype.continueOkay = function (state, callback) {
	if (this.assertAuthorization(state, 'card')) {
		callback('ok');
	}
	else {
		callback('error');
	}
};

VisualStateModel.prototype.logout = function (stateObject, resultFunc) {
	if (!stateObject) {
		stateObject = {};
	}
	this.clearState(stateObject);
	resultFunc(stateObject.visualState);
};

VisualStateModel.prototype.selectCard = function(stateObject, cardId, resultFunc) {
	if (!this.assertAuthorization(stateObject, 'card')) {
		this.pushError(stateObject, 21);
		return this.setNothingSelectedState(stateObject, resultFunc);
	}

	if (cardId) {
		var id = parseInt(cardId);
		if ('NaN'==id.toString()) {
			this.pushError(stateObject, 22);
			this.setNothingSelectedState(stateObject, resultFunc);
		}
		else {
			stateObject.selectedCard = id;
			stateObject.selectedUser = null;
			stateObject.visualState = "selectedCard";
			resultFunc(this.getDisplayModel(stateObject));
		}
	}
	else {
		this.pushError(stateObject, 23);
		this.setNothingSelectedState(stateObject, resultFunc);
	}
};

VisualStateModel.prototype.saveNewCard = function(stateObject, newCard, resultFunc) {
	if (!this.assertAuthorization(stateObject, 'card')) {
		this.pushError(stateObject, 31);
		return this.setNothingSelectedState(stateObject, resultFunc);
	}
	if (newCard) {
		var model = this;
		newCard.ownerId = stateObject.userId;
		normalizeCard(newCard);
		if ('active'===newCard.status && !newCard.versionComment) {
			model.pushError(stateObject, 32);
			return model.setNothingSelectedState(stateObject, resultFunc);
		}
		model.dao.saveNewCard({id:stateObject.userId}, newCard, function (err, newCardId) {
			if (err) {
				log.error(err);
				model.pushError(stateObject, 33);
				return model.setNothingSelectedState(stateObject, resultFunc);
			}
			if (!newCardId) {
				var error = new Error('no cardId returned');
				log.error(error);
				model.pushError(stateObject, 34);
				return model.setNothingSelectedState(stateObject, resultFunc);
			}
			model.dao.recordAction({userId: stateObject.userId, itemCardId: newCardId, effectiveDate: newCard.effectiveDate, action: 'saveCard', datetime: new Date()}, function (err) {
				if (err) {
					log.error(err);
					model.pushError(stateObject, 35);
					model.setNothingSelectedState(stateObject, resultFunc);
				}
				else {
					model.selectCard(stateObject, newCardId, resultFunc);
				}
			});
		})
	}
	else {
		this.pushError(stateObject, 36);
		this.setNothingSelectedState(stateObject, resultFunc);
	}
};

VisualStateModel.prototype.isSystemIdUsed = function (state, systemId, callback) {
	if (!this.assertAdminLoggedIn(state)) {
		return callback(new Error('user auth'));
	}
	if (!systemId) {
		return callback(new Error('systemId missing'));
	}
	this.dao.getCurrentUserIdsBySysid(systemId, function (err, ids) {
		if (err) {
			return callback(err);
		}
		var result = false;
		if (ids && ids.length) {
			if (state.selectedUser) {
				if (1==ids.length) {
					if (ids[0]!=state.selectedUser) {
						result = true;
					}
				}
				else {
					result = true;
				}
			}
			else {
				result = true;
			}
		}
		callback(null, result);
	});
};

VisualStateModel.prototype.selectUser = function(stateObject, userId, resultFunc) {
	if (!this.assertAdminLoggedIn(stateObject)) {
		this.pushError(stateObject, 41);
		return this.setNothingSelectedState(stateObject, resultFunc);
	}
	if (userId) {
		var id = parseInt(userId);
		if ('NaN'==id.toString()) {
			this.pushError(stateObject, 42);
			this.setNothingSelectedState(stateObject, resultFunc);
		}
		else {
			stateObject.selectedUser = id;
			stateObject.selectedCard = null;
			stateObject.visualState = "selectedUser";
			resultFunc(this.getDisplayModel(stateObject));
		}
	}
	else {
		this.pushError(stateObject, 43);
		this.setNothingSelectedState(stateObject, resultFunc);
	}
};

VisualStateModel.prototype.saveNewUser = function(stateObject, newUser, resultFunc) {
	if (!this.assertAdminLoggedIn(stateObject)) {
		this.pushError(stateObject, 51);
		return this.setNothingSelectedState(stateObject, resultFunc);
	}
	if (!newUser) {
		this.pushError(stateObject, 52);
		return this.setNothingSelectedState(stateObject, resultFunc);
	}
	if (!newUser.status) {
		newUser.status = 'active';
	}
	if (!newUser.effectiveDate) {
		newUser.effectiveDate = new Date();
	}
	if (!newUser.activeDirId || !clientserverCommon.sysidRegex.test(newUser.activeDirId)) {
		this.pushError(stateObject, 53);
		return this.setNothingSelectedState(stateObject, resultFunc);
	}
	if (!newUser.lastname || !clientserverCommon.nameRegex.test(newUser.lastname)) {
		this.pushError(stateObject, 58);
		return this.setNothingSelectedState(stateObject, resultFunc);
	}
	if (!newUser.firstname || !clientserverCommon.nameRegex.test(newUser.firstname)) {
		this.pushError(stateObject, 59);
		return this.setNothingSelectedState(stateObject, resultFunc);
	}
	newUser.activeDirId = newUser.activeDirId.toLowerCase();

	var model = this;
	var datetime = new Date();
	this.dao.getCurrentUserBySysid(newUser.activeDirId, function (err, users) {
		if (err) {
			log.error(err);
			model.pushError(stateObject, 71);
			return model.setNothingSelectedState(stateObject, resultFunc);
		}
		if (users && users.length) {
			model.pushError(stateObject, 72);
			return model.setNothingSelectedState(stateObject, resultFunc);
		}
		model.dao.recordAction({userId: stateObject.userId, action: 'insert new user ' + JSON.stringify(newUser), effectiveDate: newUser.effectiveDate, datetime: datetime}, function (err) {
			if (err) {
				log.error(err);
				model.pushError(stateObject, 54);
				return model.setNothingSelectedState(stateObject, resultFunc);
			}
			model.dao.saveNewUser({id: stateObject.userId}, newUser, function (err, newUserId) {
				if (err) {
					log.error(err);
					model.pushError(stateObject, 55);
					return model.setNothingSelectedState(stateObject, resultFunc);
				}

				if (!newUserId) {
					var error = new Error('no userId returned');
					log.error(error);
					model.pushError(stateObject, 56);
					return model.setNothingSelectedState(stateObject, resultFunc);
				}

				model.dao.recordAction({userId: stateObject.userId, action: 'added user', effectiveDate: newUser.effectiveDate, datetime: datetime, itemUserId: newUserId}, function (err) {
					if (err) {
						log.error(err);
						model.pushError(stateObject, 57);
						return model.setNothingSelectedState(stateObject, resultFunc);
					}
					model.selectUser(stateObject, newUserId, resultFunc);
				});
			});
		});
	});
};

VisualStateModel.prototype.getSelectedUser = function (stateObject, callback) {
	if (this.assertAdminLoggedIn(stateObject)) {
		if (stateObject.selectedUser) {
			this.dao.getCurrentUserById(stateObject.selectedUser, function (err, result) {
				if (err) {
					log.error(err);
					return callback(err);
				}
				if (!result) {
					return callback(new Error('user not found for ID'));
				}
				callback(null, result);
			});
		}
		else {
			callback(new Error('user not selected'));
		}
	}
	else {
		callback(new Error('not authorized'));
	}
};

VisualStateModel.prototype.getSelectedUserModel = function (state, callback) {
	if (this.assertAdminLoggedIn(state)) {
		if (state.selectedUser) {
			var _this = this;
			this.getSelectedUser(state, function (err, user) {
				if (err) {
					_this.pushError(state, 83);
				}
				else if (!user) {
					_this.pushError(state, 84);
				}
				state.visualState = "userSelected";
				var dModel = _this.getDisplayModel(state);
				dModel.user = user;
				callback(dModel);
			});
		}
		else {
			this.pushError(state, 82);
			this.setNothingSelectedState(state, callback);
		}
	}
	else {
		this.pushError(state, 81);
		this.setNothingSelectedState(state, callback);
	}
};

VisualStateModel.prototype.getSelectedCardModel = function (state, callback) {
	if (this.assertAuthorization(state, 'card')) {
		if (state.selectedCard) {
			var _this = this;
			this.getSelectedCard(state, function (err, card) {
				if (err) {
					log.error(err);
					_this.pushError(state, 93);
				}
				if (!card) {
					_this.pushError(state, 94);
					return _this.setNothingSelectedState(state, callback);
				}
				state.visualState = "selectedCard";
				var dModel = _this.getDisplayModel(state);
				dModel.card = card;
				callback(dModel);
			});
		}
		else {
			this.pushError(state, 92);
			this.setNothingSelectedState(state, callback);
		}
	}
	else {
		this.pushError(state, 91);
		this.setNothingSelectedState(state, callback);
	}
};

VisualStateModel.prototype.getSelectedCard = function (stateObject, callback) {
	if (this.assertAuthorization(stateObject, 'card') && stateObject.selectedCard) {
		var model = this;
		model.dao.getCurrentCard(stateObject.selectedCard, function (err, card) {
			if (err) {
				log.error(err);
				return callback(err);
			}
			if (!card) {
				return callback(new Error('card not found'));
			}
			if (model.assertAdminLoggedIn(stateObject) || (card.ownerId && stateObject.userId==card.ownerId)) {
				card.canEdit = true;
			}

			if (card.ownerId) {
				model.dao.getCurrentUserById(card.ownerId, function (err, user) {
					if (err) {
						log.error(err);
						card.ownerName = '<error getting name>';
						return callback(err, card);
					}
					if (!user) {
						card.ownerName = '<user name not found>';
						err = new Error('user not found');
					}
					else {
						card.ownerName = user.firstname + ' ' + user.lastname;
					}
					callback(err, card);
				});
			}
			else {
				card.ownerName = '<no user assigned>';
				callback(new Error('no owner ID'), card);
			}
		});
	}
	else {
		callback(new Error('user state error'));
	}
};

VisualStateModel.prototype.updateCard = function (stateObject, newCard, callback) {
	if (this.assertAuthorization(stateObject, 'card')) {
		var model = this;
		model.getSelectedCard(stateObject, function (err, card) {
			if (err) {
				model.pushError(stateObject, 111);
				return model.setNothingSelectedState(stateObject, callback);
			}
			if (!card || !card.ownerId) {
				model.pushError(stateObject, 112);
				return model.setNothingSelectedState(stateObject, callback);
			}
			if (model.assertCardAdminLoggedIn(stateObject) && stateObject.userId!=card.ownerId) {
				model.pushError(stateObject, 113);
				return model.selectCard(stateObject, stateObject.selectedCard, callback);
			}
			newCard.id = card.id;
			if (newCard.replaceOwner) {
				delete newCard.replaceOwner;
				var newOwner = parseInt(newCard.ownerId);
				if ('NaN'===newOwner.toString()) {
					model.pushError(stateObject, 114);
					return model.selectCard(stateObject, stateObject.selectedCard, callback);
				}
				if (newOwner<1) {
					model.pushError(stateObject,115);
					return model.selectCard(stateObject, stateObject.selectedCard, callback);
				}
				newCard.ownerId = newOwner;
			}
			else {
				newCard.ownerId = card.ownerId;
			}
			reconcileFileUpdate(newCard, card);
			normalizeCard(newCard);
			if ('active'===newCard.status && !newCard.versionComment) {
				model.pushError(stateObject, 116);
				return model.selectCard(stateObject, stateObject.selectedCard, callback);
			}
			if ('active'===card.status && !newCard.versionComment) {
				model.pushError(stateObject, 117);
				return model.selectCard(stateObject, stateObject.selectedCard, callback);
			}
			model.dao.recordAction({userId: stateObject.userId, itemCardId: newCard.id, action: 'updateCard', effectiveDate: newCard.effectiveDate, datetime: new Date()}, function (err) {
				if (err) {
					log.error(err);
					model.pushError(stateObject, 118);
					return model.setNothingSelectedState(stateObject, callback);
				}
				model.dao.updateCard(newCard, function (err) {
					if (err) {
						log.error(err);
						model.pushError(stateObject, 119);
						model.setNothingSelectedState(stateObject, callback);
					}
					else {
						model.selectCard(stateObject, newCard.id, callback);
					}
				});
			});
		});
	}
	else {
		this.pushError(stateObject, 120);
		this.setNothingSelectedState(stateObject, callback);
	}
};

VisualStateModel.prototype.updateUser = function (state, newUser, callback) {
	if (this.assertAdminLoggedIn(state)) {
		if (state.selectedUser && newUser) {
			var model = this;
			if (!newUser.firstname || !clientserverCommon.nameRegex.test(newUser.firstname)) {
				model.pushError(state, 107);
				return model.selectUser(state, state.selectedUser, callback);
			}
			if (!newUser.lastname || !clientserverCommon.nameRegex.test(newUser.lastname)) {
				model.pushError(state, 108);
				return model.selectUser(state, state.selectedUser, callback);
			}
			if (!newUser.activeDirId || !clientserverCommon.sysidRegex.test(newUser.activeDirId)) {
				model.pushError(state, 109);
				return model.selectUser(state, state.selectedUser, callback);
			}
			model.getSelectedUser(state, function (err, user) {
				if (err) {
					log.error(err);
					model.pushError(state, 101);
					return model.selectUser(state, state.selectedUser, callback);
				}
				if (!user) {
					model.pushError(state, 102);
					return model.setNothingSelectedState(state, callback);
				}
				newUser.id = user.id;
				if (!newUser.effectiveDate) {
					newUser.effectiveDate = new Date();
				}
				model.dao.recordAction({userId: state.userId, action: 'updateUser', effectiveDate: newUser.effectiveDate, datetime: new Date(), itemUserId: newUser.id}, function (err) {
					if (err) {
						log.error(err);
						model.pushError(state, 103);
						return model.selectUser(state, state.selectedUser, callback);
					}
					model.dao.getCurrentUserBySysid(newUser.activeDirId, function (err, users) {
						if (err) {
							log.error(err);
							model.pushError(state, 131);
							return model.selectUser(state, state.selectedUser, callback);
						}
						if (users && users.length) {
							if (1<users.length) {
								model.pushError(state, 132);
								return model.selectUser(state, state.selectedUser, callback);
							}
							if (users[0].id!=state.selectedUser) {
								model.pushError(state, 133);
								return model.selectUser(state, state.selectedUser, callback);
							}
						}
						model.dao.updateUser(newUser, function (err) {
							if (err) {
								log.error(err);
								model.pushError(state, 104);
							}
							model.selectUser(state, state.selectedUser, callback);
						});
					});
				});
			});
		}
		else {
			this.pushError(state, 105);
			this.setNothingSelectedState(state, callback);
		}
	}
	else {
		this.pushError(state, 106);
		this.setNothingSelectedState(state, callback);
	}
};

VisualStateModel.prototype.popError = function(stateObject) {
	assertStateObject(stateObject);
	var result = null;
	if (stateObject && stateObject.errors && stateObject.errors.length) {
		result = stateObject.errors.pop();
	}
	return result;
};

VisualStateModel.prototype.popAllErrors = function (stateObject) {
	assertStateObject(stateObject);
	var result = null;
	if (stateObject && stateObject.errors && stateObject.errors.length) {
		result = stateObject.errors;
		stateObject.errors = [];
	}
	return result;
};

VisualStateModel.prototype.getCurrentCards = function (stateObject, callback) {
	var model = this;
	if (this.assertAuthorization(stateObject, 'card')) {
		var userId = null;
		if (this.assertCardAdminLoggedIn(stateObject)) {
			userId = stateObject.userId;
		}
		model.dao.getCurrentCards(userId, function (err, results) {
			if (err) {
				log.error(err);
				model.pushError(stateObject, 61);
				return callback(err, null);
			}
			if (null==results) {
				results = [];
			}
			callback(null, results);
		});
	}
	else {
		model.pushError(stateObject, 62);
		callback(new Error('not authorized'));
	}
};

VisualStateModel.prototype.getAllCurrentCards = function (stateObject, callback) {
	var model = this;
	if (this.assertAuthorization(stateObject, 'card')) {
		model.dao.getCurrentCards(null, function (err, results) {
			if (err) {
				log.error(err);
				model.pushError(stateObject, 11);
				return callback(err, null);
			}
			if (null==results) {
				results = [];
			}
			callback(null, results);
		});
	}
	else {
		model.pushError(stateObject, 12);
		callback(new Error('not authorized'));
	}
};

VisualStateModel.prototype.retrieveAndSetupState = function (state, user, callback) {
	var model = this;
	var debugId = (Math.random()+'').split('.')[1];
	log.debug({retrieveAndSetupState:debugId, user: user});
	if (!(user && user.id)) {
		this.pushError(state, 121);
		return callback(new Error('no user id'));
	}
	model.dao.getCurrentUserBySysid(user.id, function (err, results) {
		log.debug({retrieveAndSetupState: debugId, result: results});
		if (err) {
			model.pushError(state, 122);
			return callback(err);
		}
		if (!results || results.length<1) {
			model.pushError(state, 123);
			return callback();
		}
		if (results.length>1) {
			model.pushError(state, 124);
			return callback();
		}
		var result = results[0];
		if ('active'!=result.status) {
			model.pushError(state, 125);
			return callback();
		}
		state.userType = result.userType;
		state.userId = result.id;
		state.userName = result.firstname + ' ' + result.lastname;
		state.activeDirId = result.activeDirId;
		state.lastCheckTimestamp = new Date().getTime();
		model.dao.recordAction({userId: state.userId, action: 'authorize', datetime: new Date()}, function (err) {
			if (err) {
				model.pushError(state, 126);
				return callback(err);
			}
			callback();
		});
	});
};

VisualStateModel.prototype.getUsers = function (state, callback) {
	if (!this.assertAdminLoggedIn(state)) {
		return callback(new Error('authorization error'));
	}
	state.selectedUser = null;
	state.selectedCard = null;
	this.dao.getUsers(function (err, results) {
		if (err) {
			callback(err);
		}
		else {
			if (results && results.length) {
				callback(null, results);
			}
			else {
				callback(null, []);
			}
		}
	});
};

VisualStateModel.prototype.getCardFile = function (state, instanceId, callback) {
	if (!this.assertAuthorization(state, 'card')) {
		return callback(new Error('not authorized'));
	}
	if (!state.selectedCard) {
		return callback(new Error('no card selected'));
	}
	var cardId = parseInt(state.selectedCard);

	if (instanceId) {
		this.dao.getCardFileInstance(cardId, instanceId, function (err, result) {
			if (err) {
				log.error(err);
				return callback(err);
			}
			if (!result) {
				return callback(new Error('card file not found'));
			}
			callback(null, result);
		});
	}
	else {
		this.dao.getCardFile(cardId, function (err, result) {
			if (err) {
				log.error(err);
				return callback(err);
			}
			if (!result) {
				return callback(new Error('card file not found'));
			}
			callback(null, result);
		});
	}
};

VisualStateModel.prototype.getCardActivity = function (state, callback) {
	if (!this.assertAuthorization(state, 'card')) {
		return callback(131);
	}
	if (!state.selectedCard) {
		return callback(132);
	}
	var cardId = parseInt(state.selectedCard);

	var model = this;
	this.dao.getCardHistory(cardId, function (err, histlist) {
		if (err) {
			return callback(err);
		}
		histlist.sort(function (i1, i2) {
			return i2.effectiveDate.getTime()-i1.effectiveDate.getTime();
		});
		histlist.forEach(function (item, index, ary) {
			if ((index-1)>-1) {
				item.nextEffectiveDate = ary[index-1].effectiveDate;
			}
			else {
				item.nextEffectiveDate = new Date();
			}
		});
		if (histlist && histlist.length) {
			var personIdNameSet = {};
			setPersonIdsFromAry(histlist, 'ownerId', personIdNameSet);
			model.dao.getCardAuditHistory(cardId, function (err, auditlist) {
				if (err) {
					return callback(err);
				}
				setEditorIdOnCards(state.selectedCard, histlist, auditlist);
				setPersonIdsFromAry(auditlist, 'userId', personIdNameSet);
				model.dao.setPersonNames(personIdNameSet, function (err) {
					if (err) {
						return callback(err);
					}
					setNamesOnCards(histlist, personIdNameSet);
					callback(null, histlist);
				});
			});
		}
		else {
			callback();
		}
	});
};

VisualStateModel.prototype.getCardVersions = function (state, callback) {
	if (!this.assertAuthorization(state, 'card')) {
		return callback(141);
	}
	if (!state.selectedCard) {
		return callback(142);
	}
	var cardId = parseInt(state.selectedCard);

	var model = this;
	model.dao.getCardHistory(cardId, function (err, histlist) {
		if (err) {
			return callback(err);
		}
		var versions = [];
		if (histlist && histlist.length) {
			for (var i = histlist.length - 1; i >= 0; i--) {
				if (histlist[i].version) {
					versions.push(histlist[i].version);
				}
			};
		}
		callback(null, versions);
	});
};

VisualStateModel.prototype.getAllCurrentUserNames = function (state, callback) {
	if (!this.assertAdminLoggedIn(state)) {
		return callback(151);
	}

	var model = this;
	model.dao.getAllCurrentUserNames(function (err, results) {
		callback(err, results);
	});
};

VisualStateModel.prototype.getMetaMetricsIterator = function (state, fromDate, toDate, callback) {
	if (!this.assertAuthorization(state, 'card')) {
		return callback(161);
	}
	if (!fromDate || !toDate || !fromDate.getTime || !toDate.getTime) {
		return callback(162);
	}
	this.dao.getMetaMetricsIterator(fromDate, toDate, function (err, iter) {
		callback(err, iter);
	});
};

VisualStateModel.prototype.getCardMetricsIterator = function (state, fromDate, toDate, callback) {
	if (!this.assertAuthorization(state, 'card')) {
		return callback(171);
	}
	this.dao.getCardMetricsIterator(state.selectedCard, fromDate, toDate, function (err, iter) {
		callback(err, iter);
	});
};

VisualStateModel.prototype.setManageUsersState = function (state, callback) {
	if (!this.assertAdminLoggedIn(state)) {
		return this.setNothingSelectedState(state, callback);
	}
	state.selectedCard = null;
	state.selectedUser = null;
	state.visualState = "manageUsers";
	callback(this.getDisplayModel(state));
};

VisualStateModel.prototype.setManageCardsState = function (state, callback) {
	if (!this.assertAuthorization(state, 'card')) {
		return this.setNothingSelectedState(state, callback);
	}
	state.selectedCard = null;
	state.selectedUser = null;
	state.visualState = 'manageCards';
	callback(this.getDisplayModel(state));
};

//TODO: finish moving display models to main model
VisualStateModel.prototype.setAddUserState = function (state, callback) {
	if (!this.assertAdminLoggedIn(state)) {
		return this.setNothingSelectedState(state, callback);
	}
	state.selectedCard = null;
	state.selectedUser = null;
	state.visualState = "addUser";
	callback(this.getDisplayModel(state));
};

VisualStateModel.prototype.setAddCardState = function (state, callback) {
	if (!this.assertCardAdminLoggedIn(state) && !this.assertAdminLoggedIn(state)) {
		return this.setNothingSelectedState(state, callback);
	}
	state.selectedCard = null;
	state.selectedUser = null;
	state.visualState = "addCard";
	callback(this.getDisplayModel(state));
};

VisualStateModel.prototype.setEditCardState = function (state, callback) {
	if ((!this.assertCardAdminLoggedIn(state) && !this.assertAdminLoggedIn(state)) || !state.selectedCard) {
		return this.setNothingSelectedState(state, callback);
	}
	state.selectedUser = null;
	state.visualState = 'editCard';
	callback(this.getDisplayModel(state));
};

VisualStateModel.prototype.setEditUserState = function (state, callback) {
	if (!this.assertAdminLoggedIn(state) || !state.selectedUser) {
		return this.setNothingSelectedState(state, callback);
	}
	state.selectedCard = null;
	state.visualState = 'editUser';
	callback(this.getDisplayModel(state));
};

VisualStateModel.prototype.setViewCardHistoryState = function (state, callback) {
	if (!this.assertAuthorization(state, 'card') || !state.selectedCard) {
		return this.setNothingSelectedState(state, callback);
	}
	state.selectedUser = null;
	state.visualState = 'cardHistory';
	callback(this.getDisplayModel(state));
};

VisualStateModel.prototype.setHelpState = function (state, callback) {
	if (this.assertLoggedIn(state)) {
		state.visualState = 'help';
		callback(this.getDisplayModel(state));
	}
	else {
		this.setNothingSelectedState(state, callback);
	}
};

VisualStateModel.prototype.setAboutState = function (state, callback) {
	if (this.assertLoggedIn(state)) {
		state.visualState = 'about';
		callback(this.getDisplayModel(state));
	}
	else {
		this.setNothingSelectedState(state, callback);
	}
};

VisualStateModel.prototype.setCardsMetricsDownloadState = function (state, callback) {
	if (this.assertAuthorization(state, 'card')) {
		state.selectedCard = null;
		state.selectedUser = null;
		state.visualState = 'cardsmetrics';
		callback(this.getDisplayModel(state));
	}
	else {
		this.pushError(state, 191);
		this.setNothingSelectedState(state, callback);
	}
};

VisualStateModel.prototype.setCardMetricsDownloadState = function (state, callback) {
	if (this.assertAuthorization(state, 'card')) {
		state.selectedUser = null;
		state.visualState = 'cardmetrics';
		if (state.selectedCard) {
			var _this = this;
			this.getSelectedCard(state, function (err, card) {
				if (!card) {
					_this.pushError(state, 181);
					return _this.setNothingSelectedState(state, callback);
				}
				if (err) {
					_this.pushError(state, 182);
				}
				var dModel = _this.getDisplayModel(state);
				dModel.card = card;
				callback(dModel);
			});
		}
		else {
			callback(this.getDisplayModel(state));
		}
	}
	else {
		this.pushError(state, 183);
		this.setNothingSelectedState(state, callback);
	}
}

VisualStateModel.prototype.assertAuthorization = function (stateObject, type) {
	assertStateObject(stateObject);
	if (type) {
		if ('user'==type) {
			return this.assertAdminLoggedIn(stateObject);
		}
		else if ('card'==type) {
			return this.assertAdminLoggedIn(stateObject) || this.assertCardAdminLoggedIn(stateObject);
		}
		else {
			return false;
		}
	}
	else {
		return false;
	}
};

VisualStateModel.prototype.assertLoggedIn = function(stateObject) {
	assertStateObject(stateObject);
	var result = (null!=stateObject && null!=stateObject.userId && null!=stateObject.userType);
	return result;
};

VisualStateModel.prototype.assertAdminLoggedIn = function(stateObject) {
	assertStateObject(stateObject);
	var result = (this.assertLoggedIn(stateObject) && 'user'===stateObject.userType);
	return result;
};

VisualStateModel.prototype.assertCardAdminLoggedIn = function (stateObject) {
	assertStateObject(stateObject);
	var result = (this.assertLoggedIn(stateObject) && 'card'===stateObject.userType);
	return result;
};

/**********************************************************************************/
/**********************************************************************************/
/**********************************************************************************/

/* Clears everything except the error array */
VisualStateModel.prototype.clearState = function (stateObject) {
	assertStateObject(stateObject);
	stateObject.visualState = "";
	stateObject.userId = null;
	stateObject.userType = null;
	stateObject.userName = null;
	this.clearSelections(stateObject);
};

/* Resets the model to all defaults */
VisualStateModel.prototype.resetState = function (stateObject) {
	assertStateObject(stateObject);
	stateObject.visualState = "";
	stateObject.errors = [];
};

VisualStateModel.prototype.pushError = function(stateObject, message) {
	assertStateObject(stateObject);
	if (!message) return;
	if (!stateObject.errors) {
		stateObject.errors = [];
	}
	stateObject.errors.push(message);
};

VisualStateModel.prototype.clearSelections = function(stateObject) {
	assertStateObject(stateObject);
	stateObject.selectedUser = null;
	stateObject.selectedCard = null;
};

VisualStateModel.prototype.setNothingSelectedState = function (stateObject, resultFunc) {
	assertStateObject(stateObject);
	if (this.assertLoggedIn(stateObject)) {
		this.clearSelections(stateObject);
		if (this.assertAdminLoggedIn(stateObject)) {
			stateObject.visualState = 'homeUser';
			return resultFunc(this.getDisplayModel(stateObject));
		}
		if (this.assertCardAdminLoggedIn(stateObject)) {
			stateObject.visualState = 'homeCard';
			return resultFunc(this.getDisplayModel(stateObject));
		}
	}
	this.logout(stateObject, resultFunc);
};

VisualStateModel.prototype.getDisplayModel = function (state, includeAllErrors) {
	var st = state || {};
	var model = {userName: st.userName, userId: st.userId, userType: st.userType};
	if (includeAllErrors) {
		model.errors = this.popAllErrors(state);
	}
	return model;
}


function assertStateObject (state) {
	if (null==state) {
		throw new Error('null state object');
	}
}

function normalizeCard (card) {
	if (!card) {
		return;
	}

	if (!card.status || 
		('active'!=card.status && 'inactive'!=card.status) ||
		(!(card.file && card.file.originalname))) {
		card.status = 'inactive';
	}

	if (card.title) {
		card.title = card.title.substr(0, global.maxTitleLength);
	}

	card.effectiveDate = new Date();
}

function reconcileFileUpdate (newCard, oldCard) {
	if (!newCard.replaceFile) {
		newCard.file = oldCard.file;
	}
	delete newCard.replaceFile;
}

function setPersonIdsFromAry(ary, propName, idNameSet) {
	if (ary && ary.length) {
		for (var i = ary.length - 1; i >= 0; i--) {
			idNameSet[ary[i][propName]] = 'Unknown';
		};
	}
}

function setEditorIdOnCards (cardId, cards, audits) {
	if (cards && cards.length && audits && audits.length) {
		var auditSet = {};
		for (var i = audits.length - 1; i >= 0; i--) {
			if (cardId!=audits[i].itemCardId) {
				throw new Error('audit card ID unexpected: expected ' + cardId + ' and got ' + audits[i].itemCardId);
			}
			if (audits[i].effectiveDate) {
				auditSet[audits[i].effectiveDate.getTime()] = audits[i];
			}
		};
		for (var i = cards.length - 1; i >= 0; i--) {
			if (cardId!=cards[i].id) {
				throw new Error('card set card ID enexpected: expected ' + cardId + ' and got ' + cards[i].id);
			}
			if (cards[i].effectiveDate && auditSet[cards[i].effectiveDate.getTime()]) {
				cards[i].editorId = auditSet[cards[i].effectiveDate.getTime()].userId;
			}
		};
	}
}

function setNamesOnCards(cardAry, personIdNameOb) {
	if (personIdNameOb && cardAry && cardAry.length) {
		for (var i = cardAry.length - 1; i >= 0; i--) {
			if (cardAry[i].editorId) {
				cardAry[i].editorName = personIdNameOb[cardAry[i].editorId];
			}
			if (cardAry[i].ownerId) {
				cardAry[i].ownerName = personIdNameOb[cardAry[i].ownerId];
			}
		};
	}
}

function getEmptyCard() {
	return {
		title: '',
		descr: '',
		tags: '',
		version: '',
		createdBy: '',
		versionComment: '',
		status: 'inactive',
		ownerId: '',
		ownerName: '',
		canEdit: false,
		effectiveDate: new Date()
	}
}

module.exports = VisualStateModel;
