/* global process */
var express = require('express');
var router = express.Router();
var bodyParser = require('body-parser');
var cookieParser = require('cookie-parser');
var session = require('cookie-session');
//var favicon = require('serve-favicon');
var parseArgs = require('minimist');
var args = parseArgs(process.argv.slice(1));
var log4js = require('log4js');
var multiparty = require('multiparty');
var uiHelper = require('./ui_helper');

var log = log4js.getLogger('admin');
var global = require('../models/global');
var config_version = require('../../common/config_version');
var VisualStateModel = require('../models/VisualStateModel');
var messageMap = require('../models/errorMessages_en');
var clientserverCommon = require('../public/js/clientserverCommon');
var _ = require('underscore');

//** IAM Variables **
var EMAIL_HEADER = 'from';
var NAME_HEADER = 'x-name';
var ID_HEADER = 'adsamaccountname';

log.debug('args=' + JSON.stringify(args));
var config = null;
if (args.c) {
	args.config = args.c;
}

var dao = null;
if (args.config) {
	var config = require('../' + args.config);
	var Dao = require(config.dao);
	dao = new Dao();
}
else {
	throw new Error("No DAO configuration specified");
}

var vsm = new VisualStateModel(dao);

// uncomment after placing your favicon in /public
//app.use(favicon(__dirname + '/public/favicon.ico'));
var jsonParser = bodyParser.json();
var urlParser = bodyParser.urlencoded({ extended: false });
router.use(jsonParser);
router.use(urlParser);

router.use(cookieParser());
router.use(session({
	name: "pocketcard",
	keys: ['key1', 'key2'],
	maxage: 3600000,	//1 hour
	secureProxy: false
}));

router.all('*', function (req, res, next) {
	setupNocache(res);
	log.debug('=-=-=-=-=-=-=-= headers');
	log.debug(req.headers);
	vsm.authorize(uiHelper.getSessionState(req), {id: req.get(ID_HEADER), name: req.get(NAME_HEADER), email: req.get(EMAIL_HEADER)}, function (err) {
		next();
	});
});

router.get('/', function (req, res, next) {
	log.debug('go to home page...');
	var state = uiHelper.getSessionState(req);
	vsm.setNothingSelectedState(state, function () {
		if ('homeUser'===state.visualState) {
			res.redirect('manageusers');
		}
		else if ('homeCard'===state.visualState) {
			res.redirect('managecards');
		}
		else if (''==state.visualState) {
			res.redirect('noauth');
		}
		else {
			next();
		}
	});
});

router.get('/ping', function (req, res) {
	log.debug('ping :-)');
	var state = uiHelper.getSessionState(req);
	vsm.continueOkay(state, function (result) {
		if ('ok'==result) {
			return res.send('ok');
		}
		res.status(500).end('error');
	});
});

router.get('/noauth', function (req, res) {
	vsm.logout(uiHelper.getSessionState(req, true), function () {
		res.render('noauth');
	});
});

router.get('/exitout', function (req, res) {
	vsm.logout(uiHelper.getSessionState(req, true), function () {
		var redir = uiHelper.getExitUrl(process.env.APP_ENV, process.env.NODE_ENV);

		if (redir) {
			res.redirect(redir);
		}
		else {
			res.render('logout');
		}
	});
});

router.get('/logout', function (req, res) {
	vsm.logout(uiHelper.getSessionState(req, true), function () {
		var redir = uiHelper.getLogoutUrl(process.env.APP_ENV, process.env.NODE_ENV);
		
		if (redir) {
			res.redirect(redir);
		}
		else {
			res.render('logout');
		}
	});
});

router.get('/manageusers', function (req, res) {
	log.debug('goto manageusers...');
	var state = uiHelper.getSessionState(req);
	vsm.setManageUsersState(state, function (dModel) {
		if ('manageUsers'===state.visualState) {
			res.render('manageusers', {model: dModel, errors: getFilledInErrors(vsm.popAllErrors(state))});
		}
		else {
			res.redirect('./');
		}
	});
});

router.get('/managecards', function (req, res) {
	log.debug('/managecards');
	var state = uiHelper.getSessionState(req);
	vsm.setManageCardsState(state, function (dModel) {
		if ('manageCards'===state.visualState) {
			res.render('managecards', {model: dModel, errors: getFilledInErrors(vsm.popAllErrors(state))});
		}
		else {
			res.redirect('./');
		}
	});
});

router.get('/adduser', function (req, res) {
	log.debug('go to add user page...');
	var state = uiHelper.getSessionState(req);
	vsm.setAddUserState(state, function (dModel) {
		if ('addUser'==state.visualState) {
			res.render('adduser', {model: dModel, errors: getFilledInErrors(vsm.popAllErrors(state)), clientserverCommon: clientserverCommon});
		}
		else {
			res.redirect('./');
		}
	});
});

router.post('/adduser', function (req, res) {
	log.debug('posting user addition...');
	var newUser = {};
	var form = new multiparty.Form(getMultipartySettings());
	form.parse(req, function (err, fields, files) {
		//TODO: handle unexpected files
		//TODO: handle error
		uiHelper.setupNewUser2(newUser, fields);
		var state = uiHelper.getSessionState(req);
		vsm.saveNewUser(state, newUser, function (result) {
			if ('selectedUser'===state.visualState) {
				res.redirect('selecteduser');
			}
			else if ('nothingSelected'===state.visualState) {
				res.redirect('adduser');
			}
			else {
				res.redirect('./');
			}
		});
	});
});

router.get('/selecteduser', function (req, res) {
	log.debug('display selected user');
	var state = uiHelper.getSessionState(req);
	vsm.getSelectedUserModel(state, function (dModel) {
		if ('userSelected'===state.visualState) {
			res.render('selecteduser', {model: dModel, errors: getFilledInErrors(vsm.popAllErrors(state))});
		}
		else {
			res.redirect('./');
		}
	});
});

router.get('/addcard', function (req, res) {
	log.debug('go to add card page...');
	var state = uiHelper.getSessionState(req);
	vsm.setAddCardState(state, function (dModel) {
		if ('addCard'===state.visualState) {
			res.render('addcard', {model: dModel, global: global, errors: getFilledInErrors(vsm.popAllErrors(state))});
		}
		else {
			res.redirect('./');
		}
	});
});

router.post('/addcard', function (req, res, next) {
	log.debug('posting card addition...');
	var form = new multiparty.Form(getMultipartySettings());
	form.parse(req, function (err, fields, files) {
		//TODO: handle unexpected files
		//TODO: handle error
		uiHelper.getCard2(fields, files, function (err, newCard) {
			var state = uiHelper.getSessionState(req);
			if (err) {
				log.error(err);
			}
			vsm.saveNewCard(state, newCard, function (result) {
				if ('selectedCard'===state.visualState) {
					res.redirect('selectedcard');
				}
				else if ('nothingSelected'===state.visualState) {
					res.redirect('addcard');
				}
				else {
					res.redirect('./');
				}
			});
		});
	});
});

router.get('/selectedcard', function (req, res, next) {
	log.debug('display selected card');
	var state = uiHelper.getSessionState(req);
	vsm.getSelectedCardModel(state, function (dModel) {
		if ('selectedCard'===state.visualState) {
			res.render('selectedcard', {model: dModel, uiHelper: uiHelper, errors: getFilledInErrors(vsm.popAllErrors(state)), clientserverCommon: clientserverCommon});
		}
		else {
			res.redirect('./');
		}
	});
});

router.get('/selectcard', function (req, res, next) {
	log.debug('select-card action: query=');
	log.debug(req.query);
	if (req.query.cardId) {
		var state = uiHelper.getSessionState(req);
		vsm.selectCard(state, req.query.cardId, function (dModel) {
			if ('selectedCard'===state.visualState) {
				res.redirect('selectedcard');
			}
			else {
				res.redirect('./');
			}
		});
	}
	else {
		next();
	}
});

router.get('/selectuser', function (req, res, next) {
	log.debug('selectuser action: query=%s', req.query);
	if (req.query.userId) {
		var state = uiHelper.getSessionState(req);
		vsm.selectUser(state, req.query.userId, function (dModel) {
			if ('selectedUser'===state.visualState) {
				res.redirect('selecteduser');
			}
			else {
				res.redirect('./');
			}
		});
	}
	else {
		next();
	}
});

router.get('/cards', function (req, res) {
	log.debug('/cards');
	vsm.getAllCurrentCards(req.session.stateObject, function (err, result) {
		if (err) {
			res.status(500).end();
		}
		else {
			res.send(result);
		}
	});
});

router.get('/editcard', function (req, res) {
	log.debug('editcard GET...');
	var state = uiHelper.getSessionState(req);
	vsm.setEditCardState(state, function (dModel) {
		if ('editCard'===state.visualState) {
			res.render('editcard', {model: dModel, global: global, errors: getFilledInErrors(vsm.popAllErrors(state))});
		}
		else {
			res.redirect('./');
		}
	});
});

router.post('/editcard', function (req, res) {
	log.debug('editcard POSTing changes...');
	var form = new multiparty.Form(getMultipartySettings());

	form.on('error', function(err){
		console.log("Error parsin from: " + err.stack);
	});


	form.parse(req, function (err, fields, files) {
		//TODO: handle unexpected files
		//TODO: handle error
		uiHelper.getCard2(fields, files, function (err, newCard) {
			var state = uiHelper.getSessionState(req);
			vsm.updateCard(state, newCard, function (dModel) {
				if ('selectedCard'===state.visualState) {
					res.redirect('selectedcard');
				}
				else {
					res.redirect('./');
				}
			});
		});
	});
});

router.get('/edituser', function (req, res) {
	log.debug('edituser GET...');
	var state = uiHelper.getSessionState(req);
	vsm.setEditUserState(state, function (dModel) {
		if ('editUser'===state.visualState) {
			res.render('edituser', {model: dModel, errors: getFilledInErrors(vsm.popAllErrors(state)), clientserverCommon: clientserverCommon});
		}
		else {
			res.redirect('./');
		}
	});
});

router.post('/edituser', function (req, res, next) {
	log.debug('edituser POSTing changes...');
	var form = new multiparty.Form(getMultipartySettings());
	form.parse(req, function (err, fields, files) {
		//TODO: handle unexpected files
		//TODO: handle error
		var newUser = uiHelper.getUser2(fields);
		var state = uiHelper.getSessionState(req);
		vsm.updateUser(state, newUser, function (dModel) {
			if ('selectedUser'===state.visualState) {
				res.redirect('selecteduser');
			}
			else if (''===state.visualState) {
				res.redirect('./');
			}
			else {
				next();
			}
		});
	});
});

router.get('/carddata', function (req, res) {
	log.debug('/carddata');
	vsm.getSelectedCard(req.session.stateObject, function (err, card) {
		if (card) {
			res.send(card);
		}
		else if (err) {
			res.status(500).end(err.toString());
		}
		else {
			res.status(500).end('not found');
		}
	});
});

router.get('/users', function (req, res) {
	log.debug('/users');
	vsm.getUsers(req.session.stateObject, function (err, results) {
		if (err) {
			res.status(500).end(err.toString());
		}
		else {
			res.send(results);
		}
	});
});

router.get('/userdata', function (req, res) {
	log.debug('/userdata');
	vsm.getSelectedUser(req.session.stateObject, function (err, result) {
		if (err) {
			return res.status(500).end(err.toString());
		}
		res.send(result);
	});
});

router.get('/downloadcardcontent', function (req, res) {
	log.debug('/downloadcardcontent');
	log.debug('...with query ');
	log.debug(req.query);
	vsm.getCardFile(req.session.stateObject, req.query.id, function (err, cardfile) {
		if (err) {
			return res.status(500).end(err.toString());
		}
		if (!cardfile || !cardfile.buffer) {
			return res.status(500).end(new Error('card file not found'));
		}
		res.setHeader('Content-disposition', 'attachment; filename=' + cardfile.originalname);
		res.setHeader('Content-type', 'text/text');
		res.send(cardfile.buffer.buffer);
	});
});

router.get('/getcardcontent', function (req, res) {
	log.debug('/getcardcontent');
	log.debug('...with query ');
	log.debug(req.query);
	vsm.getCardFile(req.session.stateObject, req.query.id, function (err, cardfile) {
		if (err) {
			return res.status(500).end();
		}
		if (!cardfile || !cardfile.buffer) {
			return res.status(500).end(new Error('card file not found'));
		}
		res.setHeader('Content-disposition', 'filename=' + cardfile.originalname);
		res.setHeader('Content-type', 'text/html');
		res.send(cardfile.buffer.buffer);
	});
});

router.get('/cardactivity', function (req, res) {
	log.debug('/cardactivity');
	var state = uiHelper.getSessionState(req);
	vsm.setViewCardHistoryState(state, function (dModel) {
		if ('cardHistory'===state.visualState) {
			res.render('cardactivity', {model: dModel, errors: getFilledInErrors(vsm.popAllErrors(state))});
		}
		else {
			res.redirect('./');
		}
	});
});

router.get('/cardactivitydata', function (req, res) {
	log.debug('/cardactivitydata');
	vsm.getCardActivity(req.session.stateObject, function (err, results) {
		if (err) {
			return res.status(500).end(err.toString());
		}
		res.send(results);
	});
});

router.get('/cardversions', function (req, res) {
	log.debug('/cardversions');
	vsm.getCardVersions(req.session.stateObject, function (err, results) {
		if (err) {
			return res.status(500).end(err.toString());
		}
		res.send(results);
	});
});

router.get('/help', function (req, res) {
	log.debug('/help');
	var state = uiHelper.getSessionState(req)
	vsm.setHelpState(state, function (dModel) {
		if ('help'===state.visualState) {
			res.render('help', {model: dModel});
		}
		else {
			res.redirect('./');
		}
	});
});

router.get('/about', function (req, res) {
	log.debug('/about');
	var state = uiHelper.getSessionState(req);
	vsm.setAboutState(state, function (dModel) {
		if ('about'===state.visualState) {
			res.render('about', {model: dModel, version: config_version.version, lastUpdate: uiHelper.getMonthYear(config_version.lastUpdate)});
		}
		else {
			res.redirect('./');
		}
	});
});

router.get('/allcurrentusersnames', function (req, res) {
	log.debug('/allCurrentUsersNames');
	vsm.getAllCurrentUserNames(req.session.stateObject, function (err, results) {
		if (err) {
			return res.status(500).end(err.toString());
		}
		res.send(results);
	});
});

router.get('/cardmetametrics', function (req, res) {
	log.debug('/cardmetametrics');
	var state = uiHelper.getSessionState(req);
	vsm.setCardsMetricsDownloadState(state, function (dModel) {
		if ('cardsmetrics'===state.visualState) {
			res.render('cardmetametrics', {model: dModel, errors: getFilledInErrors(vsm.popAllErrors(state))});
		}
		else {
			res.redirect('./');
		}
	});
});

router.get('/downloadmetametrics', function (req, res) {
	log.debug('/downloadmetametrics');
	log.debug('...with query ');
	log.debug(req.query);
	vsm.getMetaMetricsIterator(req.session.stateObject, uiHelper.getTimeDate(req.query.from), uiHelper.getTimeDate(req.query.to), function (err, iter) {
		if (err) {
			if (iter) {
				iter.done();
			}
			return res.status(500).end(err.toString());
		}
		res.setHeader('Content-disposition', 'attachment; filename=metametrics.csv');
		res.setHeader('Content-type', 'text/text');
		uiHelper.writeMetricsWithIterator2(res, iter, function (err) {
			if (err) {
				res.status(500).end(err.toString());
			}
			else {
				res.end();
			}
		}, true);
	});
});

router.get('/downloadcardmetrics', function (req, res) {
	log.debug('/downloadcardmetrics');
	vsm.getCardMetricsIterator(req.session.stateObject, uiHelper.getTimeDate(req.query.from), uiHelper.getTimeDate(req.query.to), function (err, iter) {
		if (err) {
			if (iter) {
				iter.done();
			}
			return res.status(500).end(err.toString());
		}
		res.setHeader('Content-disposition', 'attachment; filename=cardmetrics.csv');
		res.setHeader('Content-type', 'text/text');
		uiHelper.writeMetricsWithIterator2(res, iter, function (err) {
			if (err) {
				res.status(500).end(err.toString());
			}
			else {
				res.end();
			}
		}, true);
	});
});

router.get('/cardmetrics', function (req, res, next) {
	log.debug('/cardmetrics');
	var state = uiHelper.getSessionState(req);
	vsm.setCardMetricsDownloadState(state, function (dModel) {
		if ('cardmetrics'===state.visualState) {
			res.render('cardmetrics', {model: dModel, uiHelper: uiHelper, errors: getFilledInErrors(vsm.popAllErrors(state))});
		}
		else {
			res.redirect('./');
		}
	});
});

router.get('/isidused', function (req, res) {
	log.debug('/');
	if (req.query && req.query.id) {
		vsm.isSystemIdUsed(uiHelper.getSessionState(req), req.query.id, function (err, result) {
			if (err) {
				res.status(500).end(err.toString());
			}
			else {
				res.send(result);
			}
		});
	}
	else {
		res.status(500).end();
	}
});

function setupNocache(res) {
	res.setHeader('Last-Modified', (new Date()).toUTCString());
	res.setHeader('Cache-Control', 'no-cache, no-store, must-revalidate');
	res.setHeader('Pragma', 'no-cache');
	res.setHeader('Expires', 0);
}

function getMultipartySettings() {
	return {
		uploadDir: './uploads',
		maxFilesSize: 10000000,
		maxFields: 100
	};
}

function getFilledInErrors(ary) {
	if (ary && Array.isArray(ary) && ary.length) {
		ary = _.map(ary, function (key) {return key + ' - ' + messageMap.getMessage(key);});
	}
	return ary;
}

module.exports = router;
