//Unit tests for Dummy_dao.js
//
var should = require('should');
var Dummy_dao = require('../dao/AdminDao_Dummy');
var log4js = require('log4js');
var log = log4js.getLogger('AdminDao_Dummy_test');
var helper = require('../../common/utils');
var Daoi = require('../dao/IAdminDao.js');

describe('Dummy_dao_test', function () {
	describe('saveNewUser()', function () {
		describe('happy path', function () {
			it ('save new card user', function (done) {
				var dao = new Daoi(new Dummy_dao());
				dao.dao.db.reset();
				var newUser = {
					'userType': 'card',
					'lastname': 'jones',
					'firstname': 'fred',
					'status': 'active',
					'effectiveDate': new Date(),
					'activeDirId': 'vhaiszxxx'
				};
				dao.saveNewUser({'id':1234}, newUser, function (err, result) {
					if (err) {
						throw err;
					}
					result.should.equal(5001);
					done();
				});
			});
		});
	});

	describe('updateUser()', function () {
		describe('existing user', function() {
			describe('happy path', function () {
				var dao, userId = null;
				var rootUser = getRootUser();

				beforeEach(function (done) {
					dao = new Daoi(new Dummy_dao());
					dao.dao.db.reset();
					dao.saveNewUser({id:1234}, rootUser, function (err, result) {
						if (err) {
							throw err;
						}
						userId = result;
						done();
					});
				});

				it ('saves with only initial record', function (done) {
					var updatedUser = helper.getMerged(rootUser, {'id': userId, 'firstname': 'fredrick', 'effectiveDate': new Date('2014-11-02')});
					dao.updateUser(updatedUser, function (err) {
						if (err) {
							throw err;
						}
						var recs = getUserRecords(dao, userId);
						recs.length.should.equal(2);
						recs[0].id.should.equal(recs[1].id);
						recs[0].lastname.should.equal(recs[1].lastname);
						recs[0].firstname.should.not.equal(recs[1].firstname);
						recs[0].user_type.should.equal(recs[1].user_type);
						recs[0].status.should.equal(recs[1].status);
						should(isDateInRange(recs[0].effective_date)).be.ok;
						should(isDateInRange(recs[1].effective_date)).be.ok;
						recs[0].active_dir_id.should.equal(recs[1].active_dir_id);
						should(recs[0].effective_date != recs[1].effective_date).be.ok;
						done();
					});
				});

				it ('save with more than one record', function (done) {
					var updatedUser = helper.getMerged(rootUser, {'id': userId, 'firstname': 'fredrick', 'effectiveDate': new Date('2014-11-02')});
					dao.updateUser(updatedUser, function (err) {
						if (err) {
							throw err;
						}
						var update1 = helper.getMerged(updatedUser, {'status': 'inactive'});
						dao.updateUser(update1, function (err) {
							if (err) {
								throw err;
							}
							var recs = getUserRecords(dao, userId);
							recs.length.should.equal(3);
							done();
						});
					});
				});

			});
		});
	});

	describe('getCurrentUserById()', function () {
		var dao, userId = null;
		var adminUser = {'id': 1234};
		var rootUser = getRootUser();

		beforeEach(function () {
			dao = new Daoi(new Dummy_dao());
			dao.dao.db.reset();
		});

		it ('returns empty with no records in DB', function (done) {
			dao.getCurrentUserById(5001, function (err, result) {
				if (err) {
					throw err;
				}
				should(result).be.not.ok;
				done();
			});
		});

		it ('returns a single record with an older entry', function (done) {
			dao.saveNewUser(adminUser, rootUser, function (err, result) {
				if (err) {
					throw err;
				}
				userId = result;
				dao.getCurrentUserById(userId, function (err, currUser) {
					if (err) {
						throw err;
					}
					currUser.id.should.equal(userId);
					currUser.firstname.should.equal(rootUser.firstname);
					currUser.lastname.should.equal(rootUser.lastname);
					currUser.activeDirId.should.equal(rootUser.activeDirId);
					currUser.status.should.equal(rootUser.status);
					currUser.effectiveDate.should.equal(rootUser.effectiveDate);
					currUser.userType.should.equal(rootUser.userType);
					done();
				});
			});
		});

		it ('returns empty with one future dated record', function (done) {
			rootUser.effectiveDate = addDays(new Date(), 1);
			dao.saveNewUser(adminUser, rootUser, function (err, result) {
				if (err) {
					throw err;
				}
				userId = result;
				dao.getCurrentUserById(userId, function (err, currUser) {
					if (err) {
						throw err;
					}
					should(currUser).be.not.ok;
					done();
				});
			});
		});

		it ('returns latest record with multiple past dated records', function (done) {
			dao.saveNewUser(adminUser, rootUser, function (err, result) {
				if (err) {
					throw err;
				}
				userId = result;
				var nextUser = helper.getMerged(rootUser, {'id': userId, 'firstname': 'fredrick', 'effectiveDate': new Date('2014-11-01 00:00:10')});
				dao.updateUser(nextUser, function (err) {
					if (err) {
						throw err;
					}
					dao.getCurrentUserById(userId, function (err, result) {
						if (err) {
							throw err;
						}
						result.firstname.should.equal('fredrick');
						should(result.effectiveDate>new Date('2014-11-01 00:00:00')).be.ok;
						done();
					});
				});
			});
		});
	});

	describe('getUsers()', function () {
		it ('happy path', function (done) {
			var dao = new Daoi(new Dummy_dao());
			dao.dao.db.reset();
			dao.getUsers(function (err, results) {
				if (err) {
					throw err;
				}
				results.length.should.equal(5);
				done();
			});
		});
	});

	describe ('getCurrentUserBySysid()', function () {
		var dao = null;

		beforeEach(function () {
			dao = new Daoi(new Dummy_dao());
			dao.dao.db.reset();
		});

		it ('happy path - with historical duplicates', function (done) {
			dao.getCurrentUserBySysid('vhaisxwinchc', function (err, results) {
				if (err) throw err;
				results.length.should.equal(1);
				results[0].id.should.equal(4321);
				done();
			});
		});

		it ('happy path - with no historical duplicates', function (done) {
			dao.getCurrentUserBySysid('vhaisxjonesf', function (err, results) {
				if (err) throw err;
				results.length.should.equal(1);
				results[0].id.should.equal(1234);
				done();
			});
		});

		it ('can\'t find old system ID', function (done) {
			dao.getCurrentUserBySysid('vhaisxjonesj', function (err, users) {
				if (err) throw err;
				should(users).be.not.ok;
				done();
			});
		});
	});

	describe('getCardFile()', function () {
		var dao = null;

		beforeEach(function () {
			dao = new Daoi(new Dummy_dao());
			dao.dao.db.reset();
			dao.dao.db.cards.push(getMockCard());
		});

		it ('happy path', function (done) {
			dao.getCardFile(1, function (err, result) {
				if (err) throw err;
				result.originalname.should.equal('small.html');
				done();
			});
		});
	});

});

function getMockCard () {
	return {
		id: 1,
		descr: 'test card',
		name: 'test',
		effectiveDate: new Date(),
		file: {
			fieldname: 'file1',
			originalname: 'small.html',
			encoding: '7bit',
			mimetype: 'text/html',
			path: '/',
			extension: 'html',
			size: 88,
			truncated: false,
			buffer: {
				_bsontype: 'Binary',
				sub_type: 0,
				position: 88,
				buffer: new Buffer('This is a test')
			}
		}
	};
}

function assertUser (adid, lastname, user) {
	user.lastname.should.equal(lastname);
	user.activeDirId.should.equal(adid);
}

function addDays(aDate, numberOfDays) {
	var ms = aDate.getTime() + (86400000 * numberOfDays);
	return new Date(ms);
}

function getRootUser() {
	return {
		'userType': 'card',
		'lastname': 'jones',
		'firstname': 'fred',
		'status': 'active',
		'effectiveDate': new Date('2014-11-01 00:00:00'),
		'activeDirId': 'vhaiszxxx'
	};

}

function isDateInRange(aDate) {
	return aDate > new Date('2014-10-31') && aDate < new Date('2015-01-01');
}

function getUserRecords(dao,id) {
	var results = [];
	for (var i=0; i<dao.dao.db.admin_users.length; i++) {
		if (id==dao.dao.db.admin_users[i].id) {
			results.push(dao.dao.db.admin_users[i]);
		}
	}
	return results;
}

function addTestMetrics(list) {
	list.push({deviceHash : "abcdefghijk", clientIp : "127.0.0.1", request : "/pocketcardv1/user/abcdefghijk/cards", timestamp : new Date("2015-04-01 08:00:00")});
	list.push({deviceHash : "abcdefghijk", clientIp : "127.0.0.1", request : "/pocketcardv1/user/abcdefghijk/cards?state=active", timestamp : new Date("2015-04-01 09:00:00")});
	list.push({deviceHash : "abcdefghijk", clientIp : "127.0.0.1", request : "/pocketcardv1/user/abcxyz112233/cards/10101", timestamp : new Date("2015-04-01 10:00:00")});
}
