/* global Buffer */
/* global afterEach */
/* global beforeEach */
/* global it */
/* global describe */
var should = require('should');
var Dao = require('../dao/AdminDao_Dummy');
var log4js = require('log4js');
var log = log4js.getLogger('VisualStateModel_test');
var daoUtils = require('../dao/dao_utils');
log4js.getLogger('VisualStateModel').setLevel('FATAL');
log4js.getLogger('IAdminDao').setLevel('FATAL');
log.setLevel('FATAL');

var VisualStateModel = require('../models/VisualStateModel');

describe('VisualStateModel instance', function () {
	describe('resetState()', function () {
		var dao = new Dao();
		var model = new VisualStateModel(dao);
		var state = {};
		model.resetState(state);
		it('should have empty defaults', function () {
			state.errors.should.be.empty;
			state.visualState.should.equal("");
		});
	});

	describe('setNothingSelectedState()', function () {
		var model = null;

		beforeEach(function () {
			model = new VisualStateModel(new Dao());
		});

		it ('goes to logoff if no one logged in', function (done) {
			var state = {};
			model.setNothingSelectedState(state, function (result) {
				state.visualState.should.equal('');
				should(result).be.not.ok;
				done();
			});
		});

		it ('goes to logoff if user type not recognized', function (done) {
			var state = {userName: 'fred', userId: 111, userType: 'jan'};
			model.setNothingSelectedState(state, function (result) {
				state.visualState.should.equal('');
				should(result).be.not.ok;
				done();
			});
		});

		it ('goes to user admin home page', function (done) {
			var state = {userName: 'fred', userId: 111, userType: 'user'};
			model.setNothingSelectedState(state, function (result) {
				state.visualState.should.equal('homeUser');
				assertBasicDisplayModel(result, state.userName, state.userId, state.userType);
				done();
			});
		});

		it ('goes to card admin home page', function (done) {
			var state = {userName: 'fred', userId: 111, userType: 'card'};
			model.setNothingSelectedState(state, function (result) {
				state.visualState.should.equal('homeCard');
				assertBasicDisplayModel(result, state.userName, state.userId, state.userType);
				done();
			});
		});

	});
	
	describe('logout()', function () {
		it ('does nothing if there is no state', function (done) {
			var model = new VisualStateModel(new Dao());
			model.logout(null, function (result) {
				result.should.equal('');
				done();
			});
		});
	});

	describe('all state changes lead to login if not logged in', function () {
		var model = new VisualStateModel(new Dao());
		var state = {};

		beforeEach(function () {
			model.resetState(state);
		});

		it('should goto login from logout', function (done) {
			model.logout(state, function (result) {
				result.should.equal("");
				done();
			});
		});

		it ('should goto login from selectCard', function (done) {
			model.selectCard(state, 1234, function (result) {
				result.should.equal("");
				done();
			});
		});

		it ('should goto login from saveNewCard', function (done) {
			model.saveNewCard(state, {}, function (result) {
				should(model.popError(state)).equal(31);
				should(model.popError(state)).be.not.ok;
				result.should.equal("");
				done();
			});
		});

		it ('should goto login from selectUser', function (done) {
			model.selectUser(state, '4321', function (result) {
				result.should.equal("");
				done();
			});
		});

		it ('should goto login from saveNewUser', function (done) {
			model.saveNewUser(state, {}, function (result) {
				result.should.equal("");
				done();
			});
		});

		it ('should goto login from updateCard', function (done) {
			model.updateCard(state, {}, function (result) {
				result.should.equal("");
				model.popError(state).should.equal(120);
				done();
			});
		});

		it ('should goto login from help', function (done) {
			model.setHelpState(state, function (result) {
				result.should.equal('');
				should(model.popError(state)).be.not.ok;
				done();
			});
		});

		it ('should goto login from about', function (done) {
			model.setAboutState(state, function (result) {
				result.should.equal('');
				should(model.popError(state)).be.not.ok;
				done();
			});
		});

		it ('should goto login from manageCards state', function (done) {
			model.setManageCardsState(state, function (result) {
				state.visualState.should.equal('');
				should(model.popError(state)).be.not.ok;
				done();
			});
		});
	});

	describe("saveNewCard()", function () {
		var model = new VisualStateModel(new Dao());
		var state = null;

		beforeEach(function () {
			state = {};
			model.resetState(state);
			state.userId = 1234;
			model.dao.dao.db.cards = [];
		});

		var userTypes1 = getUserTypeArray();

		for (var i=userTypes1.length-1;i>=0;i--) {
			it("goes to selectedCard when saved successfully userType='" + userTypes1[i] + "'", function (done) {
				var userType = userTypes1.pop();
				state.userType = userType;
				//log.info('userType=' + userType);
				var testDate = new Date('2013-01-01 00:00:00');
				var newCard = getMockCard1(testDate);
				newCard.title = 'This title is longer than 50 characters blah blah blah blah blah';
				model.saveNewCard(state, newCard, function (result) {
					var error = model.popError(state);
					if (error) {
						log.debug('found error! :');
						log.debug(error);
					}
					should(error).be.not.ok;
					state.visualState.should.equal('selectedCard');
					state.selectedCard.should.equal(1);
					assertBasicDisplayModel(result, null, 1234, userType);
					var savedCard = model.dao.dao.db.cards[0];
					savedCard.ownerId.should.equal(state.userId);
					savedCard.effectiveDate.should.not.equal(testDate);
					savedCard.status.should.equal('active');
					savedCard.title.should.equal('This title is longer than 50 characters blah blah ');
					done();
				});
			});
		}

		var userTypes2 = getUserTypeArray();

		for (var i=userTypes2.length-1; i>=0; i--) {
			it("goes to addCard when it has a failure userType='" + userTypes2[i] + "'", function (done) {
				state.userType = userTypes2.pop();
				var newCard = getMockCard1();
				delete newCard.versionComment;
				model.saveNewCard(state, newCard, function (result) {
					state.visualState.substr(0,4).should.equal('home');
					model.popError(state).should.equal(32);
					should(model.popError(state)).be.not.ok;
					should(state.selectedCard).be.not.ok;
					assertBasicDisplayModel(result, null, 1234, state.userType);
					done();
				});
			});
		}

		it ("changes 'active' status to 'inactive' if file property is not found", function (done) {
			var newCard = getMockCard1();
			newCard.file = null;
			state.userType = 'card';
			model.saveNewCard(state, newCard, function (result) {
				state.visualState.should.equal('selectedCard');
				var sCard = model.dao.dao.db.cards[0];
				sCard.ownerId.should.equal(state.userId);
				sCard.status.should.equal('inactive');
				assertBasicDisplayModel(result, null, 1234, 'card');

				model.dao.dao.db.reset();
				newCard = getMockCard1();
				newCard.file.originalname = null;
				model.saveNewCard(state, newCard, function (result) {
					state.visualState.should.equal('selectedCard');
					sCard = model.dao.dao.db.cards[0];
					sCard.ownerId.should.equal(state.userId);
					sCard.status.should.equal('inactive');
					assertBasicDisplayModel(result, null, 1234, 'card');
					done();
				});
			});
		});

		it ("sets status to 'inactive' if status property is invalid", function (done) {
			var newCard = getMockCard1();
			newCard.status = 'fred';
			state.userType = 'card';
			model.saveNewCard(state, newCard, function (result) {
				state.visualState.should.equal('selectedCard');
				var sCard = model.dao.dao.db.cards[0];
				sCard.ownerId.should.equal(state.userId);
				sCard.status.should.equal('inactive');
				assertBasicDisplayModel(result, null, 1234, 'card');
				done();
			});
		});

		it ('throws an error if status="active" and there is no versionComment', function (done) {
			var newCard = getMockCard1();
			delete newCard.versionComment;
			state.userType = 'card';
			model.saveNewCard(state, newCard, function (result) {
				model.popError(state).should.equal(32);
				state.visualState.should.equal('homeCard');
				assertBasicDisplayModel(result, null, 1234, 'card');
				done();
			});
		});

		it ('saves a new card with status="inactive" with no versionComment', function (done) {
			var newCard = getMockCard1();
			delete newCard.versionComment;
			newCard.status = 'inactive';
			state.userType = 'card';
			model.saveNewCard(state, newCard, function(result) {
				should(model.popError(state)).be.not.ok;
				state.visualState.should.equal('selectedCard');
				assertBasicDisplayModel(result, null, 1234, 'card');
				model.dao.dao.db.cards.length.should.equal(1);
				done();
			});
		});

	});

	describe("selectUser()", function () {
		var model = new VisualStateModel(new Dao());
		var state = {};

		beforeEach(function () {
			model.resetState(state);
			state.userId = '5432';
		});

		it("works for an admin user", function (done) {
			state.userType = 'user';
			model.selectUser(state, '112233', function (result) {
				state.selectedUser.should.equal(112233);
				state.visualState.should.equal('selectedUser');
				assertBasicDisplayModel(result, null, '5432', 'user');
				model.selectUser(state, 332211, function (result) {
					state.selectedUser.should.equal(332211);
					state.visualState.should.equal('selectedUser');
					assertBasicDisplayModel(result, null, '5432', 'user');
					done();
				});
			});
		});

		it("fails for any non admin user", function (done) {
			state.userType = 'jelly';
			model.selectUser(state, '112233', function (result) {
				should(state.selectedUser).be.not.ok;
				model.popError(state).should.equal(41);
				should(model.popError(state)).be.not.ok;
				result.should.equal('');
				done();
			});
		});
	});

	describe("saveNewUser()", function () {
		var model = new VisualStateModel(new Dao());
		var state = {};
		var starttime = null;

		beforeEach(function () {
			model.resetState(state);
			state.userId = 1234;
			state.userType = 'user';
			model.dao.dao.db.reset();
			starttime = new Date();
		});

		it("works for an admin user", function (done) {
			var newUser = getMockUser1();
			model.saveNewUser(state, newUser, function (result) {
				state.visualState.should.equal('selectedUser');
				assertBasicDisplayModel(result, null, 1234, 'user');
				should(model.popError(state)).not.be.ok;
				state.selectedUser.should.equal(5001);

				var admin_users = model.dao.dao.db.admin_users;
				var savedUser = null;
				for (var i = admin_users.length - 1; i >= 0; i--) {
					if (5001==admin_users[i].id) {
						savedUser = admin_users[i];
						break;
					}
				};
				savedUser.active_dir_id.should.equal(newUser.activeDirId.toLowerCase());
				var audits = model.dao.dao.db.audits;
				audits.length.should.equal(2);
				audits[0].userId.should.equal(1234);
				audits[0].action.substr(0,16).should.equal('insert new user ');
				audits[0].effectiveDate.should.equal(newUser.effectiveDate);
				should(audits[0].itemUserId).be.not.ok;
				var now = new Date();
				should(starttime.getTime()<=audits[0].datetime.getTime() && audits[0].datetime.getTime() <= now.getTime()).be.ok;
				audits[1].userId.should.equal(1234);
				audits[1].itemUserId.should.equal(5001);
				audits[1].action.should.equal('added user');
				audits[1].datetime.should.equal(audits[0].datetime);
				audits[1].effectiveDate.should.equal(newUser.effectiveDate);
				done();
			});
		});

		it ('automatically adds an effectiveDate', function (done) {
			var mockUser = getMockUser1();
			mockUser.effectiveDate = null;
			model.saveNewUser(state, mockUser, function (result) {
				state.visualState.should.equal('selectedUser');
				assertBasicDisplayModel(result, null, 1234, 'user');
				var newUser = null;
				for (var i=0; i<model.dao.dao.db.admin_users.length; i++) {
					if (state.selectedUser==model.dao.dao.db.admin_users[i].id) {
						newUser = model.dao.dao.db.admin_users[i];
					}
				}
				if (!newUser) {
					throw new Error('newUser not found');
				}
				Object.prototype.toString.call(newUser.effective_date).should.equal('[object Date]');
				done();
			});
		});

		it("goes to addUser when it has a failure", function (done) {
			model.saveNewUser(state, {}, function (result) {
				state.visualState.should.equal('homeUser');
				model.popError(state).should.equal(53);
				should(model.popError(state)).not.be.ok;
				should(state.selectedUser).not.be.ok;
				done();
			});
		});

		it("fails on a non-admin user", function (done) {
			state.userType = 'wow';
			model.saveNewUser(state, {userType: 'user', lastname: 'Jones', firstname: 'Fred', activeDirId: 'VHAISZXXX', effectiveDate: '2014-12-02 00:00:00'}, function (result) {
				result.should.equal('');
				model.popError(state).should.equal(51);
				should(model.popError(state)).be.not.ok;
				should(state.selectedUser).be.not.ok;
				done();
			});
		});

		it ('fails on missing lastname', function (done) {
			invalidSaveNewUserTest(model, state, {userType:'user', firstname: 'Fred', activeDirId:'vhaiszxxxx', effectiveDate: new Date()}, 58, done);
		});

		// it ('fails on invalid lastname', function (done) {
		// 	invalidSaveNewUserTest(model, state, {userType:'user', firstname: 'Fred', lastname:'woa@', activeDirId:'vhaiszxxxx', effectiveDate: new Date()}, 58, done);
		// });

		it ('fails on missing firstname', function (done) {
			invalidSaveNewUserTest(model, state, {userType:'user', lastname: 'Fred', activeDirId:'vhaiszxxxx', effectiveDate: new Date()}, 59, done);
		});

		// it ('fails on invalid firstname', function (done) {
		// 	invalidSaveNewUserTest(model, state, {userType:'user', firstname: 'yikes!', lastname: 'Fred', activeDirId:'vhaiszxxxx', effectiveDate: new Date()}, 59, done);
		// });

		function invalidSaveNewUserTest(aModel, state, newUser, errorCode, done) {
			aModel.saveNewUser(state, newUser, function (result) {
				state.visualState.should.equal('homeUser');
				aModel.popError(state).should.equal(errorCode);
				should(aModel.popError(state)).be.not.ok;
				should(state.selectedUser).be.not.ok;
				done();
			});
		}

		it ('fails on missing activeDirId', function (done) {
			var mUser = getMockUser1();
			delete mUser.activeDirId;
			invalidSaveNewUserTest(model, state, mUser, 53, done);
		});

		// it ('fails on invalid activeDirId', function (done) {
		// 	var mUser = getMockUser1();
		// 	mUser.activeDirId = "fred 1";
		// 	invalidSaveNewUserTest(model, state, mUser, 53, done);
		// });

		it ('fails if activeDirId already exists', function (done) {
			var mUser = getMockUser1();
			model.saveNewUser(state, mUser, function (result) {
				state.selectedUser.should.equal(5001);
				state.visualState.should.equal('selectedUser');
				mUser = getMockUser1();
				mUser.firstname = 'jane';
				model.saveNewUser(state, mUser, function (result) {
					should(state.selectedUser).be.not.ok;
					model.popError(state).should.equal(72);
					state.visualState.should.equal('homeUser');
					should(model.popError(state)).be.not.ok;
					done();
				});
			});
		});
	});

	describe("popAllErrors()", function () {
		var model = new VisualStateModel(new Dao());
		var state = {};

		beforeEach(function () {
			model.pushError(state, 1);
			model.pushError(state, 2);
		});

		it("returns and clears errors", function () {
			state.errors.length.should.equal(2);
			var errors = model.popAllErrors(state);
			state.errors.length.should.equal(0);
			errors[0].should.equal(1);
			errors[1].should.equal(2);
			errors.length.should.equal(2);
		});
	});

	describe('assertLoggedIn()', function () {
		var model = null;
		var state = {};

		beforeEach(function () {
			model = new VisualStateModel(new Dao());
		});

		it ('fails if not logged in', function () {
			should(model.assertLoggedIn(state)).be.not.ok;
		});

		it ('should succeed if logged in', function () {
			state.userId = 1234;
			state.userType = 'card';
			should(model.assertLoggedIn(state)).be.ok;
		});
	});

	describe('assertAdminLoggedIn()', function () {
		var model = null;
		var state = {};

		beforeEach(function () {
			model = new VisualStateModel(new Dao());
		});

		it('fails when logged in as non-user admin', function () {
			state.userId = 1234;
			state.userType = 'card';
			should(model.assertAdminLoggedIn(state)).be.not.ok;
		});

		it('succeeds when logged in as user admin', function () {
			state.userId = 1234;
			state.userType = 'user';
			should(model.assertAdminLoggedIn(state)).be.ok;
		});
	});

	describe('getSelectedUser()', function () {
		var model = null;
		var state = {};

		beforeEach(function () {
			model = new VisualStateModel(new Dao());
			model.dao.dao.db.reset();
		});

		it ('pushes an error if not logged in', function (done) {
			model.getSelectedUser(state, function (err, result) {
				should(result).be.not.ok;
				should(model.popAllErrors(state)).be.not.ok;
				err.toString().should.equal('Error: not authorized');
				done();
			});
		});

		it ('pushes an error if not logged in as admin', function (done) {
			state.userId = 1234;
			state.userType = 'card';
			model.getSelectedUser(state, function (err, result) {
				should(result).be.not.ok;
				should(model.popAllErrors(state)).be.not.ok;
				err.toString().should.equal('Error: not authorized');
				done();
			});
		});

		it ('pushes an error if no user selected', function (done) {
			state.userId = 1234;
			state.userType = 'user';
			model.getSelectedUser(state, function (err, result) {
				should(result).be.not.ok;
				should(model.popAllErrors(state)).be.not.ok;
				err.toString().should.equal('Error: user not selected');
				done();
			});
		});

		it ('pushes an error if user record not found', function (done) {
			state.userId = 1234;
			state.userType = 'user';
			state.selectedUser = 3333;
			model.getSelectedUser(state, function (err, result) {
				should(result).be.not.ok;
				should(model.popAllErrors(state)).be.not.ok;
				err.toString().should.equal('Error: user not found for ID');
				done();
			});
		});

		it ('returns record if found', function (done) {
			state.userId = 1234;
			state.userType = 'user';
			var mockNewUser = getMockUser1();
			model.saveNewUser(state, mockNewUser, function (result) {
				state.visualState.should.equal('selectedUser');
				state.selectedUser.should.equal(5001);
				assertBasicDisplayModel(result, null, 1234, 'user');
				model.getSelectedUser(state, function (err, result) {
					should(result).be.ok;
					should(model.popAllErrors(state)).be.not.ok;
					result.id.should.equal(5001);
					result.lastname.should.equal(mockNewUser.lastname);
					result.firstname.should.equal(mockNewUser.firstname);
					result.activeDirId.should.equal(mockNewUser.activeDirId);
					result.effectiveDate.should.equal(mockNewUser.effectiveDate);
					done();
				});
			});
		});

	});

	describe('getSelectedCard()', function () {
		var model = null;
		var state = {};

		describe('error handling', function () {
			beforeEach(function () {
				model = new VisualStateModel(new Dao());
				model.dao.dao.db.reset();
				model.dao.dao.db.cards.push(getMockCard2());
				state = {userId: 1234, userType: 'card', selectedCard: 1111};
			});

			it ('returns error if user not logged in', function (done) {
				state.userType = '???';
				model.getSelectedCard(state, function (err, card) {
					should(card).be.not.ok;
					err.toString().should.equal('Error: user state error');
					should(model.popError(state)).be.not.ok;
					done();
				});
			});

			it ('returns error if card DAO has error', function (done) {
				model.dao.getCurrentCard = function (cardId, callback) {callback(new Error('ouch'));};
				model.getSelectedCard(state, function (err, card) {
					should(card).be.not.ok;
					should(model.popError(state)).be.not.ok;
					err.toString().should.equal('Error: ouch');
					done();
				});
			});

			it ('returns error if card not found', function (done) {
				state.selectedCard = 9999;
				model.getSelectedCard(state, function (err, card) {
					should(card).be.not.ok;
					should(model.popError(state)).be.not.ok;
					err.toString().should.equal('Error: card not found');
					done();
				});
			});

			it ('returns error & card if card has no ownerId', function (done) {
				model.dao.dao.db.cards[0].ownerId = null;
				model.getSelectedCard(state, function (err, card) {
					should(model.popError(state)).be.not.ok;
					err.toString().should.equal('Error: no owner ID');
					card.id.should.equal(1111);
					card.ownerName.should.equal('<no user assigned>');
					done();
				});
			});

			it ('returns error & card if user DAO has error', function (done) {
				model.dao.getCurrentUserById = function (uid, callback) {callback(new Error('OMGosh!'));};
				model.getSelectedCard(state, function (err, card) {
					should(model.popError(state)).be.not.ok;
					err.toString().should.equal('Error: OMGosh!');
					card.id.should.equal(1111);
					card.ownerName.should.equal('<error getting name>');
					done();
				});
			});

			it ('returns error & card if user not found', function (done) {
				model.dao.dao.db.cards[0].ownerId=909891278;
				model.getSelectedCard(state, function (err, card) {
					should(model.popError(state)).be.not.ok;
					err.toString().should.equal('Error: user not found');
					card.id.should.equal(1111);
					card.ownerName.should.equal('<user name not found>');
					done();
				});
			});
		});

		describe("userType='card'", function () {
			beforeEach(function () {
				model = new VisualStateModel(new Dao());
				model.dao.dao.db.reset();
				model.dao.dao.db.cards.push(getMockCard2());
				state = {userId: 1234, userType: 'card', selectedCard: 1111};
			});

			it ('returns card if found and owned by user', function (done) {
				model.getSelectedCard(state, function (err, card) {
					card.id.should.equal(1111);
					card.ownerId.should.equal(1234);
					card.canEdit.should.equal(true);
					card.ownerName.should.equal('fred jones');
					should(model.popError(state)).be.not.ok;
					done();
				});
			});

			it ("returns card if found and not owned by user", function (done) {
				state.userId = 1233;
				model.getSelectedCard(state, function (err, card) {
					card.id.should.equal(1111);
					should(card.canEdit).be.not.ok;
					card.ownerName.should.equal('fred jones');
					should(model.popError(state)).be.not.ok;
					done();
				});
			});

			afterEach(function () {
				model.dao.dao.db.cards = [];
			});
		});

		describe("userType='admin'", function () {
			beforeEach(function () {
				model = new VisualStateModel(new Dao());
				model.dao.dao.db.reset();
				model.dao.dao.db.cards.push(getMockCard2());
				state = {userId: 1234, userType: 'user', selectedCard: 1111};
			});

			it ('returns card if found (owner user)', function (done) {
				model.getSelectedCard(state, function (err, card) {
					card.id.should.equal(1111);
					card.ownerId.should.equal(1234);
					card.ownerName.should.equal('fred jones');
					should(model.popError(state)).be.not.ok;
					done();
				});
			});

			it ("returns card if found (non-owner user)", function (done) {
				state.userId = 1233;
				model.getSelectedCard(state, function (err, card) {
					card.id.should.equal(1111);
					card.ownerId.should.equal(1234);
					card.ownerName.should.equal('fred jones');
					should(model.popError(state)).be.not.ok;
					done();
				});
			});

			afterEach(function () {
				model.dao.dao.db.cards = [];
			});
		});
	});

	describe('assertAuthorization()', function () {
		describe('not logged in', function () {
			var model = new VisualStateModel(new Dao());
			it ('fails for user admin', function () {
				should(model.assertAuthorization('user')).be.not.ok;
			});

			it ('fails for card admin', function () {
				should(model.assertAuthorization('card')).be.not.ok;
			});

			it ('fails for other', function () {
				should(model.assertAuthorization('other')).be.not.ok;
			});
		});

		describe('user admin', function () {
			var model = new VisualStateModel(new Dao());
			var state = {
				userId: 2222,
				userType: 'user'
			}
			it ("'user' auth is okay", function () {
				should(model.assertAuthorization(state, 'user')).be.ok;
			});

			it ("'card' auth is okay", function () {
				should(model.assertAuthorization(state, 'card')).be.ok;
			});

			it ("'other' auth fails", function () {
				should(model.assertAuthorization(state, 'other')).be.not.ok;
			});
		});

		describe('card admin', function () {
			var model = new VisualStateModel(new Dao());
			var state = {
				userId: 2222,
				userType: 'card'
			}
			it ("'user' auth fails", function () {
				should(model.assertAuthorization(state, 'user')).be.not.ok;
			});

			it ("'card' auth is okay", function () {
				should(model.assertAuthorization(state, 'card')).be.ok;
			});

			it ("'other' auth fails", function () {
				should(model.assertAuthorization(state, 'other')).be.not.ok;
			});
		});
	});

	describe('getCurrentCards()', function () {
		it ('fails if not logged in', function (done) {
			var model = new VisualStateModel(new Dao());
			var state = {};
			model.getCurrentCards(state, function (err, results) {
				model.popError(state).should.equal(62);
				done();
			});
		});

		it ("returns all cards for userType='user'", function (done) {
			var model = new VisualStateModel(new Dao());
			var state = {
				userId: 1234,
				userType: 'user'
			};
			model.dao.dao.db.cards = getMockCards4();
			model.getCurrentCards(state, function (err, results) {
				results.length.should.equal(4);
				done();
			});
		});

		it ("returns only the users cards for userType='card'", function (done) {
			var model = new VisualStateModel(new Dao());
			var state = {
				userId: 1234,
				userType: 'card'
			};
			model.getCurrentCards(state, function (err, results) {
				results.length.should.equal(2);
				done();
			});
		});

	});

	describe('getAllCurrentCards()', function () {
		it ('fails if not logged in', function (done) {
			var model = new VisualStateModel(new Dao());
			var state = {};
			model.getAllCurrentCards(state, function (err, results) {
				model.popError(state).should.equal(12);
				done();
			});
		});

		it ("returns all cards for userType='user'", function (done) {
			var model = new VisualStateModel(new Dao());
			var state = {
				userId: 1234,
				userType: 'user'
			};
			model.dao.dao.db.cards = getMockCards4();
			model.getAllCurrentCards(state, function (err, results) {
				results.length.should.equal(4);
				done();
			});
		});

		it ("returns all cards for userType='card'", function (done) {
			var model = new VisualStateModel(new Dao());
			var state = {
				userId: 1234,
				userType: 'card'
			};
			model.getAllCurrentCards(state, function (err, results) {
				results.length.should.equal(4);
				done();
			});
		});

	});

	describe('selectCard()', function () {
		describe('must be logged in as valid user', function () {
			var userTypes1 = getUserTypeArray();

			for (var i=userTypes1.length-1; i>=0; i--) {
				it ('works for admin=' + userTypes1[i] + ' (numeric)', function (done) {
					var model = new VisualStateModel(new Dao());
					var userType = userTypes1.pop();
					var state = {
						userId: 1234,
						userType: userType
					};
					//log.debug('setting user type to ' + userType);
					model.selectCard(state, 1111, function (result) {
						state.visualState.should.equal('selectedCard');
						state.selectedCard.should.equal(1111);
						assertBasicDisplayModel(result, null, 1234, userType);
						done();
					});
				});
			}

			var userTypes2 = getUserTypeArray();

			for (var i=userTypes2.length-1; i>=0; i--) {
				it ('works for admin=' + userTypes2[i] + ' (string)', function (done) {
					var model = new VisualStateModel(new Dao());
					var userType = userTypes2.pop();
					var state = {
						userId: 1234,
						userType: userType
					}
					//log.debug('setting user type to ' + userType);
					model.selectCard(state, '2222', function (result) {
						state.visualState.should.equal('selectedCard');
						state.selectedCard.should.equal(2222);
						assertBasicDisplayModel(result, null, 1234, userType);
						done();
					});
				});
			}

		});
	});

	describe('updateUser()', function () {
		var model = null;
		var state = null;
		var newUser = null;

		beforeEach(function () {
			model = new VisualStateModel(new Dao());
			state = {selectedUser: 4321, userId: 1111, userType: 'user'};
			model.dao.dao.db.reset();
			newUser = {lastname: 'Winchester', firstname: 'Charles', status: 'active', activeDirId: 'vhaisxzzzz1', email: 'cewinchester@abc.defg', userType: 'card'};
		});

		it ('must be authorized', function (done) {
			state.userId = null;
			state.userType = null;
			model.updateUser(state, newUser, function (result) {
				result.should.equal('');
				model.popError(state).should.equal(106);
				done();
			});
		});

		it ('must have a selected user', function (done) {
			state.selectedUser = null;
			model.updateUser(state, newUser, function (result) {
				state.visualState.should.equal('homeUser');
				model.popError(state).should.equal(105);
				done();
			});
		});

		it ('handles error getting old info', function (done) {
			model.dao.getCurrentUserById = function (id, callback) {callback (new Error('bang'));};
			model.updateUser(state, newUser, function (result) {
				state.visualState.should.equal('selectedUser');
				model.popError(state).should.equal(101);
				should(model.popError(state)).be.not.ok;
				assertBasicDisplayModel(result, null, 1111, 'user');
				done();
			});
		});

		it ('handles user not found', function (done) {
			state.selectedUser = 11111111;
			model.updateUser(state, newUser, function (result) {
				state.visualState.should.equal('selectedUser');
				model.popError(state).should.equal(101);
				done();
			});
		});

		it ("handles 'recordAction' error", function (done) {
			model.dao.recordAction = function (info, callback) {callback(new Error('Bang'));};
			model.updateUser(state, newUser, function (result) {
				state.visualState.should.equal('selectedUser');
				model.popError(state).should.equal(103);
				assertBasicDisplayModel(result, null, 1111, 'user');
				done();
			});
		});

		it ('handles dao.updateUser error', function (done) {
			model.dao.updateUser = function (data, callback) {callback(new Error('Bang'));};
			model.updateUser(state, newUser, function (result) {
				state.visualState.should.equal('selectedUser');
				model.popError(state).should.equal(104);
				assertBasicDisplayModel(result, null, 1111, 'user');
				done();
			});
		});

		it ('handles missing firstname', function (done) {
			newUser.firstname = null;
			testUserUpdateErrors(model, state, newUser, 107, done);
		});

		// it ('handles invalid firstname', function (done) {
		// 	newUser.firstname = 'crap*';
		// 	testUserUpdateErrors(model, state, newUser, 107, done);
		// });

		it ('handles missing lastname', function (done) {
			newUser.lastname = null;
			testUserUpdateErrors(model, state, newUser, 108, done);
		});

		// it ('handles invalid lastname', function (done) {
		// 	newUser.lastname = 'cra)p';
		// 	testUserUpdateErrors(model, state, newUser, 108, done);
		// });

		function testUserUpdateErrors (aModel, state, aUser, errorCode, done) {
			aModel.updateUser(state, aUser, function (result) {
				state.visualState.should.equal('selectedUser');
				aModel.popError(state).should.equal(errorCode);
				should(aModel.popError(state)).be.not.ok;
				assertBasicDisplayModel(result, null, 1111, 'user');
				state.selectedUser.should.equal(4321);
				done();
			});
		}

		it ('throws error for an empty activeDirId', function (done) {
			delete newUser.activeDirId;
			testUserUpdateErrors(model, state, newUser, 109, done);
		});

		// it ('throws error for invalid activeDirId', function (done) {
		// 	newUser.activeDirId = 'fred 01';
		// 	testUserUpdateErrors(model, state, newUser, 109, done);
		// });

		it ('throws error if looking up activeDirId fails with an error', function (done) {
			model.dao.getCurrentUserBySysid = function (data, callback) {callback(new Error('ick'));};
			testUserUpdateErrors(model, state, newUser, 131, done);
		});

		it ('throws error if looking up activeDirId returns an item not the current user', function (done) {
			model.dao.getCurrentUserBySysid = function (data, callback) {callback(null, [{id: 38904, active_dir_id: 'vhaisxwinchc', user_type: 'user', firstname: 'charles', lastname: 'winchester', effective_date: new Date('2014-10-10 10:00:00'), status: 'inactive'}]);};
			testUserUpdateErrors(model, state, newUser, 133, done);
		});

		it ('throws error if looking up activityDirId returns more than one user', function (done) {
			model.dao.getCurrentUserBySysid = function (data, callback) {callback(null, [{}, {}]);};
			testUserUpdateErrors(model, state, newUser, 132, done);
		});

		it ('happy path', function (done) {
			model.updateUser(state, newUser, function (result) {
				state.visualState.should.equal('selectedUser');
				should(model.popError(state)).be.not.ok;
				assertBasicDisplayModel(result, null, 1111, 'user');
				model.dao.dao.db.audits.length.should.equal(1);
				var audit = model.dao.dao.db.audits[0];
				audit.userId.should.equal(state.userId);
				audit.itemUserId.should.equal(newUser.id);
				audit.action.should.equal('updateUser');
				audit.effectiveDate.should.equal(newUser.effectiveDate);
				var rec = getUserFromDate(model.dao.dao.db, audit.effectiveDate);
				rec.lastname.should.equal(newUser.lastname);
				rec.firstname.should.equal(newUser.firstname);
				rec.userType.should.equal(newUser.userType);
				rec.email.should.equal(newUser.email);
				rec.activeDirId.should.equal(newUser.activeDirId);
				rec.status.should.equal(newUser.status);
				done();
			});
		});

	});

	describe('updateCard()', function () {
		var model = null;
		var state = null;
		var newCard = null;

		beforeEach(function () {
			model = new VisualStateModel(new Dao());
			state = {selectedCard: 1, userId: 9999, userType: 'user'};
			newCard = getMockCard3();
			model.dao.dao.db.reset();
			model.dao.dao.db.cards = getMockCards4();
		});

		it ('happy path - card owner', function (done) {
			state.userId = 4321;
			state.userType = 'card';
			var now = new Date();
			newCard.replaceFile = true;
			newCard.replaceOwner = true;
			newCard.ownerId = 1029384;
			newCard.title = 'This title is longer than 50 characters blah blah blah blah bluck';
			model.updateCard(state, newCard, function (result) {
				state.visualState.should.equal('selectedCard');
				should(model.popAllErrors(state)).be.not.ok;
				assertBasicDisplayModel(result, null, 4321, 'card');
				model.dao.dao.db.audits.length.should.equal(1);
				var audit = model.dao.dao.db.audits[0];
				audit.userId.should.equal(4321);
				audit.action.should.equal('updateCard');
				audit.itemCardId.should.equal(1);
				audit.effectiveDate.should.equal(newCard.effectiveDate);
				var then = new Date();
				should(isDateWithinDatesInclusive(audit.datetime, now, then)).be.ok;
				var nc = model.dao.dao.db.cards[model.dao.dao.db.cards.length-1];
				nc.title.should.equal('This title is longer than 50 characters blah blah ');
				nc.descr.should.equal('descr');
				nc.version.should.equal('version x');
				nc.versionComment.should.equal('v comment');
				nc.id.should.equal(1);
				nc.ownerId.should.equal(1029384);
				nc.status.should.equal('inactive');
				should(nc.effectiveDate.getTime()>=now.getTime() && nc.effectiveDate.getTime()<=then.getTime()).be.ok;
				JSON.stringify(nc.file).should.equal('{"originalname":"fred.htm"}');
				nc.tags.length.should.equal(3);
				done();
			});
		});

		it ('happy path - user admin', function (done) {
			var now = new Date();
			newCard.replaceFile = true;
			model.updateCard(state, newCard, function (result) {
				state.visualState.should.equal('selectedCard');
				should(model.popAllErrors(state)).be.not.ok;
				assertBasicDisplayModel(result, null, 9999, 'user');
				model.dao.dao.db.audits.length.should.equal(1);
				var audit = model.dao.dao.db.audits[0];
				audit.userId.should.equal(9999);
				audit.action.should.equal('updateCard');
				audit.itemCardId.should.equal(1);
				audit.effectiveDate.should.equal(newCard.effectiveDate);
				var then = new Date();
				should(isDateWithinDatesInclusive(audit.datetime, now, then)).be.ok;
				var nc = model.dao.dao.db.cards[model.dao.dao.db.cards.length-1];
				nc.title.should.equal('title');
				nc.descr.should.equal('descr');
				nc.version.should.equal('version x');
				nc.versionComment.should.equal('v comment');
				nc.id.should.equal(1);
				nc.ownerId.should.equal(4321);
				nc.status.should.equal('inactive');
				should(nc.effectiveDate.getTime()>=now.getTime() && nc.effectiveDate.getTime()<=then.getTime()).be.ok;
				JSON.stringify(nc.file).should.equal('{"originalname":"fred.htm"}');
				nc.tags.length.should.equal(3);
				done();
			});
		});

		it ("changes status to 'inactive' if file info is not there", function (done) {
			newCard.status = 'active';
			newCard.file.originalname = null;
			model.updateCard(state, newCard, function (result) {
				state.visualState.should.equal('selectedCard');
				should(model.popError(state)).be.not.ok;
				assertBasicDisplayModel(result, null, 9999, 'user');
				model.dao.dao.db.cards.length.should.equal(5);
				model.dao.dao.db.audits.length.should.equal(1);
				var ncard = model.dao.dao.db.cards[model.dao.dao.db.cards.length-1];
				ncard.id.should.equal(newCard.id);
				ncard.status.should.equal('inactive');
				ncard.title.should.equal(newCard.title);
				ncard.descr.should.equal(newCard.descr);
				done();
			});
		});

		it ('card admin cannot alter card thats not theirs', function (done) {
			state.userType = 'card';
			model.updateCard(state, newCard, function (result) {
				state.visualState.should.equal('selectedCard');
				model.dao.dao.db.cards.length.should.equal(4);
				model.dao.dao.db.audits.length.should.equal(0);
				model.popError(state).should.equal(113);
				assertBasicDisplayModel(result, null, 9999, 'card');
				done();
			});
		});

		it("adds a file when there wasn't one before", function (done) {
			newCard.replaceFile = true;
			model.updateCard(state, newCard, function (result) {
				state.visualState.should.equal('selectedCard');
				should(model.popError(state)).be.not.ok;
				assertBasicDisplayModel(result, null, 9999, 'user');
				model.dao.dao.db.cards.length.should.equal(5);
				model.dao.dao.db.audits.length.should.equal(1);
				model.getSelectedCard(state, function (err, card) {
					should(err).be.not.ok;
					should(model.popError(state)).be.not.ok;
					card.file.originalname.should.equal('fred.htm');
					done();
				});
			});
		});

		it("replaces a file", function (done) {
			newCard.replaceFile = true;
			setFileOriginalName(model.dao.dao.db.cards);
			model.updateCard(state, newCard, function (result) {
				state.visualState.should.equal('selectedCard');
				should(model.popError(state)).be.not.ok;
				assertBasicDisplayModel(result, null, 9999, 'user');
				model.dao.dao.db.cards.length.should.equal(5);
				model.dao.dao.db.audits.length.should.equal(1);
				model.getSelectedCard(state, function (err, card) {
					should(err).be.not.ok;
					should(model.popError(state)).be.not.ok;
					card.file.originalname.should.equal('fred.htm');
					done();
				});
			});
		});

		it("removes a file", function (done) {
			newCard.replaceFile = true;
			setFileOriginalName(model.dao.dao.db.cards);
			newCard.file = null;
			model.updateCard(state, newCard, function (result) {
				state.visualState.should.equal('selectedCard');
				should(model.popError(state)).be.not.ok;
				assertBasicDisplayModel(result, null, 9999, 'user');
				model.dao.dao.db.cards.length.should.equal(5);
				model.dao.dao.db.audits.length.should.equal(1);
				model.getSelectedCard(state, function (err, card) {
					should(err).be.not.ok;
					should(model.popError(state)).be.not.ok;
					should(card.file).be.not.ok;
					done();
				});
			});
		});

		it("keeps the old file", function (done) {
			newCard.replaceFile = false;
			setFileOriginalName(model.dao.dao.db.cards);
			model.updateCard(state, newCard, function (result) {
				state.visualState.should.equal('selectedCard');
				should(model.popError(state)).be.not.ok;
				assertBasicDisplayModel(result, null, 9999, 'user');
				model.dao.dao.db.cards.length.should.equal(5);
				model.dao.dao.db.audits.length.should.equal(1);
				model.getSelectedCard(state, function (err, card) {
					should(err).be.not.ok;
					should(model.popError(state)).be.not.ok;
					card.file.originalname.should.equal('stuff.htm');
					done();
				});
			});
		});

		it ("throws error if no versionComment and if new status is 'active'", function (done) {
			delete newCard.versionComment;
			newCard.status = 'active';
			setFileOriginalName(model.dao.dao.db.cards);
			model.updateCard(state, newCard, function (result) {
				model.popError(state).should.equal(116);
				state.visualState.should.equal('selectedCard');
				assertBasicDisplayModel(result, null, 9999, 'user');
				done();
			});
		});

		it ("throws error if no versionComment and if old status is 'active'", function (done) {
			delete newCard.versionComment;
			model.updateCard(state, newCard, function (result) {
				model.popError(state).should.equal(117);
				state.visualState.should.equal('selectedCard');
				assertBasicDisplayModel(result, null, 9999, 'user');
				done();
			});
		});

		it ('throws error if old card cant be retrieved', function (done) {
			model.getSelectedCard = function (state, callback) {callback();};
			model.updateCard(state, newCard, function (result) {
				model.popError(state).should.equal(112);
				state.visualState.should.equal('homeUser');
				done();
			});
		});

		it ('throws error if audit cannot be recorded', function (done) {
			model.dao.recordAction = function (action, callback) {callback(new Error('whoops'));};
			model.updateCard(state, newCard, function (result) {
				model.popError(state).should.equal(118);
				state.visualState.should.equal('homeUser');
				done();
			});
		});

		it ('throws error if actual update call fails', function (done) {
			model.dao.updateCard = function (newCard, callback) {callback(new Error('oh oh'));};
			model.updateCard(state, newCard, function (result) {
				model.popError(state).should.equal(119);
				state.visualState.should.equal('homeUser');
				done();
			});
		});

		it ('throws error if replacement owner ID is not an integer', function (done) {
			newCard.replaceOwner = true;
			newCard.ownerId = 'fred';
			model.updateCard(state, newCard, function (result) {
				model.popError(state).should.equal(114);
				state.visualState.should.equal('selectedCard');
				assertBasicDisplayModel(result, null, 9999, 'user');
				done();
			});
		});

		it ('throws error if replacement owner ID is not higher than 0', function (done) {
			newCard.replaceOwner = true;
			newCard.ownerId = -1;
			model.updateCard(state, newCard, function (result) {
				model.popError(state).should.equal(115);
				state.visualState.should.equal('selectedCard');
				assertBasicDisplayModel(result, null, 9999, 'user');
				done();
			});
		});

	});

	describe('authorize()', function () {
		var model = new VisualStateModel(new Dao());

		beforeEach(function () {
			model.dao.dao.db.reset();
		});

		it ('sets up state if not prior authorized', function (done) {
			var state = {};
			var before = new Date();
			var reqUser = getMockRequestUser1();
			model.authorize(state, reqUser, function (err) {
				if (err) throw err;
				var after = new Date();
				state.userId.should.equal(1234);
				state.userType.should.equal('user');
				state.userName.should.equal('fred jones');
				state.activeDirId.should.equal('vhaisxjonesf');
				should(state.lastCheckTimestamp>=before.getTime() && state.lastCheckTimestamp<=after.getTime()).be.ok;

				model.dao.dao.db.audits.length.should.equal(1);
				var audit = model.dao.dao.db.audits[0];
				audit.userId.should.equal(1234);
				audit.action.should.equal('authorize');
				should(audit.datetime.getTime()>=before.getTime() && audit.datetime.getTime()<=after.getTime()).be.ok;
				done();
			});
		});

		it ('does nothing if user is setup and within timeout', function (done) {
			var now = new Date();
			var stateDate = new Date(now.getTime()-(5*60*1000));		//5 minutes
			var state = {
				userId : 1234,
				userType : 'user',
				userName : 'fred jones',
				activeDirId : 'vhaisxjonesf',
				lastCheckTimestamp: stateDate.getTime()
			};
			var reqUser = getMockRequestUser1();
			model.authorize(state, reqUser, function (err) {
				if (err) throw err;
				state.userId.should.equal(1234);
				state.activeDirId.should.equal('vhaisxjonesf');
				state.lastCheckTimestamp.should.equal(stateDate.getTime());
				model.dao.dao.db.audits.length.should.equal(0);
				done();
			});
		});

		it ('reauthorizes if lastCheckTimestamp is overdue', function (done) {
			var before = new Date();
			var state = {
				userId : 1234,
				userType : 'user',
				userName : 'fred jones',
				activeDirId : 'vhaisxjonesf',
				lastCheckTimestamp: before.getTime()-900001
			};
			var reqUser = getMockRequestUser1();
			model.authorize(state, reqUser, function (err) {
				if (err) throw err;
				var after = new Date();
				state.userId.should.equal(1234);
				state.userType.should.equal('user');
				state.userName.should.equal('fred jones');
				state.activeDirId.should.equal('vhaisxjonesf');
				should(state.lastCheckTimestamp>=before.getTime() && state.lastCheckTimestamp<=after.getTime()).be.ok;

				model.dao.dao.db.audits.length.should.equal(1);
				var audit = model.dao.dao.db.audits[0];
				audit.userId.should.equal(1234);
				audit.action.should.equal('authorize');
				should(audit.datetime.getTime()>=before.getTime() && audit.datetime.getTime()<=after.getTime()).be.ok;
				done();
			});
		});

		it ('does nothing in state if user is not found', function (done) {
			var state = {};
			var reqUser = getMockRequestUser2();
			model.authorize(state, reqUser, function (err) {
				if (err) throw err;
				log.debug('authorize() result for not found user:');
				log.debug(state);
				state.visualState.should.equal('');
				should(state.userId).be.not.ok;
				should(state.userType).be.not.ok;
				model.dao.dao.db.audits.length.should.equal(0);
				done();
			});
		});

		it ('does nothing in state if user is not active', function (done) {
			var state = {};
			var reqUser = getMockRequestUser3();
			model.authorize(state, reqUser, function (err) {
				if (err) throw err;
				log.debug('authorize() result for inactive user:');
				log.debug(state);
				state.visualState.should.equal('');
				should(state.userId).be.not.ok;
				should(state.userType).be.not.ok;
				model.dao.dao.db.audits.length.should.equal(0);
				done();
			});
		});

	});

	describe('getUsers()', function () {
		var model = new VisualStateModel(new Dao());

		beforeEach(function () {
			model.dao.dao.db.reset();
		});

		it ('does not allow unauthorized users to call method', function (done) {
			model.getUsers({}, function (err, results) {
				err.should.be.ok;
				should(results).be.not.ok;
				var state = {userId: 112233, userType: 'card'};
				model.getUsers(state, function (err, results) {
					err.should.be.ok;
					should(results).be.not.ok;
					done();
				});
			});

		});

		it ('returns users when authorized user calls method', function (done) {
			var state = {userId:112233, userType: 'user', selectedUser: 332211};
			model.getUsers(state, function (err, results) {
				if (err) throw err;
				results.length.should.equal(5);
				should(state.selectedUser).be.not.ok;
				done();
			});
		});
	});

	describe('setManageUsersState()', function () {
		var model = new VisualStateModel(new Dao());

		it ('allows authorized users to set state (no errors cleared)', function (done) {
			var state = {userId:1122, userType:'user', userName: 'fred', errors: []};
			model.setManageUsersState(state, function (result) {
				state.visualState.should.equal('manageUsers');
				assertBasicDisplayModel(result, state.userName, state.userId, state.userType);
				should(result.errors).be.not.ok;
				done();
			});
		});

		it ('allows authorized users to set state', function (done) {
			var state = {userId:1122, userType:'user', userName: 'fred', errors: [1,2,3]};
			model.setManageUsersState(state, function (result) {
				state.visualState.should.equal('manageUsers');
				assertBasicDisplayModel(result, state.userName, state.userId, state.userType);
				done();
			});
		});

		it('disallows unauthorized users', function (done) {
			var state = {userId:1122, userType:"card", userName: 'red'};
			model.setManageUsersState(state, function (result) {
				state.visualState.should.equal('homeCard');
				assertBasicDisplayModel(result, 'red', 1122, 'card');
				done();
			});
		});
	});

	describe('setAddUserState()', function () {
		var model = new VisualStateModel(new Dao());

		it ('allows authorized users to set state', function (done) {
			var state = {userId:1122, userType:'user'};
			model.setAddUserState(state, function (result) {
				state.visualState.should.equal('addUser');
				assertBasicDisplayModel(result, null, 1122, 'user');
				done();
			});
		});

		it('disallows unauthorized users', function (done) {
			var state = {userId:1122, userType:"card"};
			model.setAddUserState(state, function (result) {
				state.visualState.should.equal('homeCard');
				done();
			});
		});
	});

	describe('setAddCardState()', function () {
		var model = new VisualStateModel(new Dao());

		it ('allows authorized users to set state', function (done) {
			var state = {userId:1122, userType:'user'};
			model.setAddCardState(state, function (result) {
				state.visualState.should.equal('addCard');
				assertBasicDisplayModel(result, null, 1122, 'user');
				state.userType = 'card';
				model.setAddCardState(state, function (result) {
					state.visualState.should.equal('addCard');
					assertBasicDisplayModel(result, null, 1122, 'card');
					done();
				});
			});
		});

		it('disallows unauthorized users', function (done) {
			model.setAddCardState({userId:1122, userType:"fred"}, function (result) {
				result.should.equal('');
				done();
			});
		});
	});

	describe('setEditCardState()', function () {
		var model = new VisualStateModel(new Dao());

		it ('allows authorized users to set state', function (done) {
			var state = {userId:1122, userType:'user', selectedCard: 1122, selectedUser: 332211};
			model.setEditCardState(state, function (result) {
				should(state.selectedUser).be.not.ok;
				state.visualState.should.equal('editCard');
				state.selectedCard.should.equal(1122);
				state.userType = 'card';
				state.selectedUser = 332211;
				assertBasicDisplayModel(result, null, 1122, 'user');
				model.setEditCardState(state, function (result) {
					should(state.selectedUser).be.not.ok;
					state.visualState.should.equal('editCard');
					assertBasicDisplayModel(result, null, 1122, 'card');
					done();
				});
			});
		});

		it('disallows unauthorized users', function (done) {
			model.setEditCardState({userId:1122, userType:"fred", selectedCard:999}, function (result) {
				result.should.equal('');
				done();
			});
		});

		it("won't allow if no card selected", function (done) {
			var state = {userId:123, userType:'card'};
			model.setEditCardState(state, function (result) {
				state.visualState.should.equal('homeCard');
				done();
			});
		});
	});

	describe('setEditUserState()', function () {
		var model = new VisualStateModel(new Dao());

		it ('allows authorized users to set state', function (done) {
			var state = {userId:1122, userType:'user', selectedCard: 1122, selectedUser: 332211};
			model.setEditUserState(state, function (result) {
				should(state.selectedCard).be.not.ok;
				state.selectedUser.should.equal(332211);
				state.visualState.should.equal('editUser');
				assertBasicDisplayModel(result, null, 1122, 'user');
				done();
			});
		});

		it('disallows unauthorized users', function (done) {
			model.setEditUserState({userId:1122, userType:"fred", selectedUser:999}, function (result) {
				result.should.equal('');
				done();
			});
		});

		it("won't allow if no user selected", function (done) {
			var state = {userId:123, userType:'user'};
			model.setEditUserState(state, function (result) {
				state.visualState.should.equal('homeUser');
				done();
			});
		});
	});

	describe('retrieveAndSetupState()', function() {
		var model = null;
		var state = null;

		beforeEach(function () {
			model = new VisualStateModel(new Dao());
			state = {};
			model.dao.dao.db.reset();
		});

		it ('reports user ID errors', function (done) {
			model.retrieveAndSetupState(state, {}, function (result) {
				result.toString().should.equal('Error: no user id');
				model.popError(state).should.equal(121);
				assertUserStateNotSet(state);
				done();
			});
		});

		it ('handles errors', function (done) {
			model.dao.getCurrentUserBySysid = function (uid, callback) {callback(new Error('ouch'));};
			model.retrieveAndSetupState(state, {id:112233}, function (result) {
				result.toString().should.equal('Error: ouch');
				model.popError(state).should.equal(122);
				assertUserStateNotSet(state);
				done();
			});
		});

		it ('handles user not found', function (done) {
			model.retrieveAndSetupState(state, {id:111222333}, function (result) {
				should(result).be.not.ok;
				model.popError(state).should.equal(123);
				assertUserStateNotSet(state);
				done();
			});
		});

		it ('handles too many users found for domain ID', function (done) {
			model.dao.dao.db.admin_users.push({id: 555555, active_dir_id: 'vhaisxjonesf', user_type: 'user', firstname: 'jack', lastname: 'black', effective_date: new Date('2014-04-01 08:00:00'), status: 'active'});
			model.retrieveAndSetupState(state, {id:'vhaisxjonesf'}, function (result) {
				should(result).be.not.ok;
				model.popError(state).should.equal(124);
				assertUserStateNotSet(state);
				done();
			});
		});

		it ('handles inactive user', function (done) {
			model.retrieveAndSetupState(state, {id:'vhaisxwinchc'}, function (result) {
				should(result).be.not.ok;
				model.popError(state).should.equal(125);
				assertUserStateNotSet(state);
				done();
			});
		});

		it ('sets up state for active found user', function (done) {
			var beforeTime = new Date();
			model.retrieveAndSetupState(state, {id:'vhaisxjonesf'}, function (result) {
				should(result).be.not.ok;
				should(model.popError(state)).be.not.ok;
				state.userType.should.equal('user');
				state.userId.should.equal(1234);
				state.userName.should.equal('fred jones');
				state.activeDirId.should.equal('vhaisxjonesf');
				var afterTime = new Date();
				should(beforeTime.getTime() <= state.lastCheckTimestamp && afterTime.getTime() >= state.lastCheckTimestamp).be.ok;
				done();
			});
		});
	});

	describe('getCardFile()', function () {
		var model = null;
		var state = null;

		beforeEach(function () {
			model = new VisualStateModel(new Dao());
			model.dao.dao.db.reset();
			state = {userId: 1, userType: 'card', selectedCard: 1};
		});

		it ('returns error if user not logged in', function (done) {
			state.userType = null;
			model.getCardFile(state, null, function (err, file) {
				err.toString().should.equal('Error: not authorized');
				should(file).be.not.ok;
				should(model.popError(state)).be.not.ok;
				done();
			});
		});

		it ('returns error if no card selected', function (done) {
			state.selectedCard = null;
			model.getCardFile(state, null, function (err, file) {
				err.toString().should.equal('Error: no card selected');
				should(file).be.not.ok;
				should(model.popError(state)).be.not.ok;
				done();
			});
		});

		it ('returns error if dao returns error', function (done) {
			model.dao.getCardFile = function (id, callback) {callback(new Error('whoa!'))};
			model.getCardFile(state, null, function (err, file) {
				should(err).be.ok;
				should(file).be.not.ok;
				should(model.popError(state)).be.not.ok;
				done();
			});
		});

		it ('returns error if empty result (not found)', function (done) {
			state.selectedCard = 11223300;
			model.getCardFile(state, null, function (err, file) {
				err.toString().should.equal('Error: card file not found');
				should(file).be.not.ok;
				should(model.popError(state)).be.not.ok;
				done();
			});
		});

		it ('throws error with non-numeric ID', function (done) {
			state.selectedCard = 'whatever';
			model.getCardFile(state, null, function (err, file) {
				should(file).be.not.ok;
				should(model.popError(state)).be.not.ok;
				should(err).be.ok;
				done();
			});
		});

		it ('finds file from string ID', function (done) {
			state.selectedCard = '1';
			model.dao.dao.db.cards = setupFileNames(getMockCards4());
			model.getCardFile(state, '', function (err, file) {
				if (err) throw err;
				file.originalname.should.equal('something.htm');
				should(model.popError(state)).be.not.ok;
				done();
			});
		});

		it ('finds file from numeric ID', function (done) {
			model.dao.dao.db.cards = setupFileNames(getMockCards4());
			model.getCardFile(state, null, function (err, file) {
				if (err) throw err;
				file.originalname.should.equal('something.htm');
				should(model.popError(state)).be.not.ok;
				done();
			});
		});

		it ('returns error if dao gives error', function (done) {
			model.dao.getCardFileInstance = function (id, instId, callback) {callback(new Error('ouch'))};
			model.getCardFile(state, 'xyzpdq', function (err, result) {
				err.toString().should.equal('Error: ouch');
				should(result).be.not.ok;
				should(model.popError(state)).be.not.ok;
				done();
			});
		});

		it ('returns error if result is empty', function (done) {
			model.getCardFile(state, 'xyzpdq', function (err, result) {
				err.toString().should.equal('Error: card file not found');
				should(result).be.not.ok;
				should(model.popError(state)).be.not.ok;
				done();
			});
		});

		it ('returns instance file', function (done) {
			var aDate = new Date('2014-08-15 01:02:03');
			model.dao.dao.db.cards = setupFileNames(getMockCards4(aDate));
			var instanceId = state.selectedCard.toString() + '_' + aDate.getTime().toString();
			model.getCardFile(state, instanceId, function (err, result) {
				if (err) throw err;
				should(model.popError(state)).be.not.ok;
				result.originalname.should.equal('something.htm');
				done();
			});
		});
	});

	describe('getCardActivity()', function () {
		var model = null;
		var state = null;

		beforeEach(function () {
			model = new VisualStateModel(new Dao());
			model.dao.dao.db.reset();
			state = {userId: 1, userType: 'card', selectedCard: 1};
		});

		it ('returns error if user not logged in', function (done) {
			state.userType = 'fred';
			model.getCardActivity(state, function (err, result) {
				err.should.equal(131);
				done();
			});
		});

		it ('returns error if card not selected', function (done) {
			state.selectedCard = null;
			model.getCardActivity(state, function (err, result) {
				err.should.equal(132);
				done();
			});
		});

		it ('happy path', function (done) {
			var now = new Date();
			var date1 = new Date('2014-09-01 00:00:00')
			var cards1 = getMockCards4(date1);

			var audits = [];
			for (var i = cards1.length - 1; i >= 0; i--) {
				var audit = {
					datetime: cards1[i].effectiveDate,
					effectiveDate: cards1[i].effectiveDate,
					userId: 5000,
					action: 'saveNewCard',
					itemCardId: cards1[i].id
				}
				audits.push(audit);
			};

			var date2 = new Date('2014-10-01 00:00:00');
			var cards2 = getMockCards4(date2);
			for (var i = cards2.length - 1; i >= 0; i--) {
				audits.push({
					datetime: cards2[i].effectiveDate,
					effectiveDate: cards2[i].effectiveDate,
					userId: 1234,
					action: 'updateCard',
					itemCardId: cards2[i].id
				});
			};

			var date3 = new Date('2014-11-01 00:00:00');
			var cards3 = getMockCards4(date3);

			for (var i = cards3.length - 1; i >= 0; i--) {
				audits.push({
					datetime: cards3[i].effectiveDate,
					effectiveDate: cards3[i].effectiveDate,
					userId:4321,
					action: 'updateCard',
					itemCardId: cards3[i].id
				});
			};

			model.dao.dao.db.cards = cards2.concat(cards1).concat(cards3);
			model.dao.dao.db.audits = audits;

			model.getCardActivity(state, function (err, result) {
				if (err) throw err;
				log.debug('+++card activity+++');
				log.debug(result[0]);
				result.length.should.equal(3);
				result[0].id.should.equal(1);
				result[0].title.should.equal('A Pocket Card 2');
				result[0].ownerId.should.equal(4321);
				result[0].ownerName.should.equal('charles winchester');
				result[0].editorId.should.equal(4321);
				result[0].editorName.should.equal('charles winchester');
				result[0].effectiveDate.getTime().should.equal(date3.getTime());
				result[0].instanceId.should.equal(result[0].effectiveDate.getTime().toString());
				should(result[0].nextEffectiveDate.getTime()>=now.getTime()).be.ok;

				result[1].id.should.equal(1);
				result[1].title.should.equal('A Pocket Card 2');
				result[1].ownerId.should.equal(4321);
				result[1].ownerName.should.equal('charles winchester');
				result[1].editorId.should.equal(1234);
				result[1].editorName.should.equal('fred jones');
				result[1].effectiveDate.getTime().should.equal(date2.getTime());
				result[1].instanceId.should.equal(result[1].effectiveDate.getTime().toString());
				result[1].nextEffectiveDate.should.equal(result[0].effectiveDate);

				result[2].id.should.equal(1);
				result[2].title.should.equal('A Pocket Card 2');
				result[2].ownerId.should.equal(4321);
				result[2].ownerName.should.equal('charles winchester');
				result[2].editorId.should.equal(5000);
				result[2].editorName.should.equal('weird user');
				result[2].effectiveDate.getTime().should.equal(date1.getTime());
				result[2].instanceId.should.equal(result[2].effectiveDate.getTime().toString());
				result[2].nextEffectiveDate.should.equal(result[1].effectiveDate);
				done();
			});
		});
	});

	describe('getCardVersions()', function () {
		var model = null;
		var state = null;

		beforeEach(function () {
			model = new VisualStateModel(new Dao());
			model.dao.dao.db.reset();
			state = {userId: 1, userType: 'card', selectedCard: 1};
		});

		it ('returns error if not logged in', function (done) {
			delete state.userType;
			model.getCardVersions(state, function (err, results) {
				err.should.equal(141);
				should(results).be.not.ok;
				done();
			});
		});

		it ('returns error if no card selected', function (done) {
			delete state.selectedCard;
			model.getCardVersions(state, function (err, results) {
				err.should.equal(142);
				should(results).be.not.ok;
				done();
			});
		});

		it ('returns error if data call has error', function (done) {
			model.dao.getCardHistory = function (cardId, callback) {callback(new Error('stop it!'));};
			model.getCardVersions(state, function (err, results) {
				err.toString().should.equal('Error: stop it!');
				should(results).be.not.ok;
				done();
			});
		});

		it ('returns data if card exists', function (done) {
			model.dao.dao.db.cards = getMockCards4();

			model.getCardVersions(state, function (err, results) {
				should(err).be.not.ok;
				results.length.should.equal(1);
				results[0].should.equal('1.0.0');
				done();
			});
		});
	});

	describe('setAboutState()', function () {
		it ('allows state if logged in', function (done) {
			var model = new VisualStateModel(new Dao());
			var state = {userId:1, userType: 'fred'};
			model.setAboutState(state, function (result) {
				state.visualState.should.equal('about');
				assertBasicDisplayModel(result, null, 1, 'fred');
				done();
			});
		});
	});

	describe('setHelpState()', function () {
		it ('allows state if logged in', function (done) {
			var model = new VisualStateModel(new Dao());
			var state = {userId:1, userType: 'fred'};
			model.setHelpState(state, function (result) {
				state.visualState.should.equal('help');
				assertBasicDisplayModel(result, null, 1, 'fred');
				done();
			});
		});
	});

	describe('setManageCardsState()', function () {
		var model = null;
		var state = null;

		beforeEach(function () {
			model = new VisualStateModel(new Dao());
			state = {userId: 1, errors: [], userName:'todd'};
		});

		it ('allows user managegers', function (done) {
			state.userType = 'user';
			model.setManageCardsState(state, function (result) {
				state.visualState.should.equal('manageCards');
				should(model.popError(state)).be.not.ok;
				should(result.errors).be.not.ok;
				assertBasicDisplayModel2(result, state);
				done();
			});
		});

		it ('allows user managegers', function (done) {
			state.userType = 'card';
			model.setManageCardsState(state, function (result) {
				state.visualState.should.equal('manageCards');
				should(model.popError(state)).be.not.ok;
				should(result.errors).be.not.ok;
				assertBasicDisplayModel2(result, state);
				done();
			});
		});

	});

	describe ('getAllCurrentUserNames()', function () {
		var model = null;
		var state = null;

		beforeEach(function () {
			model = new VisualStateModel(new Dao());
			state = {userId:1, userType:'user'};
		});

		it ('fails if not user admin', function (done) {
			state.userType = 'card';
			model.getAllCurrentUserNames(state, function (err, users) {
				should(users).be.not.ok;
				err.should.equal(151);
				done();
			});
		});

		it ('happy path', function (done) {
			model.getAllCurrentUserNames(state, function (err, users) {
				users.length.should.equal(5);
				should(err).be.not.ok;
				for (var i = users.length - 1; i >= 0; i--) {
					Object.getOwnPropertyNames(users[i]).length.should.equal(3);
				};
				users.sort(function (i1, i2) {
					return i1.id-i2.id;
				});
				//TODO: finish this
				users[0].id.should.equal(888);
				users[0].lastname.should.equal('jones');
				users[0].firstname.should.equal('robert');

				users[1].id.should.equal(889);
				users[1].lastname.should.equal('jones');
				users[1].firstname.should.equal('richard');

				users[2].id.should.equal(1234);
				users[2].lastname.should.equal('jones');
				users[2].firstname.should.equal('fred');

				users[3].id.should.equal(4321);
				users[3].lastname.should.equal('winchester');
				users[3].firstname.should.equal('charles');

				users[4].id.should.equal(5000);
				users[4].lastname.should.equal('user');
				users[4].firstname.should.equal('weird');
				done();
			});
		});
	});

	describe ('setCardsMetricsDownloadState()', function () {
		var model = null;
		var state = null;

		beforeEach(function () {
			model = new VisualStateModel(new Dao());
			state = {userId:1, userType:'user', selectedCard: 2};
		});

		it('happy path', function (done) {
			model.setCardsMetricsDownloadState(state, function (result) {
				state.visualState.should.equal('cardsmetrics');
				should(state.selectedCard).be.not.ok;
				should(model.popAllErrors(state)).be.not.ok;
				done();
			});
		});

		it ('fails if user not logged in', function (done) {
			state.userType = 'fred';
			model.setCardsMetricsDownloadState(state, function (result) {
				state.visualState.should.equal('');
				var errors = model.popAllErrors(state);
				errors.length.should.equal(1);
				done();
			});
		});
	});

	describe ('getMetaMetricsIterator()', function () {
		var model = null;
		var state = null;

		beforeEach(function () {
			model = new VisualStateModel(new Dao());
			state = {userId:1, userType:'card'};
			model.dao.dao.db.reset();
		});

		it ('returns a valid iterator', function (done) {
			setupMetrics(model.dao.dao.db.metrics);
			model.getMetaMetricsIterator(state, new Date(2015, 3, 1, 0, 0, 0), new Date(2015, 3, 1, 23, 59, 59), function (err, iter) {
				var list = [];
				iter.getNext(function (err, item) {
					list.push(item);
					iter.getNext(function (err, item) {
						list.push(item);
						iter.getNext(function (err, item) {
							list.push(item);
							iter.getNext(function (err, item) {
								should(item).be.not.ok;
								iter.done();
								list.sort(function (i1, i2) {
									return i1.timestamp.getTime()-i2.timestamp.getTime();
								});
								list.length.should.equal(3);
								list[0].request.should.equal('/pocketcardv1/user/abcdefghijk/cards');
								list[1].request.should.equal('/pocketcardv1/user/abcdefghijk/cards?state=active');
								list[2].request.should.equal('/pocketcardv1/user/abcdefghijk/cards?state=inactive');
								done();
							});
						});
					});
				});
			});
		});
	});

	describe('getCardMetricsIterator()', function () {
		var model = null;
		var state = null;

		beforeEach(function () {
			model = new VisualStateModel(new Dao());
			state = {userId:1, userType:'card'};
			model.dao.dao.db.reset();
			setupMetrics(model.dao.dao.db.metrics);
		});

		it ('returns a working iterator when card is selected (no dates)', function (done) {
			state.selectedCard = '10101';
			model.getCardMetricsIterator(state, null, null, function (err, iter) {
				iter.getNext(function (err, item) {
					item.request.should.equal('/pocketcardv1/user/abcxyz112233/cards/10101/contents');
					iter.getNext(function (err, item2) {
						item2.request.should.equal('/pocketcardv1/user/abcxyz112233/cards/10101/contents');
						iter.getNext(function (err, item3) {
							if (err) throw err;
							should(item3).be.not.ok;
							iter.done();
							done();
						});
					});
				});
			});
		});

		it ('returns a working iterator when card is selected (with dates #1)', function (done) {
			state.selectedCard = '10101';
			model.getCardMetricsIterator(state, new Date('1-1-2015'), new Date('12-31-2015'), function (err, iter) {
				iter.getNext(function (err, item) {
					item.request.should.equal('/pocketcardv1/user/abcxyz112233/cards/10101/contents');
					iter.getNext(function (err, item2) {
						item.request.should.equal('/pocketcardv1/user/abcxyz112233/cards/10101/contents');
						iter.getNext(function (err, item3) {
							if (err) throw err;
							should(item3).be.not.ok;
							iter.done();
							done();
						});
					});
				});
			});
		});

		it ('returns a working iterator when card is selected (with dates #2)', function (done) {
			state.selectedCard = '10101';
			model.getCardMetricsIterator(state, new Date('5-15-2015'), new Date('12-31-2015'), function (err, iter) {
				iter.getNext(function (err, item) {
					item.request.should.equal('/pocketcardv1/user/abcxyz112233/cards/10101/contents');
					item.timestamp.getMonth().should.equal(5);
					iter.getNext(function (err, item2) {
						if (err) throw err;
						should(item2).be.not.ok;
						iter.done();
						done();
					});
				});
			});
		});

		it ('returns a working iterator when card is not selected (no dates)', function (done) {
			model.getCardMetricsIterator(state, null, null, function (err, iter) {
				var items = [];
				iter.getNext(function (err, item) {
					items.push(item);
					iter.getNext(function (err, item2) {
						items.push(item2);
						iter.getNext(function (err, item3) {
							items.push(item3);
							iter.getNext(function (err, item4) {
								items.push(item4);
								items.length.should.equal(4);
								iter.done();
								items[0].timestamp.getMonth().should.equal(3);
								items[1].timestamp.getMonth().should.equal(4);
								items[2].timestamp.getMonth().should.equal(5);
								items[3].timestamp.getMonth().should.equal(6);
								done();
							});
						});
					});
				});
			});
		});

		it ('returns a working iterator when card is not selected (with dates)', function (done) {
			model.getCardMetricsIterator(state, new Date('4-15-2015'), new Date('6-15-2015'), function (err, iter) {
				iter.getNext(function (err, item) {
					item.timestamp.getMonth().should.equal(4);
					iter.getNext(function (err, item2) {
						item2.timestamp.getMonth().should.equal(5);
						iter.getNext(function (err, item3) {
							should(item3).be.not.ok;
							iter.done();
							done();
						});
					});
				});
			});
		});

		it ('returns an error if wrong user type logged in', function (done) {
			state.userType = 'fred';
			state.selectedCard = 10101;
			model.getCardMetricsIterator(state, null, null, function (err, iter) {
				err.toString().should.equal('171');
				should(iter).be.not.ok;
				done();
			});
		});
	});

	describe('setCardMetricsDownloadState()', function () {
		var model = null;
		var state = null;

		beforeEach(function () {
			model = new VisualStateModel(new Dao());
			state = {userId:1, userType:'card'};
			model.dao.dao.db.reset();
		});

		it ('handles unauthorized users', function (done) {
			state.userType = 'fred';
			model.setCardMetricsDownloadState(state, function (result) {
				state.visualState.should.equal('');
				var errors = model.popAllErrors(state);
				errors.length.should.equal(1);
				done();
			});
		});

		it ('returns expected state when no card selected', function (done) {
			model.setCardMetricsDownloadState(state, function (result) {
				state.visualState.should.equal('cardmetrics');
				should(result.card).be.not.ok;
				done();
			});
		});

		it ('returns error if card not found', function (done) {
			state.selectedCard = 11001100;
			model.setCardMetricsDownloadState(state, function (result) {
				state.visualState.should.equal('homeCard');
				should(result.card).be.not.ok;
				model.popError(state).should.equal(181);
				done();
			});
		});

		it ('returns card if selected', function (done) {
			state.selectedCard = 11001100;
			var mockCard1 = getMockCard1(new Date('1/1/2015'));
			mockCard1.id = state.selectedCard;
			model.dao.dao.db.cards.push(mockCard1);
			model.setCardMetricsDownloadState(state, function (result) {
				state.visualState.should.equal('cardmetrics');
				result.card.id.should.equal(state.selectedCard);
				done();
			});
		});
	});

	describe ('getSelectedUserModel()', function () {
		var model = null;
		var state = null;

		beforeEach(function () {
			model = new VisualStateModel(new Dao());
			state = {userId:1, userType:'user', selectedUser: 123};
			model.dao.dao.db.reset();
		});

		it ('pushes error if user not logged in', function (done) {
			state.userId = null;
			model.getSelectedUserModel(state, function (result) {
				state.visualState.should.equal('');
				should(result).be.not.ok;
				model.popError(state).should.equal(81);
				should(model.popError(state)).be.not.ok;
				done();
			});
		});

		it ('pushes error if user not selected', function (done) {
			state.selectedUser = null;
			model.getSelectedUserModel(state, function (result) {
				state.visualState.should.equal('homeUser');
				model.popError(state).should.equal(82);
				should(model.popError(state)).be.not.ok;
				assertBasicDisplayModel(result, null, 1, 'user');
				done();
			});
		});

		it ('pushes error if there is an error getting selected user', function (done) {
			model.getSelectedUser = function (state, callback) {callback(new Error(''));};
			model.getSelectedUserModel(state, function (result) {
				state.visualState.should.equal('userSelected');
				model.popError(state).should.equal(83);
				should(model.popError(state)).be.not.ok;
				assertBasicDisplayModel(result, null, 1, 'user');
				should(result.user).be.not.ok;
				done();
			});
		});

		it ('pushes error if user not found', function (done) {
			model.getSelectedUser = function (state, callback) {callback();};
			model.getSelectedUserModel(state, function (result) {
				state.visualState.should.equal('userSelected');
				model.popError(state).should.equal(84);
				should(model.popError(state)).be.not.ok;
				assertBasicDisplayModel(result, null, 1, 'user');
				should(result.user).be.not.ok;
				done();
			});
		});

		it ('happy path', function (done) {
			model.getSelectedUser = function (state, callback) {callback(null, {naaame: 'fred'});};
			model.getSelectedUserModel(state, function (result) {
				state.visualState.should.equal('userSelected');
				should(model.popError(state)).be.not.ok;
				assertBasicDisplayModel(result, null, 1, 'user');
				result.user.naaame.should.equal('fred');
				done();
			});
		});
	});

	describe ('getSelectedCardModel()', function () {
		var state = null;
		var model = null;

		beforeEach(function () {
			model = new VisualStateModel(new Dao());
			state = {userId: 12, userType: 'card', selectedCard: 21};
			model.dao.dao.db.reset();
		});

		it ('pushes an error if the user not logged in', function (done) {
			state.userId = null;
			model.getSelectedCardModel(state, function (result) {
				state.visualState.should.equal('');
				should(result).be.not.ok;
				model.popError(state).should.equal(91);
				should(model.popError(state)).be.not.ok;
				done();
			});
		});

		it ('pushed an error if card not selected', function (done) {
			state.selectedCard = null;
			model.getSelectedCardModel(state, function (result) {
				state.visualState.should.equal('homeCard');
				model.popError(state).should.equal(92);
				should(model.popError(state)).be.not.ok;
				assertBasicDisplayModel(result, null, 12, 'card');
				done();
			});
		});

		it ('pushes 2 errors and saves null card if there is no card returned with an error', function (done) {
			model.getSelectedCard = function (state, callback) {callback(new Error('oops'));};
			model.getSelectedCardModel(state, function (result) {
				state.visualState.should.equal('homeCard');
				model.popError(state).should.equal(94);
				model.popError(state).should.equal(93);
				should(model.popError(state)).be.not.ok;
				assertBasicDisplayModel(result, null, 12, 'card');
				should(result.card).be.not.ok;
				done();
			});
		});

		it ('pushes error if card not found', function (done) {
			model.getSelectedCard = function (state, callback) {callback();};
			model.getSelectedCardModel(state, function (result) {
				state.visualState.should.equal('homeCard');
				model.popError(state).should.equal(94);
				should(model.popError(state)).be.not.ok;
				assertBasicDisplayModel(result, null, 12, 'card');
				should(result.card).be.not.ok;
				done();
			});
		});

		it ('pushes an error and saves card if card call returns an error and card', function (done) {
			model.getSelectedCard = function (state, callback) {callback(new Error('oops'), {id:21, title: 'howdy'});};
			model.getSelectedCardModel(state, function (result) {
				state.visualState.should.equal('selectedCard');
				model.popError(state).should.equal(93);
				should(model.popError(state)).be.not.ok;
				assertBasicDisplayModel(result, null, 12, 'card');
				result.card.title.should.equal('howdy');
				done();
			});
		});

		it ('happy path', function (done) {
			model.getSelectedCard = function (state, callback) {callback(null, {id:21, title: 'howdy'});};
			model.getSelectedCardModel(state, function (result) {
				state.visualState.should.equal('selectedCard');
				should(model.popError(state)).be.not.ok;
				assertBasicDisplayModel(result, null, 12, 'card');
				result.card.title.should.equal('howdy');
				done();
			});
		});

	});

	describe('continueOkay()', function () {
		var state = null;
		var model = null;

		beforeEach(function () {
			model = new VisualStateModel(new Dao());
			state = {userId: 12, userType: 'card'};
			model.dao.dao.db.reset();
		});

		it ('returns ok for happy path', function (done) {
			model.continueOkay(state, function (result) {
				result.should.equal('ok');
				done();
			});
		});

		it ('returns error if user not logged in or wrong type', function (done) {
			state.userType = 'fred';
			model.continueOkay(state, function (result) {
				result.should.equal('error');
				done();
			});
		});
	});

	describe('isSystemIdUsed()', function () {
		var state = null;
		var model = null;

		beforeEach(function () {
			model = new VisualStateModel(new Dao());
			state = {userId: 12, userType: 'user'};
			model.dao.dao.db.reset();
		});

		it ('returns false if user not selected and not found', function (done) {
			systemIdUsedTestX(model, state, 'vhaisxjonesj', false, done);
		});

		it ('returns true if user not selected and found', function (done) {
			systemIdUsedTestX(model, state, 'vhaisxwinchc', true, done);
		});

		it ('returns false if user is selected and is found', function (done) {
			state.selectedUser = 4321;
			systemIdUsedTestX(model, state, 'vhaisxwinchc', false, done);
		});

		it ('returns true if user is selected and found and there is another user with the same ID', function (done) {
			state.selectedUser = 888;
			systemIdUsedTestX(model, state, 'vhaisxjonesr', true, done);
		});

		function systemIdUsedTestX (aModel, state, id, value, done) {
			aModel.isSystemIdUsed(state, id, function (err, result) {
				if (err) throw err;
				result.should.equal(value);
				done();
			});
		}
	});
});

function assertUserStateNotSet(state) {
	should(state).be.ok;
	should(state.userType).be.not.ok;
	should(state.userId).be.not.ok;
	should(state.userName).be.not.ok;
	should(state.activeDirId).be.not.ok;
	should(state.lastCheckTimestamp).be.not.ok;
}

function getUserFromDate(db, aDate) {
	for (var i = db.admin_users.length - 1; i >= 0; i--) {
		if (aDate === db.admin_users[i].effective_date) {
			return daoUtils.getNormalUserEntry(db.admin_users[i]);
		}
	};
	return null;
}

function setupFileNames(ary) {
	if (ary && ary.length) {
		for (var i = ary.length - 1; i >= 0; i--) {
			if (!ary[i].file) {
				ary[i].file = {};
			}
			ary[i].file.originalname = 'something.htm';
		};
	}
	return ary;
}

function getMockCards4(aDate) {
	var cards = [];
	cards.push(getMockCard2(aDate));
	cards.push(getMockCard2(aDate));
	cards.push(getMockCard2(aDate));
	cards.push(getMockCard2(aDate));
	cards[0].ownerId = 4321;
	cards[3].ownerId = 4321;
	cards[0].id = 1;
	cards[1].id = 2;
	cards[2].id = 3;
	cards[3].id = 4;
	return cards;
}

function getMockUser1() {
	return {userType: 'user', lastname: 'Jones', firstname: 'Fred', activeDirId: 'VHAISZXXX', effectiveDate: new Date('2014-12-01 00:00:00')};
}

function getMockCard1(effDate) {
	if (!effDate) {
		effDate = new Date();
	}
	var card = {
		ownerId: 44, 
		title:'A Pocket Card 1', 
		descr: 'Dummy card 1', 
		tags: ['dummy', 'test', '1'], 
		version: '1.0.0',
		versionComment: 'initial 1',
		status:'active', 
		effectiveDate: effDate,
		file: {fieldname: 'file1',
			originalname: 'test.html',
			name: 'test1419363434354.html',
			encoding: '7bit',
			mimetype: 'text/html',
			path: 'uploads/test1419363434354.html',
			extension: 'html',
			size: 8996,
			truncated: false,
			buffer: new Buffer('<html><body>This is a test</body></html>')}
	};
	return card;
}

function getMockCard2(aDate) {
	if (!aDate) {
		aDate = new Date('2013-01-01 00:00:00');
	}
	return {
		id: 1111,
		ownerId: 1234, 
		title:'A Pocket Card 2', 
		descr: 'Dummy card 2', 
		tags: ['dummy', 'test', '2'], 
		version: '1.0.0',
		versionComment: 'initial 2',
		status:'active', 
		effectiveDate: aDate,
		file: {}
	};
}

function getMockCard3() {
	return {
		title: 'title',
		descr: 'descr',
		version: 'version x',
		versionComment: 'v comment',
		tags: ['1', '2', '3'],
		file: {originalname: 'fred.htm'}
	};
}

function getUserTypeArray() {
	return ['card', 'user'];
}

function getMockRequestUser1() {
	return {id: 'VHAISXJONESF', name: 'blahblahblah', email: 'blah@blah.com'};
}

function getMockRequestUser2() {
	return {id: 'vhaisxwhatEVA', name: 'yuckyuckyuck', email: 'yuck@yuck.com'};
}

function getMockRequestUser3() {
	return {id: 'VHAisxwinchc', name: 'charles winchester', email: 'chuck@va.gov'};
}

function isDateWithinDatesInclusive(aDate, lowerDate, upperDate) {
	assertNotNull(aDate, 'ref date');
	assertNotNull(lowerDate, 'lower date');
	assertNotNull(upperDate, 'upper date');
	var result = (aDate.getTime() >= lowerDate.getTime() && aDate.getTime() <= upperDate.getTime());
	return result;
}

function assertNotNull(item, name) {
	if (!item) {
		if (name) {
			throw new Error(name + ' was blank');
		}
		else {
			throw new Error ('item was blank');
		}
	}
}

function setupInstanceIds(ary) {
	if (ary && ary.length) {
		for (var i = ary.length - 1; i >= 0; i--) {
			ary[i].instanceId = ary[i].id.toString() + '_' + ary[i].effectiveDate.getTime().toString();
		};
	}
	return ary;
}

function setFileOriginalName(ary) {
	if (ary && ary.length) {
		for (var i = ary.length - 1; i >= 0; i--) {
			ary[i].file = {originalname:'stuff.htm'};
		};
	}
}

function setupMetrics(list) {
	list.push({deviceHash : "abcdefghijk", clientIp : "127.0.0.1", request : "/pocketcardv1/user/abcdefghijk/cards", timestamp : new Date("2015-04-01 15:58:45")});
	list.push({deviceHash : "abcdefghijk", clientIp : "127.0.0.1", request : "/pocketcardv1/user/abcdefghijk/cards?state=active", timestamp : new Date("2015-04-01 15:58:46")});
	list.push({deviceHash : "abcdefghijk", clientIp : "127.0.0.1", request : "/pocketcardv1/user/abcdefghijk/cards?state=inactive", timestamp : new Date("2015-04-01 15:58:47")});
	list.push({deviceHash : "abcxyz112233", clientIp : "127.0.0.1", request : "/pocketcardv1/user/abcxyz112233/cards/10101", timestamp : new Date("2015-04-01 15:58:48")});
	list.push({deviceHash : "abcxyz112233", clientIp : "127.0.0.1", request : "/pocketcardv1/user/abcxyz112233/cards/10101/contents", timestamp : new Date("2015-04-01 15:58:49")});
	list.push({deviceHash : "abcxyz112233", clientIp : "127.0.0.1", request : "/pocketcardv1/user/abcxyz112233/cards/10102/contents", timestamp : new Date("2015-05-01 15:58:50")});
	list.push({deviceHash : "abcxyz112233", clientIp : "127.0.0.1", request : "/pocketcardv1/user/abcxyz112233/cards/10101/contents", timestamp : new Date("2015-06-01 15:58:49")});
	list.push({deviceHash : "abcxyz112233", clientIp : "127.0.0.1", request : "/pocketcardv1/user/abcxyz112233/cards/10102/contents", timestamp : new Date("2015-07-01 15:58:50")});
}

function assertBasicDisplayModel(dModel, name, userId, userType) {
	should(dModel).be.ok;
	if (name) {
		dModel.userName.should.equal(name);
	}
	if (userId) {
		dModel.userId.should.equal(userId);
	}
	if (userType) {
		dModel.userType.should.equal(userType);
	}
}

function assertBasicDisplayModel2(dModel, other) {
	dModel.userName.should.equal(other.userName);
	dModel.userId.should.equal(other.userId);
	dModel.userType.should.equal(other.userType);
}