//setupMongo.js
var log4js = require('log4js');
var mclient = require('mongodb').MongoClient;
var _ = require('underscore');
var log = log4js.getLogger('setupMongo.js');

var FIRST_SITE_ID = 99;
var userIdsToAdmins = [4];

var asyncFunctions = [
	setupCards,
	setupUsers,
	setupIds,
	setupSites
];
var db;

log.info("started Mongo DB setup");

var dbData = promptForDbData();

function asyncFinisher(err) {
	if (err) {
		if (db) {
			db.close();
		}
		throw err;
	}
	if (asyncFunctions.length) {
		var func = asyncFunctions.pop();
		func(db, asyncFinisher);
	}
	else {
		db.close();
		log.info("completed successfully");
	}
}

mclient.connect(dbData.getConnectionString(), function (err, database) {
	db = database;
	asyncFinisher();
});

function promptForDbData() {
	return {
		getConnectionString: function () {
			return "mongodb://localhost:27017/pc2";
		}
	};
}

function setupIds(db, callback) {
	log.info("setting up IDs");
	
	db.collection('ids').find({}).toArray(function (err, ids) {
		if (err) return callback(err);
		
		//log.debug('ids length==' + ids.length);
		//new install; give it everything
		if (!ids || !ids.length) {
			db.collection('ids').insert(getNewIds(), function (err, results) {
				return callback(err);
			});
		}
		//v1 install; add the new stuff
		else if (ids.length<3) {
			db.collection('ids').insert(getExtraIds(), function (err, results) {
				return callback(err);
			});
		}
		//this is not new and not v1; this is an *error*
		else {
			callback(new Error("not a new or v1 database; can't continue"));
		}
	});
}

function setupUsers(db, callback) {
	log.info("setting up users");
	
	db.collection('adminUsers').count({}, function (err, count) {
		if (err) return callback(err);

		if (count>0) {
			modifyUsers(db, callback);
		}
		else {
			db.collection("adminUsers").insert(getNewUsers(), function (err, result) {
				callback(err);
			});
		}
	});
}

function modifyUsers(db, callback) {
	var adminIdMap = {};
	userIdsToAdmins.forEach(function (id) {
		adminIdMap[id] = 1;
	});
	(function updateNextUser() {
		db.collection("adminUsers").findOne({roles:null}, function (err, user) {
			if (err) return callback(err);
			if (user) {
				var utype = user.userType;
				delete user.userType;
				var role = {};
				if (adminIdMap[user.id]) {
					role.name = "app";
				}
				else {
					role.siteId = FIRST_SITE_ID;
					if ("card"==utype) {
						role.name = "card";
					}
					else if ("user"==utype) {
						role.name = "site";
					}
					else {
						role.name = "unknown";
					}
				}
				user.roles = [role];
				db.collection("adminUsers").findOneAndReplace({_id: user._id}, user, {upsert: false}, function (err) {
					if (err) return callback(err);
					updateNextUser();
				});
			}
			else {
				callback();
			}
		});
	})();
}

function setupCards(db, callback) {
	log.info("setting up cards");
	
	(function updateNextCard() {
		db.collection("cards").findOne({siteId: null}, function (err, card) {
			if (err) return callback(err);
			if (card) {
				card.siteId = FIRST_SITE_ID;
				db.collection("audits").find({itemCardId: card.id, effectiveDate: card.effectiveDate}).toArray(function (err, audits) {
					if (err) return callback(err);
					if (audits && audits.length) {
						if (audits.length>1) {
							card.lastEditUserId = -2;
						}
						else {
							card.lastEditUserId = audits[0].userId;
						}
					}
					else {
						card.lastEditUserId = -1;
					}
					db.collection("cards").findOneAndReplace({_id: card._id}, card, {upsert: false}, function (err) {
						if (err) return callback(err);
						updateNextCard();
					});
				});
			}
			else {
				callback();
			}
		});
	})();
}

function convertCards(cards, audits) {
	cards.forEach(function (card) {
		card.siteId = FIRST_SITE_ID;
		delete card.ownerId;
		card.lastEditUserId = getLastEditUserId(card, audits);
	});
}

function getLastEditUserId(card, audits) {
	var cardId = card.id;
	var effDtTime = card.effectiveDate.getTime();
	var cAudits = _.filter(audits, function (audit) {
		return cardId == audit.itemCardId && effDtTime == audit.effectiveDate.getTime();
	});
	if (cAudits && cAudits.length) {
		if (cAudits.length>1) {
			return -2;
		}
		return cAudits[0].userId;
	}
	return -1;
}

function setupSites(db, callback) {
	log.info("setting up sites");
	
	db.collection("sites").find({}).toArray(function (err, sites) {
		if (err || !sites || !sites.length) {
			//Good; there are no sites yet. Setup one initial site
			db.collection("sites").insert(getNewSites(), function (err, results) {
				callback(err);
			});
		}
		else {
			//We have *sites*?!? Bad bad bad.
			callback(new Error("Sites already exist? Wrong database; can't continue"));
		}
	});
}

function getNewIds() {
	var ids = [
		{name: "card_id", value: 100},
		{name: "user_id", value: 100}
	];
	var extras = getExtraIds();
	ids.push(extras[0]);
	ids.push(extras[1]);
	return ids;
}

function getExtraIds() {
	return [
		{name: "site_id", value: 100},
		{name: "version", value: 2}
	];
}

function getNewSites() {
	return [
		{
			id: FIRST_SITE_ID,
			name: "V.A. National",
			descr: "Pocket Cards for the Department of Veterans Affairs",
			status: "active",
			effectiveDate: new Date("01/01/2015") 
		}
	];
}

function getNewUsers() {
	return [
		{
			id: 50,
			lastname: "",
			firstname: "",
			email: "",
			activeDirId: "",
			roles: [
				{name: "app"}
			],
			effectiveDate: new Date("1/1/2015"),
			status: "active"
		},
		{
			id:51,
			lastname: "",
			firstname: "",
			email: "",
			activeDirId: "",
			roles: [
				{
					name: "site",
					siteId: FIRST_SITE_ID
				}
			],
			effectiveDate: new Date("1/1/2015"),
			status: "active"
		},
		{
			id: ,
			lastname: "",
			firstname: "",
			email: "",
			activeDirId: "",
			roles: [
				{
					name: "card",
					siteId: FIRST_SITE_ID
				}
			],
			effectiveDate: new Date("1/1/2015"),
			status: "active"
		}
	];
}
