/* global process */
//Mongo_dao.js
var log4js = require('log4js');
var utils = require('../../common/utils');
var log = log4js.getLogger('Mongo_dao');

var returnFields = {id:1, title:1, descr:1, tags:1, version:1, versionComment:1, status:1, effectiveDate:1, siteId:1};

var MongoDao = function (params, dbObject) {
	if (dbObject) {
		this.mongo = dbObject;
	}
	else {
		this.mongo = require('mongodb').MongoClient;
	}
	this.config = null;
	if (params) {
		this.config = params;
		log.debug('dao setup with input parameters');
	}
	else {
		log.debug('setting up from -c or --config arguments');
		var processArgs = require('minimist');
		var args = processArgs(process.argv.slice(1));
		if (args.c) {
			args.config = args.c;
		}
		this.config = require('../' + args.config);
	}

	if (!this.config.url) {
		throw new Error('no URL given');
	}
	if (!this.config.dbName) {
		throw new Error('no dbName given');
	}
	this.config.connectString = this.config.url + this.config.dbName;
};

MongoDao.prototype.getAllCurrentCards = function (reqObj, callback) {
	if (!callback) {
		callback = reqObj;
		reqObj = {};
	}
	if (reqObj.ids) {
		this.getAllCurrentCardsByIds(reqObj.ids, callback);
	}
	else {
		this.getAllCurrentCardsByStatus(reqObj.status, callback);
	}
};

MongoDao.prototype.getCardContent = function (id, callback) {
	if (!id) {
		return callback(null, null);
	}
	this.mongo.connect(this.config.connectString, function (err,db) {
		if (err) {
			log.error(err);
			return callback(err);
		}
		var now = new Date();
		db.collection('cards').findOne({id: parseInt(id), effectiveDate: {$lte: now}}, {sort: [['effectiveDate', -1]]}, function (err, result) {
			db.close();
			if (err) {
				log.error(err);
				return callback(err);
			}
			//log.debug('getCardContent result:')
			//log.debug(result);
			if (result && 'active'==result.status) {
				//log.debug('sending active result');
				callback(null, result.file);
			}
			else {
				//log.debug('sending empty non-active result');
				callback();
			}
		});
	});
};

MongoDao.prototype.saveMetrics = function(hash, ip, req, date, callback) {
	//log.debug("saveMetrics > " + hash);
	this.mongo.connect(this.config.connectString, function (err, db) {
		if (err) {
			log.error(err);
			return callback(err);
		}
		db.collection('metrics').insert({deviceHash: hash, clientIp: ip, request: req, timestamp: date}, function (err, results) {
			db.close();
			if (err) {
				log.error(err);
				return callback(err)
			}
			callback();
		});
	});
};

MongoDao.prototype.getAllCurrentCardsByIds = function (ids, callback) {
	//log.debug('getAllCurrentCardsByIds called with:');
	//log.debug(ids);
	if (!ids) {
		return callback(null, []);
	}
	if (!Array.isArray(ids)) {
		ids = [ids];
	}
	ids = normalizeToInts(ids);

	this.mongo.connect(this.config.connectString, function (err, db) {
		if (err) {
			log.error(err);
			return callback(err);
		}
		var results = [];
		getCardMeta(db, new Date(), ids, results, function (err) {
			db.close();
			if (err) {
				return callback(err);
			}
			callback(null, results);
		});
	});
};

MongoDao.prototype.getAllCurrentCardsByStatus = function (status, callback) {
	//log.debug('getAllCurrentCardsByStatus called with:');
	//log.debug(status);
	this.mongo.connect(this.config.connectString, function (err, db) {
		if (err) {
			log.error(err);
			return callback(err);
		}
		var now = new Date();
		var finalMatch = {
			match: true
		};

		if (status) {
			finalMatch['data.status'] = status;
		}

		var pipeline = [
			{
				$match: {
					effectiveDate: {$lte: now}
				}
			},
			{
				$group: {
					_id: "$id",
					effectiveDate: {$max: "$effectiveDate"},
					data: {$push: {effectiveDate: "$effectiveDate", status: "$status", title: "$title", descr: "$descr", version:"$version", versionComment:"$versionComment", tags:"$tags", siteId: "$siteId"}}
				}
			},
			{
				$unwind: "$data"
			},
			{
				$project: {
					data: 1,
					match: {$eq: ["$effectiveDate", "$data.effectiveDate"]}
				}
			},
			{
				$match: finalMatch
			},
			{
				$sort: {"data.title":1}
			}
		];
		db.collection('cards').aggregate(pipeline, {out: {inline:1}}, function (err, results) {
			db.close();
			if (err) {
				log.error(err);
				return callback(err);
			}
			log.debug('aggregate results:');
			log.debug(results);
			if (results && results.length) {
				var ary = [];
				for (var i = results.length - 1; i >= 0; i--) {
					var item = utils.getMergedInto({id: results[i]._id}, results[i].data);
					ary.push(item);
				};
				callback(null, ary);
			}
			else {
				callback(null, []);
			}
		});
	});
};

MongoDao.prototype.getSites = function (callback) {
	this.mongo.connect(this.config.connectString, function (err, db) {
		if (err) {
			log.error(err);
			return callback(err);
		}
		var now = new Date();
		var finalMatch = {
			match: true
		};
		finalMatch['data.status'] = 'active';

		var pipeline = [
			{
				$match: {
					effectiveDate: {$lte: now}
				}
			},
			{
				$group: {
					_id: "$id",
					effectiveDate: {$max: "$effectiveDate"},
					data: {$push: {effectiveDate: "$effectiveDate", status: "$status", name: "$name", descr: "$descr"}}
				}
			},
			{
				$unwind: "$data"
			},
			{
				$project: {
					data: 1,
					match: {$eq: ["$effectiveDate", "$data.effectiveDate"]}
				}
			},
			{
				$match: finalMatch
			},
			{
				$sort: {"data.name":1}
			}
		];
		db.collection('sites').aggregate(pipeline, {out: {inline:1}}, function (err, results) {
			db.close();
			if (err) {
				log.error(err);
				return callback(err);
			}
			//log.debug('aggregate results:');
			//log.debug(results);
			if (results && results.length) {
				var ary = [];
				for (var i = results.length - 1; i >= 0; i--) {
					var item = utils.getMergedInto({id: results[i]._id}, results[i].data);
					ary.push(item);
				};
				callback(null, ary);
			}
			else {
				callback(null, []);
			}
		});
	});
}

function getCardMeta(db, now, ids, results, callback) {
	var id = ids.pop();
	if (!id) {
		return callback(null, results);
	}
	db.collection('cards').findOne({id: id, effectiveDate: {$lte: now}}, returnFields, {sort: [['effectiveDate', -1]]}, function (err, result) {
		if (err) {
			return callback(err);
		}
		// log.debug('getCardMeta for ' + id + ' returned:');
		// log.debug(result);
		if (result) {
			results.push(result);
		}
		getCardMeta(db, now, ids, results, callback);
	});
}

function normalizeToInts(ary) {
	if (!ary) {
		return null;
	}
	var result = [];
	for(var i=0;i<ary.length;i++) {
		if (ary[i]) {
			var id = parseInt(ary[i]);
			if ('NaN'!=id.toString()) {
				result.push(id);
			}
		}
	}
	return result;
}

module.exports = MongoDao;