/* global process */
/*jslint node: true */

var numCPUs = require('os').cpus().length;
var cluster = require('cluster');
var parseArgs = require('minimist');
var args = parseArgs(process.argv.slice(1));
var log4js = require('log4js');
var requestIp = require('request-ip');

var log = log4js.getLogger('server.js');

process.on('uncaughtException', function(err) {
    log.error(err);
});

log.info("Running in Node %s on %s[%s] under %s", process.version, process.platform, process.arch, process.cwd());

log.debug('args=' + JSON.stringify(args));

var package = require('./package').config;
var config = null;

if (args.c) {
    args.config = args.c;
}

var dao = null;
if (args.config) {
    var config = require('./' + args.config);
    var Dao = require(config.dao);
    dao = new Dao();
}
else {
    throw new Error("No DAO configuration specified");
}

if (cluster.isMaster) {
    log.info("Master PID: %s", process.pid);

    // Temporary fix for debugging child processes.
    // See https://github.com/joyent/node/issues/5318 for more info.
    var debug = process.execArgv.indexOf('--debug') !== -1;
    cluster.setupMaster({
        execArgv: process.execArgv.filter(function (s) {
            'use strict';
            return s !== '--debug';
        })
    });

    for (var i = 0; i < numCPUs; i++) {
        if (debug) cluster.settings.execArgv.push('--debug=' + (5858 + i));
        cluster.fork();
        if (debug) cluster.settings.execArgv.pop();
    }

    cluster.on("online", function (worker) {});

    cluster.on("listening", function (worker, address) {});

    // Create a new process on disconnect event.
    cluster.on('disconnect', function (worker) {
        'use strict';

        console.error('disconnect!');
        cluster.fork();
    });
} else {
    // Node v10 cluster debugging workaround!
    // Make the debuging port unigue for each process.
    process.debugPort = 5858 + cluster.worker.id;

    log.info("Child id: %s, PID: %s, Debug Port: %s", cluster.worker && cluster.worker.id, process.pid, process.debugPort);

    var domain = require("domain");
    var http = null;
    if (package.https) {
        http = require('https');
    }
    else {
        http = require('http');
    }
    var express = require('express');

    var app = express();

    // http://stackoverflow.com/questions/11181546/node-js-express-cross-domain-scripting
    app.all('*', function (request, response, next) {
        'use strict';
        log.debug('URL:' + request.originalUrl);

        response.header('Access-Control-Allow-Origin', '*');
        response.header('Access-Control-Allow-Methods', 'PUT, GET, POST, DELETE, OPTIONS');
        response.header('Access-Control-Allow-Headers', 'Content-Type');
        next();
    });


    // Start the server
    // Port can be configured by developers using `npm config set processPort 1111`
    // This will override the port defined in package.json
    // Application should be started using `npm start`
    app.set('port', process.env.npm_config_processPort || package.processPort);

    var server = null;

    if (package.https) {
        var fs = require('fs');
        var httpsOptions = {
            key: fs.readFileSync(package.keyFileName),
            cert: fs.readFileSync(package.certFileName)
        };
        server = http.createServer(httpsOptions, app);
    }
    else {
        server = http.createServer(app);
    }

    //The 3rd parameter below is the default for this call
    server.listen(app.get('port'), '0.0.0.0', 511, function () {
        'use strict';
        // Create a Domain error handler. This provides a safe way to shutdown a process.
        // when unexpected errors occur. See //http/nodejs.org/api/domain.html
        var d = domain.create();
        d.on('error', function (error) {
            try {
                // Make sure we close down within 30 seconds...
                var killtimer = setTimeout(function () {
                    process.exit(1);
                }, 30000);

                // But don't keep the process open just for that!
                killtimer.unref();

                // Stop taking new requests...
                server.close();

                // Let the master know we're dead.  This will trigger a
                // 'disconnect' in the cluster master, and then it will fork
                // a new worker...
                cluster.worker.disconnect();
            } catch (er2) {
                // oh well, not much we can do at this point.
                console.error('Error sending 500!', er2.stack);
            }
        });
        log.info('Server listening on port ' + server.address().port);
    });

    /**
     * Function to save the device hash for all routes.
     *
     * @param  {[type]}   request
     * @param  {[type]}   response
     * @param  {Function} next
     * @param  {[type]}   device_hash
     * @return {[type]}
     */
    app.param('device_hash', function (request, response, next, device_hash) {
        'use strict';
        dao.saveMetrics(device_hash, requestIp.getClientIp(request), request.originalUrl, new Date(), function (err) {
            if (err) {
                log.error(err)
            }
            next();
        });
    });

    /**
     * Function to return all Cards. A query parameter can be specified to filter the
     * Cards returned. Ex. /pocketcardv2/user/:device_hash/cards?state=active.
     *
     * @param  {[type]} request
     * @param  {[type]} response
     * @return {[type]}
     */
    app.get('/:service_id/user/:device_hash/cards', function (request, response) {
        'use strict';

        if (!isValidServiceId(request.params.service_id)) {
            return response.send(500).end();
        }

        try {
            log.debug("Get all pocket cards [" + request.params.device_hash + "]....");

            // Check for filter criteria...
            var state = 'active';
            if (request.query.state) {
                state = encodeURI(request.query.state);
            }
            var reqObj = {
                status: state
            };
            dao.getAllCurrentCards(reqObj, function (err, result) {
                if (err) {
                    log.error(err);
                    return response.send(500).end();
                }
                log.debug('Get all cards > Number of rows returned [' + request.params.device_hash + ']: ' + result.length);
                response.status(200).send(JSON.stringify(result));
            });
        } catch (err) {
            log.error(err);
            response.send(500).end();
        }
    });

    /**
     * Function to return a Pocket Card for a given id.
     *
     * @param  {[type]} request
     * @param  {[type]} response
     * @return {[type]}
     */
    app.get('/:service_id/user/:device_hash/cards/:id', function (request, response) {
        'use strict';

        if (!isValidServiceId(request.params.service_id)) {
            return response.send(500).end();
        }

        try {
            log.debug("Get pocket card by id=" + request.params.id + " [" + request.params.device_hash + "]....");

            // Check for filter criteria...
            var ids = encodeURI(request.params.id).split(',');
            dao.getAllCurrentCards({
                ids: ids
            }, function (err, result) {
                if (err) {
                    log.error(err);
                    return response.send(500).end();
                }
                log.debug("Get cards by id=" + request.params.id + "> Number of rows returned [" + request.params.device_hash + "]: " + result.length);
                response.status(200).send(JSON.stringify(result));
            });
        } catch (err) {
            log.error(err);
            response.send(500).end();
        }
    });

    /**
     * Function to return the Pocket Card contents.
     *
     * @param  {[type]} request
     * @param  {[type]} response
     * @return {[type]}
     */
    app.get('/:service_id/user/:device_hash/cards/:id/contents', function (request, response) {
        'use strict';

        if (!isValidServiceId(request.params.service_id)) {
            return response.send(500).end();
        }

        try {
            log.debug("Get pocket card content for " + request.params.id + " [" + request.params.device_hash + "]....");
            dao.getCardContent(request.params.id, function (err, result) {
                if (err) {
                    throw err;
                }
                // Send response back to client.
                response.status(200).send(result);
            });
        } catch (err) {
            log.error(err);
            response.send(500).end();
        }
    });
    
    app.get('/pocketcardv2/user/:device_hash/sites', function (req, res) {
        'use strict';

        try {
            log.debug('get all active sites [' + req.params.device_hash + ']...');
            dao.getSites(function (err, sites) {
                if (err) throw err;
                res.status(200).send(sites);
            });
        }
        catch (err) {
            log.error(err);
            res.send(500).end();
        }
    });

}

function isValidServiceId(sid) {
    if (sid) {
        var _sid = sid.toString().toLowerCase();
        return 'pocketcardv1'===_sid || 'pocketcardv2'===_sid;
    }
    return false;
}
